/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<limits.h>
#include<float.h>
#include"globals.h"
#include"node.h"
#include"twister.h"
#include"utilities.h"
#include"event.h"

/******************************************************************************/
/*                               NewEvent                                     */
/******************************************************************************/
EVENT NewEvent(int ev_type, double ev_time, long l_id, double current_time)
{
  // DEBUGGING
  //printf("NETGEN DEBUG:  we have been asked to make a new event type %d, at time %lf, for id %ld\n", ev_type, ev_time, l_id);

  // make the new event
  EVENT ev;
  ev = (EVENT)malloc(sizeof(EVENTTYPE));

  if (ev == NULL)
  {
    printf("NETGEN ERROR: Could not allocate space for event\n");
    exit(0);
  }

  // use the values we were given
  ev->event_type = ev_type;
  ev->event_time = ev_time;
  ev->time_created = current_time;
  ev->lin_id = l_id;

  // set defaults
  ev->sched_status = 1;
  ev->previous_event = NULL;
  ev->next_event = NULL;
  ev->second_lin_id = -1;
  ev->second_parent_time = -1;

  return ev;

}  // end of NewEvent

/******************************************************************************/
/*                             EstblishEvent                                  */
/******************************************************************************/
void EstablishEvent(EVENTPTR queue_ptr, LINEAGE lin, double current_time)
{
  // using the lineage's rates for birth, death, and hybridization, this routine
  // figures out what the lineage's first event will be and places it on the
  // queue (note that a lineage will have at most one thing happen to
  // it -- if a birth or hybrid, new lineages are born and the original
  // lineage is retired, if a death, the lineage simply dies)

  int next_event;
  double time_to_event, ev_time;
  EVENT ev;

  // decide what event will be scheduled and how long until that event
  next_event = FindNextEvent(lin, &time_to_event);

  // create event
  ev_time = current_time + time_to_event;
  ev = NewEvent(next_event, ev_time, lin->id, current_time);

  // now schedule the event in the event queue
  ScheduleNewEvent(queue_ptr, ev);
}

/******************************************************************************/
/*                             FindNextEvent                                  */
/******************************************************************************/
int FindNextEvent(LINEAGE lin, double *time_to_event)
{
  int next_event_type;
  double b_time, d_time, h_time;
  
  // we need to know what the next inter-arrival time for each event is
  
  // next birth time interval
  if (lin->birth_rate > 0.0)
  {
    b_time = NextExpArrInterval(lin->birth_rate);

    // DEBUGGING
    //printf("NETGEN DEBUG:  next birth inter-arrival time for lin %ld is: %lf\n", lin->id, b_time);

  }
  else
    b_time = DBL_MAX;

  // next death time interval
  if (lin->death_rate > 0.0)
  {
    d_time = NextExpArrInterval(lin->death_rate);

    // DEBUGGING
    //printf("NETGEN DEBUG:  next death inter-arrival time for lin %ld is: %lf\n", lin->id, d_time);

  }
  else
    d_time = DBL_MAX;

  // next hybrid time interval
  if (lin->hybrid_rate > 0.0)
  {
    h_time = NextExpArrInterval(lin->hybrid_rate);

    // DEBUGGING
    //printf("NETGEN DEBUG:  next hybrid inter-arrival time for lin %ld is: %lf\n", lin->id, h_time);

  }
  else
    h_time = DBL_MAX;


  // sanity check -- if all the rates are <= 0, we have a problem and the
  // simulation should not continue
  if ( (lin->birth_rate <= 0.0) && 
       (lin->death_rate <= 0.0) && 
       (lin->hybrid_rate <= 0.0) )
  {
    printf("NETGEN ERROR:  Something is amiss with our FindNextEvent routine as all event rates are 0 or less, the simulation cannot continue\n");
    exit(0);
  }

  (*time_to_event) = MinTime(b_time, d_time, h_time);
  next_event_type = FindMinEvent(b_time, d_time, h_time);

  // sanity check -- if our time to next event is DBL_MAX, something
  // is amiss with our simulation, so abort
  if ( (*time_to_event) == DBL_MAX )
  {
    printf("NETGEN ERROR:  Something is amiss with our FindNextEvent routine as the time to next event is %lf and DBL_MAX is %lf\n", (*time_to_event), DBL_MAX);
    exit(0);
  }

  //DEBUGGING
  //printf("NETGEN DEBUG:  From FindNextEvent for lineage %ld, our time to the next event is %lf and the event type will be %d\n", lin->id, (*time_to_event), next_event_type);

  return(next_event_type);
}

/******************************************************************************/
/*                             FindMinEvent                                   */
/******************************************************************************/
int FindMinEvent(double t1, double t2, double t3)
{
  int event;
  double event_time;

  if ( (t1 <= t2) && (t1 <= t3) ) // t1 is the min
  {
    event = 1;
    event_time = t1;
  }

  if ( (t2 <= t1) && (t2 <= t3) ) // t2 is the min
  {
    event = 2;
    event_time = t2;
  }

  if ( (t3 <= t1) && (t3 <= t2) ) // t3 is the min
  {
    event = 3;
    event_time = t3;
  }

  // if we don't have a 3-way tie, look for a 2-way tie
  // and decide the event randomly
  if (! ((t1 == t2) && (t2 == t3)) ) // not 3-way
  {
    if ((event_time == t1) && (event_time == t2)) 
    {
      if (genrand_real2() < .5)
        event = 1;
      else
        event = 2;
    }
    if ((event_time == t1) && (event_time == t3)) 
    {
      if (genrand_real2() < .5)
        event = 1;
      else
        event = 3;
    }
    if ((event_time == t2) && (event_time == t3)) 
    {
      if (genrand_real2() < .5)
        event = 2;
      else
        event = 3;
    }
  }
  
  // check to see if we had a 3-way tie & decide it
  if ( (t1 == t2) && (t2 == t3) )
  {
    // get a rand int on the range of [0-2]
    event = (int) (genrand_real2()*3);

    // scale up for 1-3 range
    event = event + 1;
  }
  
  return event;
}

/******************************************************************************/
/*                          InheritParentalEvent                              */
/******************************************************************************/
void InheritParentalEvent(EVENTPTR queue_ptr, long parent_lin_id, long lin_id)
{
  EVENT ev;
  EVENT ev_i_ptr;
  EVENT ev_to_inherit;

  // here we find the parental event, make a new event with the same info
  // except for the lineage id, and schedule the new event on the queue
  
  // go through the queue and find the event which matches on the 
  // parental lineage
  ev_i_ptr = (*queue_ptr);
  ev_to_inherit = NULL;

  while (ev_i_ptr != NULL)
  { if (ev_i_ptr->lin_id == parent_lin_id)
    {
      ev_to_inherit = ev_i_ptr;
      ev_i_ptr = NULL;
    }
    else
    {
      ev_i_ptr = ev_i_ptr->next_event;
    }
  }

  // set up the new event and schedule it

  if (ev_to_inherit != NULL)
  {
    // make the new event and set all the parameters from the parental one
    // (including the time it was created -- we want time between creation
    //  and occured to be correct, even though this means the lineage will
    //  have been created AFTER its event was scheduled!) except for the 
    //  new lineage id!

    ev = NewEvent(ev_to_inherit->event_type, ev_to_inherit->event_time, lin_id, ev_to_inherit->time_created); 

    ScheduleNewEvent(queue_ptr, ev);
    
  }
  else
  {
    printf("NETGEN ERROR:  Could not find parental event for lineage %ld to inherit\n", lin_id);
    exit(0);
  }

}


/******************************************************************************/
/*                             ScheduleEvent                                  */
/******************************************************************************/
void ScheduleNewEvent(EVENTPTR queue_ptr, EVENT ev)
{
  EVENT temp_ev_ptr;
  EVENT temp_prev_ev_ptr;

  // first see if the queue is currently empty, making this the first
  // item in the queue
  if ((*queue_ptr) == NULL)
  {
    (*queue_ptr) = ev; 
  }
  // next see if this ev will become the head of the existing queue 
  //    if so: 
  //           make sure the old head gets is previous event set to the new head
  //           set event's next event correctly
  //           reset head of event queue to this event
  //
  else if (ev->event_time < (*queue_ptr)->event_time)
  {
    ev->next_event = (*queue_ptr);
    ev->next_event->previous_event = ev;
    (*queue_ptr) = ev;
  }
  else // find its correct location in the list and put it there
  {
    temp_ev_ptr = (*queue_ptr);
    temp_prev_ev_ptr = NULL;

    while ((temp_ev_ptr != NULL) && (ev->event_time >= temp_ev_ptr->event_time))
    {
      temp_prev_ev_ptr = temp_ev_ptr;
      temp_ev_ptr = temp_ev_ptr->next_event;
    }

    // see if we have the tail case and tack on the event at the end
    if (temp_ev_ptr == NULL)
    {
      temp_prev_ev_ptr->next_event = ev;
      ev->previous_event = temp_prev_ev_ptr;
    }
    else // otherwise put it in between the two events where it should be
    {
      temp_prev_ev_ptr->next_event = ev;
      ev->previous_event = temp_prev_ev_ptr;
      ev->next_event = temp_ev_ptr;
      temp_ev_ptr->previous_event = ev;
    }
  }
}


/******************************************************************************/
/*                             GetNextEvent                                   */
/******************************************************************************/
EVENT GetNextEvent(EVENTPTR queue_ptr)
{
  EVENT n_ev;
  EVENT temp_ev_ptr;
  int   chosen_ev_index, ctr, same_time_count;

  /****************************************************************/
  /*                        sanity check                          */
  /* we should only call this routine when there are events left! */
  /****************************************************************/
  if ((*queue_ptr) == NULL)
  {
    printf("NETGEN ERROR:  GetNextEvent was called for an empty queue!\n");
    exit(0);
  }

  /****************************************************************/
  /*   see if we have a distinct next event or there is a tie     */
  /****************************************************************/
  if ( ((*queue_ptr)->next_event == NULL) ||
       ((*queue_ptr)->event_time < (*queue_ptr)->next_event->event_time) )
  {
    n_ev = (*queue_ptr);
    (*queue_ptr) = n_ev->next_event;
  }
  else 
  {
    /**************************/
    /* do a random tie-breker */
    /**************************/

    // count how many events have the same next time
    temp_ev_ptr = (*queue_ptr)->next_event;
    same_time_count = 1;

    while ( (temp_ev_ptr != NULL) &&
	    ((*queue_ptr)->event_time == temp_ev_ptr->event_time) )
    {
      same_time_count++;
      temp_ev_ptr = temp_ev_ptr->next_event;
    }

    // DEBUGGING
    //printf("NETGEN DEBUG:  we have %d events tied to be the next event\n", same_time_count);
    // END DEBUGGING

    // chose an index randomly among those with the same next time
    // do random value [0,1) and then multiply by number of events
    // with the same time to give [0, num-1] random selection
    chosen_ev_index = (int) (genrand_real2() * same_time_count);

    // DEBUGGING
    //printf("NETGEN DEBUG:  chosen index for random pull from queue for tied events is %d\n", chosen_ev_index);
    // END DEBUGGING
    
    // find and pull out the chosen event based on index value
    ctr = 0;
    temp_ev_ptr = (*queue_ptr);
    while (ctr < chosen_ev_index)
    {
       ctr++;
       temp_ev_ptr = temp_ev_ptr->next_event;
    }
    n_ev = temp_ev_ptr;

    // update the surrounding events if present

    // if event is at head of queue (and we know there are at least 2 events
    // in the queue, otherwise we would not have had a tie)
    if (temp_ev_ptr == (*queue_ptr))
    {
      temp_ev_ptr->next_event->previous_event = NULL;
      (*queue_ptr) = temp_ev_ptr->next_event;
      temp_ev_ptr->next_event = NULL;
    }
    // if event is at the end of the queue
    else if (temp_ev_ptr->next_event == NULL)
    {
      temp_ev_ptr->previous_event->next_event = NULL;
      temp_ev_ptr->previous_event = NULL;
    }
    // it is between two events
    else
    {
      temp_ev_ptr->previous_event->next_event = temp_ev_ptr->next_event;
      temp_ev_ptr->next_event->previous_event = temp_ev_ptr->previous_event;
    }
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: about to return next event which is for lineage %ld\n", n_ev->lin_id);

  return n_ev;
}


/******************************************************************************/
/*                             CancelLinEvent                                 */
/******************************************************************************/
void CancelLinEvent(EVENTPTR ev_q_ptr, EVENTPTR old_ev_list_ptr, long l_id)
{
  EVENT ev_i_ptr;
  EVENT ev_to_cancel;

  // go through the queue and find the event which matches on the lineage
  // id -- there will only be one
  ev_i_ptr = (*ev_q_ptr);
  ev_to_cancel = NULL;

  while (ev_i_ptr != NULL)
  { if (ev_i_ptr->lin_id == l_id)
    {
      ev_to_cancel = ev_i_ptr;
      ev_i_ptr = NULL;
    }
    else
    {
      ev_i_ptr = ev_i_ptr->next_event;
    }
  }
  if (ev_to_cancel != NULL)
  {
    // first remove it from the event queue
    
    // if event to cancel is at the head of the queue
    if (ev_to_cancel == (*ev_q_ptr))
    {
      (*ev_q_ptr) = (*ev_q_ptr)->next_event;

      // if it wasn't the only thing remaining in the queue
      // correct the new head's previous pointer
      if ((*ev_q_ptr) != NULL)
        (*ev_q_ptr)->previous_event = NULL;
    }
    else
    {
      // event to cancel is not at the end of queue
      if (ev_to_cancel->next_event != NULL)
      {
        ev_to_cancel->next_event->previous_event = ev_to_cancel->previous_event;
        ev_to_cancel->previous_event->next_event = ev_to_cancel->next_event;
      }
      else
      {
        ev_to_cancel->previous_event->next_event = NULL;
      }

      ev_to_cancel->previous_event = NULL;
      ev_to_cancel->next_event = NULL;
    }

    // now set it up as an old event
    ev_to_cancel->sched_status = 3;
    LogOldEvent(old_ev_list_ptr, ev_to_cancel);
  }
  else
  {
    printf("NETGEN ERROR:  Event to cancel (for lin id %ld) was not found!\n", l_id);
    exit(0);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG:  Just finished cancel (for lin id %ld)\n", l_id);
}


/******************************************************************************/
/*                             LogOldEvent                                    */
/******************************************************************************/
void LogOldEvent(EVENTPTR list_ptr, EVENT ev)
{
  // each old event is ALWAYS placed at the front of the list (like a stack) 
  // so when we print them, we simply print back to front to get the order
  // correct

  ev->next_event = (*list_ptr);
  ev->previous_event = NULL;

  if ((*list_ptr) != NULL)
    (*list_ptr)->previous_event = ev;

  (*list_ptr) = ev;
}

