/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include"globals.h"
#include"twister.h"
#include"utilities.h"
#include"node.h"
#include"event.h"
#include"network.h"
#include"sequences.h"
#include"hybrids.h"
#include"output.h"
#include"eventexec.h"



/******************************************************************************/
/*                          Birth/Speciation Event Work                       */
/******************************************************************************/
void BirthEvent(NETWORK nt, EVENT ev, double current_time)
{
  long chosen_lin, first_index, second_index;
  int nhc, nsphc, sl;
  NODE parent_ptr;

  // if we are doing updates for all events, we need to update branch
  // lengths and sequences here for all active lineages, but not later
  // in this routine....
  if (update_seqs_for_all_events_yn)
  {
    UpdateAllBranchLengths(nt, current_time);
    UpdateAllSequences(nt, current_time);
  }

  first_index = (nt->node_id_ctr)++;
  second_index = (nt->node_id_ctr)++;
  
  // the event tells us which lineage will be the one speciating
  chosen_lin = ev->lin_id;

  // assign time and set inactive
  nt->NodeArray[chosen_lin]->time_created = current_time;
  nt->NodeArray[chosen_lin]->active = 0;

  // assign branch lengths and sequences only if not done at beginning 
  // of routine -- otherwise they are already done and should not 
  // repeat the process!
  if (update_seqs_for_all_events_yn == 0)
  {
    BirthDeathSingleBranch(nt, chosen_lin, current_time);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: we are doing a birth, the node's id was %ld and the current time is %f with its parent's time being %f giving us a branch length of %f\n", nt->NodeArray[chosen_lin]->id, current_time, nt->NodeArray[chosen_lin]->parent1->time_created, nt->NodeArray[chosen_lin]->brlen1_t);
  // END DEBUGGING

  // make the two new lineages which are spawned due to this birth/speciation
  // they will gain their parent's info (which is the chosen_lin here) for
  // some values and will get their rates in NewLineage -- and schedule their
  // events into the event queue
  
  nhc = nt->NodeArray[chosen_lin]->num_hom_chromos;
  nsphc = nt->NodeArray[chosen_lin]->num_strands_per_hom_chrom;
  sl =  nt->NodeArray[chosen_lin]->seq_length;
  parent_ptr = nt->NodeArray[chosen_lin];

  // first new lineage
  nt->NodeArray[first_index] = NewLineage(first_index, nhc, nsphc, sl, parent_ptr);
  nt->NodeArray[first_index]->seq_length = parent_ptr->seq_length;
  EstablishEvent(&(nt->event_queue), nt->NodeArray[first_index], current_time);
  
  // second new lineage
  nt->NodeArray[second_index] = NewLineage(second_index, nhc, nsphc, sl, parent_ptr);
  nt->NodeArray[second_index]->seq_length = parent_ptr->seq_length;
  EstablishEvent(&(nt->event_queue), nt->NodeArray[second_index], current_time);

  // now link in the new lineages
  ConnectTwoChildren(nt->NodeArray[chosen_lin], nt->NodeArray[first_index], nt->NodeArray[second_index]);

  // counter maintenance
  nt->num_act_lins += 2;
  nt->num_nodes += 2;
  nt->num_edges += 2;

  // now set the chosen lineage as finished (no longer active)
  nt->NodeArray[chosen_lin]->active = 0;

  // reduce the number of ACTIVE lineages, by the one that ended in this event 
  (nt->num_act_lins)--;

  // mark the event as processed
  ev->sched_status = 2;

  // DEBUGGING
  //printf("NETGEN DEBUG: we have just completed a speciation\n");
  
}


/******************************************************************************/
/*                     Death/Extinction Event Work                            */
/******************************************************************************/
void DeathEvent(NETWORK nt, EVENT ev, double current_time) 
{
  long chosen_lin;

  // if we are doing updates for all events, we need to update branch
  // lengths and sequences here for all active lineages, but not later
  // in this routine....
  if (update_seqs_for_all_events_yn)
  {
    UpdateAllBranchLengths(nt, current_time);
    UpdateAllSequences(nt, current_time);
  }

  // the event tells us which lineage will be the one going extinct
  chosen_lin = ev->lin_id;

  // set the time, leaf, and inactive values
  nt->NodeArray[chosen_lin]->time_created = current_time;
  nt->NodeArray[chosen_lin]->leaf = 1;
  nt->NodeArray[chosen_lin]->active = 0;

  // assign branch lengths and sequences only if not done at beginning 
  // of routine -- otherwise they are already done and should not 
  // repeat the process!
  if (update_seqs_for_all_events_yn == 0)
  {
    BirthDeathSingleBranch(nt, chosen_lin, current_time);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: we have just done an extinction, the node's id was %ld and the current time is %f with its parent's time being %f giving us a clock branch length of %f\n", nt->NodeArray[chosen_lin]->id, current_time, nt->NodeArray[chosen_lin]->parent1->time_created, nt->NodeArray[chosen_lin]->brlen1_t);
  // END DEBUGGING


  // counter maintenance
  (nt->num_act_lins)--;
  (nt->num_leaves)++;

  // mark the event as processed
  ev->sched_status = 2;

  // DEBUGGING
  //printf("NETGEN DEBUG: we have just completed a death\n");
}

/******************************************************************************/
/*                            HybridEventType                                 */
/******************************************************************************/

// This routine decides which type of hybrid event has occured.
// Currently diploid and polyploid hybridizations are implemented.  
//
//     1 : diploid hybridization
//     2 : polyploid hybridization

int HybridEventType() 
{
  double rnd_value;

  rnd_value = genrand_real2();

  if (rnd_value < dip_rate) 
    return 1;
  else if ( (dip_rate <= rnd_value) && (rnd_value < (dip_rate + poly_rate)) )
    return 2;
  else
  {
    printf("NETGEN ERROR:  Only diploid and polypolid hybridizations are implemented\n");
    printf("NETGEN ERROR:  dip rate: %f, poly rate: %f,  rand value %f\n", dip_rate, poly_rate, rnd_value);
    exit(0);
  }

}

/******************************************************************************/
/*                         BirthDeathSingleBranch                             */
/******************************************************************************/
void BirthDeathSingleBranch(NETWORK nt,  long chosen_lin, double current_time) 
{
  double time_distance;
  int prev_update_yn;

  if (fake_branch_lengths_yn)
  {
    nt->NodeArray[chosen_lin]->brlen1_ev = FakeBranchLength();
    nt->NodeArray[chosen_lin]->brlen1_t = nt->NodeArray[chosen_lin]->brlen1_ev;
  }
  else if (ultrametric_yn == 0) // doing non-ultrametric network
  {
    // evolutionary time is deviated, clock time is not
    time_distance = (nt->NodeArray[chosen_lin]->time_created - 
                     nt->NodeArray[chosen_lin]->parent1->time_created);

    nt->NodeArray[chosen_lin]->brlen1_t =  time_distance;

    // take time and scale according to desired size and finally multiply
    // by the lineage's gamma value so we will have a non-ultrametric network
    nt->NodeArray[chosen_lin]->brlen1_ev = time_distance * 
                                           height_scalar *
	       			           nt->NodeArray[chosen_lin]->gamma_value;

    // DEBUGGING
    //printf("NETGEN DEBUG:  Just made a non-ultrametric branch length from %ld to %ld of clock length: %lf, and ev length %lf\n", nt->NodeArray[chosen_lin]->id, nt->NodeArray[chosen_lin]->parent1->id, nt->NodeArray[chosen_lin]->brlen1_t, nt->NodeArray[chosen_lin]->brlen1_ev);
    // END DEBUGGING
  }
  else // regular ultrametric branch lengths
  {
    // evolutionary time is scaled, clock time is not
    time_distance = (nt->NodeArray[chosen_lin]->time_created - 
                     nt->NodeArray[chosen_lin]->parent1->time_created);

    nt->NodeArray[chosen_lin]->brlen1_t =  time_distance;

    nt->NodeArray[chosen_lin]->brlen1_ev = time_distance * height_scalar;

    // DEBUGGING
    //printf("NETGEN DEBUG:  Just made an ultrametric branch length from %ld to %ld of clock length: %lf, and ev length %lf\n", nt->NodeArray[chosen_lin]->id, nt->NodeArray[chosen_lin]->parent1->id, nt->NodeArray[chosen_lin]->brlen1_t, nt->NodeArray[chosen_lin]->brlen1_ev);
    // END DEBUGGING
  }

  // assign the sequences if needed (if not assigning artificial ones!)
  if (sequences_yn)
  {
    if (nt->NodeArray[chosen_lin]->prev_update_time == PREV_UPDATE_TIME_DEFAULT)
      prev_update_yn = 0;
    else
      prev_update_yn = 1;

    AssignSequences(1, nt->NodeArray, chosen_lin, prev_update_yn, current_time);
  } 
  else // tracking sequences artificially
  {
    if (nt->NodeArray[chosen_lin]->prev_update_time == PREV_UPDATE_TIME_DEFAULT)
      prev_update_yn = 0;
    else
      prev_update_yn = 1;

    AssignArtificialSequences(1, nt->NodeArray, chosen_lin, prev_update_yn, current_time);
  } 

}
