/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include "globals.h"
#include "node.h"
#include "event.h"
#include "network.h"
#include "eventexec.h"
#include "hybrids.h"
#include "output.h"
#include "utilities.h"
#include "twister.h"
#include "generate.h"


/******************************************************************************/
/*                           GenerateNetwork                                  */
/******************************************************************************/
void GenerateNetwork(NETWORK nt)
{
  double current_time, time_to_event;
  int loop_count, next_event, hybrid_type;
  EVENT ev;

  loop_count = 0;
  rates_redone_yn = 0;

  hyb_rate_orig = 0.0; 
  hyb_rate_var_orig = 0.0; 
  dip_rate_orig = 0.0; 
  poly_rate_orig = 0.0; 

  /*******************************************************/
  /*           set up random number stream               */
  /*******************************************************/

  if (user_seed_netgen_yn == 0) // no seed provided
    seed_netgen = time(NULL) + clock();
  else
    seed_netgen = user_seed_netgen;
  
  init_genrand(seed_netgen);


  /******************************************************/
  /*                set up time stuff                   */
  /******************************************************/

  current_time = 0.0;
  time_to_event = 0.0;


  /******************************************************/
  /*                 Set up root node                   */
  /******************************************************/

  MakeRootNode(nt, current_time);


  /*****************************************************/
  /*  deal with outgroup rooting if requested by user  */
  /*****************************************************/

  if (outgroup)
  {
    MakeSubRootNode(nt, current_time);
  }


  /*****************************************************/
  /*            create the first two lineages          */
  /*****************************************************/
  // now, no matter what, we need to create the two lineages which will be
  // our active lineages for starting the growing of the network (even if we
  // have an outgroup, we do not count it as an active lineage and it does 
  // not contribute to the desired extant taxa count)
  
  MakeFirstActLins(nt, current_time);

  // no matter what we start with 2 active lineages
  nt->num_act_lins = 2;

  // now setup the event for each these two active lineages
  EstablishEvent(&(nt->event_queue), nt->NodeArray[(nt->node_id_ctr)-2], current_time);
  EstablishEvent(&(nt->event_queue), nt->NodeArray[(nt->node_id_ctr)-1], current_time);

  /********************************************************/
  /*         now do the big processing loop               */
  /********************************************************/

  while ((nt->num_act_lins < num_extant_taxa_req) && //have not reached req size
	 (nt->event_queue != NULL)  )                //still have events
  {
    // DEBUGGING
     //printf("NETGEN DEBUG: We are starting a brand new loop, loop count is %d, act lineage count %ld, num_hybrids %ld, current_time %f  \n", loop_count, nt->num_act_lins, nt->num_hybrids, current_time);
    // END DEBUGGING
    
    /***************************************/
    /*   get the next event off the queue  */
    /***************************************/
    ev = GetNextEvent(&(nt->event_queue));

    next_event = ev->event_type;
    current_time = ev->event_time;

    /***************************************/
    /*        process the event            */
    /***************************************/

    /*********************************/
    /*  birth -- regular speciation  */
    /*********************************/
    if (next_event == 1)
    {
      BirthEvent(nt, ev, current_time);
    } 

    /*********************************/
    /* death -- extinction           */
    /*********************************/
    if (next_event == 2)
    {
      DeathEvent(nt, ev, current_time);
    }

    /*********************************/
    /*         hybrid event          */
    /*********************************/

    if (next_event == 3)
    {
      if ( (hyb_rate > 0.0) && (nt->num_act_lins > 1) )
      {
        // we can only do the hybrid if the rate is still
        // over 0 (meaning the max number of hybrids if
        // specified has not been reached) and we have
        // at least two lineages
	
        // decide what type of hybrid will occur
        hybrid_type = HybridEventType();

        // only implenting diploid and polyploid for now.....
        if ( (hybrid_type != 1) && (hybrid_type != 2) )
        {
          printf("NETGEN ERROR:  Only diploid and polyploid hybrids are implemented\n");
	  exit(0);
        }

	// call the hybrid event and pass the hybrid type
	Hybrid(nt, ev, current_time, hybrid_type);
	 
	// check on hybrid limit
        if ( (-1 < max_num_hybrids) && 
	     (nt->num_hybrids == max_num_hybrids) &&
             (rates_redone_yn == 0) ) 
        {
          // save old rates and remember we did this
          rates_redone_yn = 1;

          hyb_rate_orig = hyb_rate;
          hyb_rate_var_orig = hyb_rate_var;
          dip_rate_orig = dip_rate;
          poly_rate_orig = poly_rate;

          // reset the rates
          hyb_rate = 0.0;
          hyb_rate_var = 0.0;
	  dip_rate = 0.0;
	  poly_rate = 0.0;
        }
      }
      else
      {
        // check to see if the hybrid event did not happen 
	// because of max hybrid being met -- if so,
	// we need to reset this lineage's hybrid rate
	// to 0 so it does not keep scheduling a hybrid
	if (nt->num_hybrids == max_num_hybrids)
	{
	  nt->NodeArray[ev->lin_id]->hybrid_rate = 0.0;
	}
	
        // no matter what the reason for not executing,
	// we need to deal with getting a new event, and
	// marking this event as unexecuted
	ev->sched_status = 4;
        UnexecutedHybridEvent(nt, ev->lin_id, current_time);
      }

    }; // end of hybrid event

    // DEBUGGING birth/death process stuff -- growth analysis
     //printf("NETGEN DEBUG: Just finished loop %d and population (active lineages) is %ld at time %f\n", loop_count, nt->num_act_lins, current_time);
    // END DEBUGGING

    /************************/
    /*      log the event   */
    /************************/
    LogOldEvent(&(nt->old_event_list), ev);

    loop_count++;

  } // end of while -- still have events and have not reached desired size


  // DEBUGGING
  //printf("NETGEN DEBUG:  After big big while\n");
  //printf("NETGEN DEBUG:  ");
  //PrintFutureEvents(stdout, nt->event_queue);
  //printf("NETGEN DEBUG:  ");
  //PrintOldEvents(stdout, nt->old_event_list);
  // END DEBUGGING

  /***************************************************************/
  /*          choose the ending time of the simulation           */
  /*                                                             */
  /*  (unless we ended prematurely, then there is no way to      */
  /*   to get a random final time as all lineages have died off, */
  /*   and our end time will be the time of that last death)     */
  /***************************************************************/

  if (nt->event_queue != NULL)
    current_time = CalcEndTime(nt, current_time);


  /*************************************************/
  /*      clean up all the active lineages         */
  /*************************************************/

  // take all the active lineages and turn them 
  // into leaves now that we are done processing 
  // (except for outgroup leaf handled below)

  MakeExtantLeaves(nt, current_time);
  

  /**************************************/
  /*       outgroup leaf work           */
  /**************************************/

  // if we have an outgroup -- go ahead and make its leaf now and add it 
  // in appropriately to the NodeArray so its information will get printed
  //
  // note:  we can only do this if we have extant taxa to work with, and if
  //        the simulation ended prematurely (due to all lineages dying out) 
  //        we will not have any extant taxa with which to work

  if ( (outgroup) && (nt->num_extant_taxa > 0) )
  {
    MakeOutgroupLeaf(nt, current_time);
  }


  /********************************/
  /*       print the network      */
  /********************************/
  nt->clock_height = ClockHeight(nt->root);
  nt->ev_height = EvHeight(nt->root);
  if (sequences_yn)
  {
    UpdateAllHDsFromRoot(nt);
    nt->avg_hd_height = AvgHDHeight(nt);
    nt->parsimony_score = CalcParsScore(nt);
  }
  NetworkReport(nt);

}

/******************************************************************************/
/*                            CalcEndTime                                     */
/******************************************************************************/
double CalcEndTime(NETWORK nt, double current_time)
{
  double time_to_event, rand_time_to_end;
  EVENT next_ev;

  // if we end now immediately, some branches will have 0 branch length
  // we want to end at a time randomly between now and the next event 
  // (which we do not process as we have reached our extant taxa goal)

  // we know what the time of our next event is by looking at the queue
  next_ev = GetNextEvent(&(nt->event_queue));

  time_to_event = next_ev->event_time - current_time;

  // now choose a time randomly from [now, next event time)
  rand_time_to_end = (genrand_real2() * time_to_event); 
 
  // DEBUGGING
  //printf("NETGEN DEBUG:  calc end time -- current time is: %f, time to next event is: %f, rand choice of time to end is: %f, end time of sim will be: %f\n", current_time, time_to_event, rand_time_to_end, current_time + rand_time_to_end); 
  // END
  
  // hand back what will be the simulation end time
  return (current_time + rand_time_to_end);
}


