/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include "globals.h"
#include "node.h"
#include "event.h"
#include "network.h"
#include "sequences.h"
#include "output.h"
#include "utilities.h"
#include "twister.h"
#include "hybrids.h"


/******************************************************************************/
/*                               Hybrid                                       */
/******************************************************************************/
void Hybrid(NETWORK nt, EVENT ev, double current_time, int hybrid_type) 
{
  long chosen_lin, p2_lin, hybrid_index, node_i;
  int prev_update_yn;
  LINEAGE chosen_ptr;
  LINEAGE p2_ptr;

  int num_hybrid_hom_chromos;
  int num_hybrid_strands_per_hc;
  int hybrid_seq_length;

  /*******************************************************************/
  /*  we know from the event which lineage will be the first parent  */
  /*******************************************************************/
  chosen_lin = ev->lin_id;
  chosen_ptr = nt->NodeArray[chosen_lin];
  
  /*******************************************/
  /* Update branch lengths before proceeding */
  /*******************************************/
  UpdateAllBranchLengths(nt, current_time); 

  /******************************************/
  /* update all sequences for active nodes  */
  /******************************************/
  UpdateAllSequences(nt, current_time);


  /**************************************************************************/
  /*    find a second parent if possible -- choose based upon               */
  /*            hyb_parent_selection = 0 (default min hamming)              */
  /*                                 = 1 (randomly)                         */
  /*                                 = 2 (lineage with "min" branch length) */
  /*                                 = 3 (lineage with "max" branch length) */
  /*                                 = 4 (lineage with "min" ev distance    */
  /*                                 = 5 (lineage chosen with hamming dist  */
  /*                                      from exponential function)        */
  /**************************************************************************/

  p2_lin = -1;

  if (hyb_parent_selection == 0)
  {
    p2_lin = SelectSecondParentMinHamDist(nt, chosen_lin, hybrid_type);
  }
  else if (hyb_parent_selection == 1) 
  {
    p2_lin = SelectSecondParentRandom(nt, chosen_lin, hybrid_type);
  }
  else if (hyb_parent_selection == 2) 
  {
    p2_lin = SelectSecondParentMinBL(nt, chosen_lin, hybrid_type);
  }
  else if (hyb_parent_selection == 3) 
  {
    p2_lin = SelectSecondParentMaxBL(nt, chosen_lin, hybrid_type);
  }
  else if (hyb_parent_selection == 4) 
  {
    p2_lin = SelectSecondParentMinEvDist(nt, chosen_lin, hybrid_type);
  }
  else if (hyb_parent_selection == 5) 
  {
    p2_lin = SelectSecondParentExpFn(nt, chosen_lin, hybrid_type);
  }
  else 
  {
    printf("NETGEN ERROR: Invalid parent selection method %d\n", hyb_parent_selection);
    exit(0);
  }

  /************************************/
  /* if hybrid possible, do the work! */
  /************************************/
  if (p2_lin != -1)
  {
    p2_ptr = nt->NodeArray[p2_lin];

    // do network book-keeping (topology maintenance work)

    num_hybrid_hom_chromos = CalcNumHybridHomChromos(chosen_ptr, p2_ptr, hybrid_type); 

    num_hybrid_strands_per_hc = CalcNumHybridStrands(chosen_ptr, p2_ptr, hybrid_type);

    // all our seq lengths should be the same since these do not vary
    // during the simulation, but do a min function just in case
    hybrid_seq_length = MinInt(chosen_ptr->seq_length, p2_ptr->seq_length);

    hybrid_index = HybridCreation(nt, chosen_lin, p2_lin, hybrid_type,
                                  current_time, 
		                  num_hybrid_hom_chromos,
			          num_hybrid_strands_per_hc,
			          hybrid_seq_length);

    // now update the event so we have the information about the second
    // parent as part of the event
    ev->second_lin_id = p2_ptr->id;
    ev->second_parent_time = p2_ptr->parent1->time_created;

    // DEBUGGING
    //printf("NETGEN DEBUG:  we have just done a hybrid (id %ld) of type %d with parents %ld and %ld (%ld and %ld according to hybrid itself) and the current time is %f\n", nt->NodeArray[hybrid_index]->id, hybrid_type,  nt->NodeArray[chosen_lin]->id, nt->NodeArray[p2_lin]->id, nt->NodeArray[hybrid_index]->parent1->id, nt->NodeArray[hybrid_index]->parent2->id, current_time); 
    // END DEBUGGING

    // do sequence work to determine the sequences for the new hybrid
    if (hybrid_type == 1)
    {
      MakeDipHybridSequences(nt->NodeArray[hybrid_index], seq_gen_options);
    }
    else if (hybrid_type == 2)
    {
      MakePolyHybridSequences(nt->NodeArray[hybrid_index], seq_gen_options);
    }
    else
    {
      printf("NETGEN ERROR:  hybrid type not defined\n");
      exit(0);
    }

    // clean up network book-keeping (topology and event maintenance work)
    HybridCleanUp(nt, chosen_lin, p2_lin);

    // mark this event as processed
    ev->sched_status = 2;
  }
  else
  {
    // need to indicate this event could not be executed
    ev->sched_status = 4;
    UnexecutedHybridEvent(nt, chosen_lin, current_time);
  }

  // DEBUGGING
  // printf("NETGEN DEBUG:  just finished hybrid event type %d with parent 1 being %ld, parent 2 being %ld, with the resulting hybrid lineage being %ld\n", hybrid_type, chosen_lin, p2_lin, hybrid_index);
  // END DEBUGGING

}

/******************************************************************************/
/*               Select Second Parent -- Min Hamming DISTANCE                 */
/******************************************************************************/
long SelectSecondParentMinHamDist(NETWORK nt, long chosen_lin, int hybrid_type)
{
  // NOTE:  the min hamming distance here is calculated as an average
  
  long node_i;
  long h_dist;
  long min_h_dist;
  long thresh_val;
  long closest_lin;

  long num_candidates;
  long selected_index;

  int hamming_distance[MAX_NODES];
  int hchromo_i, strand_i, nhc, nsphc;
  char str_a[MAX_SEQ_LENGTH];
  char str_b[MAX_SEQ_LENGTH];
  LINEAGE chosen_ptr;
  LINEAGE nd;

  // DEBUGGING
  //HDDebugging(nt);

  chosen_ptr = nt->NodeArray[chosen_lin];
  closest_lin = -1;

  /*******************************************************************/
  /* create distance array for chosen lin and every other active lin */
  /*******************************************************************/
  for (node_i = 0; node_i < (nt->num_nodes); node_i++) 
  {
    nd = nt->NodeArray[node_i];

    // reset to the max for each new node no matter what
    h_dist = INT_MAX;

    if (CheckNodeEligibility(chosen_ptr, nd, hybrid_type))
    {
      h_dist = HammingDistanceAvgNodes(chosen_ptr, nd);
    }

    // set the value in the array
    hamming_distance[node_i] = h_dist;  

    // DEBUGGING
    //printf("NETGEN DEBUG:  Average Hamming Distance between chosen lineage %ld and %ld is %d\n", chosen_ptr->id, nd->id, hamming_distance[node_i]);
  }

  /**********************************************************/
  /* find the best second parent based upon a MIN selection */
  /**********************************************************/

  // set the default as high as it can be so we can look for a better value
  min_h_dist = INT_MAX;
  thresh_val = (long) (hyb_ham_thresh_rate * min_h_dist);

  // DEBUGGING
  //printf("NETGEN DEBUG:  our thresh_val for seq hybrids is %ld\n", thresh_val);

  for (node_i = 0; node_i < (nt->num_nodes); node_i++) 
  {
    if ( (hamming_distance[node_i] <=  min_h_dist) &&
         (hamming_distance[node_i] <= thresh_val) )
    {
      min_h_dist = hamming_distance[node_i];
      closest_lin = node_i;
    }
  }

  // since it is possible that more than one lineage has the "min" hamming
  // distance we want to randomly select from those...so go through this
  // extra processing (but only if there was at least one candidate found
  // above)

  num_candidates = 0;

  if (closest_lin != -1)
  {
    InitializeCandidatePool(nt);
  
    // identify the candidates
    for (node_i = 0; node_i < nt->num_nodes; node_i++)
    {
      if ( (nt->NodeArray[node_i]->active == 1) &&
           (node_i != chosen_lin) &&
	   (hamming_distance[node_i] == min_h_dist))
      {
        nt->Candidates[num_candidates] = node_i;
	num_candidates++;
      }
    }
  }

  if (num_candidates == 0)
  {
    closest_lin = -1;
  }
  else
  {
    // randomly pick one of the candidates
    selected_index = (long) (trunc((genrand_real2() * num_candidates)));
    closest_lin = nt->Candidates[selected_index];
  }
  
  return closest_lin;
}

/******************************************************************************/
/*                    Select Second Parent -- Random                         */
/******************************************************************************/
long SelectSecondParentRandom(NETWORK nt, long chosen_lin, int hybrid_type)
{
  long p2_lin;
  long node_i;
  long num_candidates;
  long selected_index;

  // first initialize the candidate pool and counter
  InitializeCandidatePool(nt);
  num_candidates = 0;

  // now go through the node array and identify all the potential candidates
  // these are all lineages which are active and not the chosen_lin

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    if (CheckNodeEligibility(nt->NodeArray[chosen_lin], nt->NodeArray[node_i], hybrid_type))
    {
      nt->Candidates[num_candidates] = node_i;
      num_candidates++;
    }
  }

  if (num_candidates == 0)
  {
    p2_lin = -1;
  }
  else
  {
    // randomly pick one of the candidates
    selected_index = (long) (trunc((genrand_real2() * num_candidates)));
    p2_lin = nt->Candidates[selected_index];
  }

  return p2_lin;
}

/******************************************************************************/
/*                   Select Second Parent -- Min Branch Length                */
/******************************************************************************/
long SelectSecondParentMinBL(NETWORK nt, long chosen_lin, int hybrid_type)
{
  /************************************************************/
  /*  This routine forces the second parent for the hybrid    */
  /*  to be a lineage that would have otherwise had a minimal */
  /*  branch length.                                          */
  /************************************************************/

  EVENT ev_i_ptr;

  long node_i;
  long num_candidates;
  long selected_index;

  long p2_lin;
  double min_bl;

  p2_lin = -1; // default
  min_bl = DBL_MAX;

  // go through queue and calculate what the branch length for the
  // event will be based upon event time - created time
  //
  // if this expected_bl beats (is less than) min_bl currently found
  // replace the p2_lin with this event's lin_id

  ev_i_ptr = nt->event_queue;
  while (ev_i_ptr != NULL)
  {
    if ((ev_i_ptr->event_time - ev_i_ptr->time_created) < min_bl)
    {
      if (CheckNodeEligibility(nt->NodeArray[chosen_lin], nt->NodeArray[ev_i_ptr->lin_id], hybrid_type))
      {
        min_bl = ev_i_ptr->event_time - ev_i_ptr->time_created;
        p2_lin = ev_i_ptr->lin_id;
      }
    }
    ev_i_ptr = ev_i_ptr->next_event;
  }

  // it is possible that there was a tie for the min branch length, so
  // if at least one candidate was found above, go through the process 
  // of identifying all candidates and picking one randomly

  num_candidates = 0;

  if (p2_lin != -1)
  {
    InitializeCandidatePool(nt);

    // go through events again looking for candidates
    ev_i_ptr = nt->event_queue;
    while (ev_i_ptr != NULL)
    {
      if ((ev_i_ptr->event_time - ev_i_ptr->time_created) == min_bl)
      {
        nt->Candidates[num_candidates] = ev_i_ptr->lin_id;
	num_candidates++;
      }
    }

    if (num_candidates == 0)
      p2_lin = -1; // redundant
    else
    {
      // randomly pick one of the candidates
      selected_index = (long) (trunc((genrand_real2() * num_candidates)));
      p2_lin = nt->Candidates[selected_index];
    }
  }

  // DEBUGGING
  // printf("NETGEN DEBUG: we are choosing %ld as our second parent, it would have otherwise resulted in a branch length of %f (for comparison the head of the queue has an expected bl of %f)\n", p2_lin, min_bl,( nt->event_queue->event_time - nt->event_queue->time_created));

  return p2_lin;
}

/******************************************************************************/
/*                   Select Second Parent -- Max Branch Length                */
/******************************************************************************/
long SelectSecondParentMaxBL(NETWORK nt, long chosen_lin, int hybrid_type)
{
  /************************************************************/
  /*  This routine forces the second parent for the hybrid    */
  /*  to be a lineage that would have otherwise had a MAXIMAL */
  /*  branch length.                                          */
  /************************************************************/

  EVENT ev_i_ptr;

  long node_i;
  long num_candidates;
  long selected_index;

  long p2_lin;
  double max_bl;

  p2_lin = -1; // default
  max_bl = -1;

  // go through queue and calculate what the branch length for the
  // event will be based upon event time - created time
  //
  // if this expected_bl beats (is MORE than) max_bl currently found
  // replace the p2_lin with this event's lin_id

  ev_i_ptr = nt->event_queue;
  while (ev_i_ptr != NULL)
  {
    if ((ev_i_ptr->event_time - ev_i_ptr->time_created) > max_bl)
    {
      if (CheckNodeEligibility(nt->NodeArray[chosen_lin], nt->NodeArray[ev_i_ptr->lin_id], hybrid_type))
      {
        max_bl = ev_i_ptr->event_time - ev_i_ptr->time_created;
        p2_lin = ev_i_ptr->lin_id;
      }
    }
    ev_i_ptr = ev_i_ptr->next_event;
  }

  // it is possible that there was a tie for the max branch length, so
  // if at least one candidate was found above, go through the process 
  // of identifying all candidates and picking one randomly

  num_candidates = 0;

  if (p2_lin != -1)
  {
    InitializeCandidatePool(nt);

    // go through events again looking for candidates
    ev_i_ptr = nt->event_queue;
    while (ev_i_ptr != NULL)
    {
      if ((ev_i_ptr->event_time - ev_i_ptr->time_created) == max_bl)
      {
        nt->Candidates[num_candidates] = ev_i_ptr->lin_id;
	num_candidates++;
      }
    }

    if (num_candidates == 0)
      p2_lin = -1; // redundant
    else
    {
      // randomly pick one of the candidates
      selected_index = (long) (trunc((genrand_real2() * num_candidates)));
      p2_lin = nt->Candidates[selected_index];
    }
  }

  // DEBUGGING
  // printf("NETGEN DEBUG: we are choosing %ld as our second parent, it would have otherwise resulted in a branch length of %f (as a comparison the head of the queue has an expected branch length of %f)\n", p2_lin, max_bl, (nt->event_queue->event_time - nt->event_queue->time_created));

  return p2_lin;
}

/******************************************************************************/
/*            Select Second Parent -- Min Evolutionary DISTANCE               */
/******************************************************************************/
long SelectSecondParentMinEvDist(NETWORK nt, long chosen_lin, int hybrid_type)
{
  long node_i;
  long closest_lin;
  double min_ev_dist;
  LINEAGE chosen_ptr;

  long num_candidates;
  long selected_index;

  chosen_ptr = nt->NodeArray[chosen_lin];
  closest_lin = -1;
  min_ev_dist = DBL_MAX;

  /*******************************************************************/
  /* find the lineage/node closest to the chosen_lin in terms of     */
  /* evolutionary distance                                           */
  /*                                                                 */
  /* to do this, we run dijkstra's algorithm, with the chosen_lin as */
  /* the "source" -- once run, each lineage's dijkstra_dist will     */
  /* contain the distance for reaching that node from the source and */
  /* we can then go through all lineages/nodes to find the min       */
  /*                                                                 */
  /* once the minimum is found, we can check it against the          */
  /* threshold and see if it will be the second lineage.....         */
  /*******************************************************************/

  DijkstraDist(nt, chosen_lin);

  for (node_i = 0; node_i < (nt->num_nodes); node_i++) 
  {
    if (CheckNodeEligibility(chosen_ptr, nt->NodeArray[node_i], hybrid_type))
    {
      if (nt->NodeArray[node_i]->dijkstra_dist <= min_ev_dist)
      {
        min_ev_dist = nt->NodeArray[node_i]->dijkstra_dist;
	closest_lin = node_i;
      }
    }
  }

  // make sure the threshold is met, otherwise, return -1
  if (nt->NodeArray[closest_lin]->dijkstra_dist > hyb_ev_dist_thresh)
  {
    closest_lin = -1;
  }

  // it is possible that there was a tie for the closest lineage so
  // if there is at least one candidate found at this point, go through
  // and identify all candidates and pick one randomly

  num_candidates = 0;

  if (closest_lin != -1)
  {
    InitializeCandidatePool(nt);

    // go through and identify candidates
    for (node_i = 0; node_i < (nt->num_nodes); node_i++)
    {
      if ( (nt->NodeArray[node_i]->dijkstra_dist == min_ev_dist) &&
           (nt->NodeArray[node_i]->active == 1) &&
	   (chosen_lin != node_i))
      {
        nt->Candidates[num_candidates] = node_i;
        num_candidates++;
      }
    }

    if (num_candidates == 0)
      closest_lin = -1; // redundant
    else
    {
      // randomly pick one of the candidates
      selected_index = (long) (trunc((genrand_real2() * num_candidates)));
      closest_lin = nt->Candidates[selected_index];
    }
  }

  return closest_lin;
}

/******************************************************************************/
/*                Select Second Parent -- Exponential Function                */
/******************************************************************************/
long SelectSecondParentExpFn(NETWORK nt, long chosen_lin, int hybrid_type)
{
  /********************************************************************/
  /*   find a lineage that meets a hamming distance drawn from the    */
  /*   exponential distribution defined by the user via lambda_hd     */
  /*                                                                  */
  /*   lambda_hd is the hamming distance which has 1/e (app 37%)      */
  /*   probability of being the second parent's hamming distance      */
  /*   from the first parent -- this defines an exponential           */
  /*   distribution                                                   */
  /*                                                                  */
  /*   once this distribution is defined, we can pick a value         */
  /*   from it, and then look for active lineages with this           */
  /*   distance (hd);  if candidates are found, choose one randomly;  */
  /*   if no candidates are found, expand the search by increasing/   */
  /*   decreasing the hd value by 1 until candidates are found or     */
  /*   the boundaries defined by the hyb_exp_fn_thresh are reached    */
  /*                                                                  */
  /*   note that we are defining hamming distance as an average here  */
  /********************************************************************/

  long max_hd;
  long hd_thresh;
  long ideal_hd;
  long calculated_hd;
  long p2_lin;
  long node_i;
  long num_candidates;
  long selected_index;
  long hd_range;
  long adj_hd_low;  // lower bound of adjusted hamming distance value
  long adj_hd_high; // upper bound of adjusted hamming distance value

  int i;
  
  ideal_hd = 0;

  max_hd = MaxHammingDistanceAvgNodes(nt->NodeArray[chosen_lin], nt->NodeArray[chosen_lin]); 

  hd_thresh = (long) (trunc(hyb_exp_fn_thresh_rate * max_hd));

  // get our ideal hamming distance for the second parent from the 
  // exponential distribution which was defined by the user inputting
  // the lambda_hd value -- the value should be in the range of 0,max_hd
  ideal_hd = CustomizedExpHD(max_hd);

  // using an array of node indexes to track the pool of candidates
  // (while it would be more elegant to sort all nodes according to
  // their hamming distance from the first parent, this is pontentially
  // very expensive -- this will save processing time in most cases
  // as in most circumstances a pool will be created very early on)

  // initialize candidate pool array and counter
  InitializeCandidatePool(nt);
  num_candidates = 0;

  // while there are no candidates, expand the range of hd and look
  // for candidates (within the boundaries)
  
  for (hd_range = 0; hd_range <= max_hd; hd_range++)
  {
    // DEBUGGING
    //printf("NETGEN DEBUG:  hd_range is %ld\n", hd_range);

    // while we are lacking candidates for the second parent keep searching
    if (num_candidates == 0)
    {
      // first set the range for this round -- remember we do not want to 
      // go beyond the boundaries defined by the hyb_exp_fn_thresh value
      adj_hd_low  = MaxLong((ideal_hd - hd_thresh), (ideal_hd - hd_range));
      adj_hd_high = MinLong((ideal_hd + hd_thresh), (ideal_hd + hd_range));

      for (node_i = 0; node_i < nt->num_nodes; node_i++)
      {
        if (CheckNodeEligibility(nt->NodeArray[chosen_lin], nt->NodeArray[node_i], hybrid_type))
        {
	  calculated_hd = HammingDistanceAvgNodes(nt->NodeArray[chosen_lin], 
			                          nt->NodeArray[node_i]);

	  // see if we have a candidate and if so, record it
	  if ( (calculated_hd >= adj_hd_low) && (calculated_hd <= adj_hd_high) )
	  {
	    nt->Candidates[num_candidates] = node_i;
	    num_candidates++;

	    // DEBUGGING
	    //printf("NETGEN DEBUG:  we have a candidate id is %ld\n", node_i);
	  }
        }
      } // end of going through node array
    } // if num_candidates == 0
    
    // DEBUGGING
    //printf("NETGEN DEBUG:  our num of candidates at this point is %d\n", num_candidates);
  } // next range step

  if (num_candidates == 0)
  {
    // we did not find any candidates at all....
    p2_lin = -1;
  }
  else
  {
    // randomly pick one of the candidates
    selected_index = (long) (trunc((genrand_real2() * num_candidates)));
    p2_lin = nt->Candidates[selected_index];
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: our selected candidate is %ld (it had an index of %ld)\n", p2_lin, selected_index);

  return p2_lin;

}


/******************************************************************************/

/******************************************************************************/
/*                        Make Diploid Hybrid Sequences                       */
/******************************************************************************/
void MakeDipHybridSequences(NODE nd, char seq_gen_options[]) 
{

  int p1_num_strands_gives, p2_num_strands_gives;
  int nhc, nsphc, hchromo_i, strand_i;

  nhc = nd->num_hom_chromos;
  nsphc = nd->num_strands_per_hom_chrom;

  // knowing we have an even number of strands, we can establish
  // how many strands from each homologous chromosome will come from each parent;
  // note that for diploid hybridizations this is always a 50/50 split
  // per chromosome contents

  p1_num_strands_gives = (int) (nsphc * 0.5);
  p2_num_strands_gives = p1_num_strands_gives;

  // just another double check that will we be okay to process sequences
  if ((p1_num_strands_gives + p2_num_strands_gives) != nsphc)
  {
    printf("NETGEN ERROR: We have a problem with doing a diploid hybridization\n" );
    printf("NETGEN ERROR: number of strands do not add up for the hybrid id %ld\n", nd->id);
    exit(0);
  }

  /****************************************************************/
  /*                                                              */
  /*  For EACH homologous chromosome, half of the sequences come  */
  /*  from each of the two parents -- resulting in the same total */
  /*  number of  strand sequences for the hybrid in the end.      */
  /*  However, the selection of strand sequences for each         */
  /*  homologous chromosome composition is random.                */
  /*                                                              */
  /****************************************************************/

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    // do the first half of sequences which will come from p1
    RandomFillStrands(nd, nd->parent1, hchromo_i, 0, p1_num_strands_gives);

    // do the second half of sequences which will come from p2
    RandomFillStrands(nd, nd->parent2, hchromo_i, p1_num_strands_gives, nsphc);
  } 

  // DEBUGGING
  //printf("NETGEN DEBUG: done assigning seqs for hybrid %ld\n", nd->id);

  //for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nd->num_strands_per_hom_chrom; strand_i++)
    //{
      //printf("NETGEN DEBUG: hchromo_i %d, strand_i %d is %s \n", hchromo_i, strand_i, nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
  // END DEBUGGING
}


/******************************************************************************/
/*                      Random Fill Strands                                   */
/******************************************************************************/
void RandomFillStrands(NODE nd, NODE parent_ptr, int hchromo_i, 
		       int start_strand, int end_strand)
{
  int strand_i;
  int rand_index, used_index_yn;
  int used_indices[MAX_NUM_STRANDS_PER_HC];


  /**************************************************************/
  /*  Here we select the correct number of sequences from the   */
  /*  specified parent.  We do this as random selection         */
  /*  without replacement and track which sequences are chosen, */
  /*  as a sequence can be chosen only once.  This keeps the    */
  /*  random selection fair as we do not adjust the size        */
  /*  of the pool, simply rechoose if a duplicate is chosen.    */
  /**************************************************************/

  // initialize the tracking array
  for (strand_i = 0; strand_i < nd->num_strands_per_hom_chrom; strand_i++)
  {
    used_indices[strand_i] = -1;
  }
  
  // do the contribution
  for (strand_i = start_strand; strand_i < end_strand; strand_i++)
  {
    // set default so we enter the loop the first time
    used_index_yn = 1;

    // look for a contribution that is not a repeat of a previous contribution
    while (used_index_yn)
    {
      // the following gives a rand in the [0,1) range multiplied by number of
      // strands and finally cast to an int -- this in effect gives 
      // us a rand int in the range of [0,num_strands_per_hom_chrom-1] 
      // which is how we want to select
      rand_index = (int) (genrand_real2()*(parent_ptr->num_strands_per_hom_chrom));

      // DEBUGGING
      //printf("NETGEN DEBUG: parent (id %ld) rand_index chosen %d (for strand_i %d)\n", parent_ptr->id, rand_index, strand_i);
      // END DEBUGGING

      // now see if rand_index has been chosen before
      if (used_indices[rand_index] == 1)
        used_index_yn = 1;
      else
        used_index_yn = 0;

    } // find a random index that is not a repeat

    // record the rand_index chosen this round
    used_indices[rand_index] = 1;

    // copy the chosen sequence into the node 
    strcpy(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
           parent_ptr->hom_chromos[hchromo_i]->strands[rand_index]->sequence);

    // record which parent gave the sequence
    nd->hom_chromos[hchromo_i]->strands[strand_i]->ancestor_node_id = 
      parent_ptr->id;

  } // next strand
}

/******************************************************************************/
/*                        Make Polyloid Hybrid Sequences                      */
/******************************************************************************/
void MakePolyHybridSequences(NODE nd, char seq_gen_options[]) 
{
  // the number of unique homologous chromosomes is the same for both parents
  // and the new hybrid (and its offspring) -- so set that value here 
  // (number of total homologous chromosomes and type has already been set 
  // correctly in HybridCreation)
  
  nd->num_unique_hom_chromos = nd->parent1->num_unique_hom_chromos;
  nd->left->num_unique_hom_chromos = nd->parent1->num_unique_hom_chromos;

  // for polyploid hybridization, the hybrid gets all the chromosomes
  // from both parents

  // for parent 1
  InheritAllStrands(nd, nd->parent1, 0);

  // for parent 2
  InheritAllStrands(nd, nd->parent2, nd->parent1->num_hom_chromos);

  // DEBUGGING
  //PolyploidDebugging(nd);

  return;
}

/******************************************************************************/
/*                          InheritAllStrands                                 */
/******************************************************************************/
void InheritAllStrands(NODE nd, NODE parent_ptr, int start_hchromo_id)
{
  int hchromo_i;
  int hchromo_j;

  int strand_i;
  int strand_j;

  int value;

  // DEBUGGING
  //printf("NETGEN:  starting inherit all strands\n");

  // go through parent chromosomes and fill correct hybrid chromosomes
  for (hchromo_i = 0; hchromo_i < parent_ptr->num_hom_chromos; hchromo_i++)
  {
     hchromo_j = hchromo_i + start_hchromo_id;

     // set the number for the hybrid's homologous chromosome
     // and also the values for the hybrid's offspring

     nd->hom_chromos[hchromo_j]->number = 
       parent_ptr->hom_chromos[hchromo_i]->number;
     nd->left->hom_chromos[hchromo_j]->number = 
       parent_ptr->hom_chromos[hchromo_i]->number;

     // DEBUGGING
     //printf("NETGEN DEBUGGING: PARENTAL number was.....%d\n", parent_ptr->hom_chromos[hchromo_i]->number);

     // set the values for the strands of this homologous chromosome
     for (strand_j = 0; 
          strand_j < parent_ptr->num_strands_per_hom_chrom; 
          strand_j++)
     {
       // set the number, ancestor id, and sequence for the strand
       // do this for both the hybrid and its offspring

       nd->hom_chromos[hchromo_j]->strands[strand_j]->number = 
         parent_ptr->hom_chromos[hchromo_i]->strands[strand_j]->number; 
       nd->left->hom_chromos[hchromo_j]->strands[strand_j]->number = 
         parent_ptr->hom_chromos[hchromo_i]->strands[strand_j]->number; 

       nd->hom_chromos[hchromo_j]->strands[strand_j]->ancestor_node_id = 
         parent_ptr->id;
       nd->left->hom_chromos[hchromo_j]->strands[strand_j]->ancestor_node_id =
         parent_ptr->id;

       strcpy(nd->hom_chromos[hchromo_j]->strands[strand_j]->sequence,
              parent_ptr->hom_chromos[hchromo_i]->strands[strand_j]->sequence);
     }
  }

  return;
}

/******************************************************************************/
/*                              HybridCreation                                */
/******************************************************************************/
long HybridCreation(NETWORK nt, long chosen_lin, long p2_lin, 
		    int hybrid_type, double current_time, int h_nhc, 
		    int h_nsphc, int h_sl)
{
  LINEAGE chosen_ptr;
  LINEAGE p2_ptr; 
  NODE hybrid;
  NODE leftmost;
  long hybrid_index, chosen_conts_ind, hybrid_conts_ind, p2_conts_ind;
  int nhc, nsphc, sl;

  /***************************************************************************/
  /*                    CONCEPTUALLY WE HAVE THE FOLLOWING:                  */
  /*                                                                         */
  /*                            <root somewhere>                             */
  /*                           .               .                             */
  /*                           .               .                             */
  /*                          /                /                             */
  /*  brlen1_ev/1[chosen_lin]/                / brlen1_ev/t[p2_lin]          */
  /*      (retired)         /                /          (retired)            */
  /*                       /     hybrid     /                                */
  /*          chosen_lin(*)-----(*)-------(*)p2_lin                          */
  /*                     /        |          \                               */
  /*                    /         |           \                              */
  /*                   /          |            \                             */
  /*                 (*)         (*)           (*)                           */
  /*   (chosen_conts_ind)  (hybrid_conts_ind)  (p2_conts_ind)                */
  /*      (continues)         (continues)       (continues)                  */
  /*                                                                         */
  /*                                                                         */
  /*                                                                         */
  /***************************************************************************/

  chosen_ptr =  nt->NodeArray[chosen_lin];
  p2_ptr = nt->NodeArray[p2_lin];

  nt->NodeArray[chosen_lin]->time_created = current_time;
  nt->NodeArray[p2_lin]->time_created = current_time;


  /******************************************************************/
  /* make the new node for the chosen node lineage to carry on      */
  /*        -- its original lineage will be "retired" later         */
  /******************************************************************/
  nhc = chosen_ptr->num_hom_chromos;
  nsphc = chosen_ptr->num_strands_per_hom_chrom;
  sl = chosen_ptr->seq_length;

  chosen_conts_ind = (nt->node_id_ctr)++;
  nt->NodeArray[chosen_conts_ind] = NewLineage(chosen_conts_ind, nhc, nsphc, sl, chosen_ptr);


  /********************************************************************/
  /*        make the new lineages for the hybrid offspring            */
  /*  (the initial hybrid merging and then the continuing lineage)    */
  /*                                                                  */
  /* the initial hybrid never is active, while the continuing one is  */
  /********************************************************************/
  hybrid_index = (nt->node_id_ctr)++;

  hybrid = NewLineage(hybrid_index, h_nhc, h_nsphc, h_sl, NULL);
  nt->NodeArray[hybrid_index] = hybrid;
  hybrid->time_created = current_time;
  hybrid->hybrid = hybrid_type;
  hybrid->active = 0; // this will not be an active lineage

  // DEBUGGING
  //printf("NETGEN DEBUG: we just assigned the hybrid id of %ld\n", hybrid->id);
  // END DEBUGGING

  // there is no clock or evolution time here as a hybrid is defined
  // to be at the same point in time with no other evolution, just
  // the mixing of the parents' info
  hybrid->brlen1_ev = 0.0; 
  hybrid->brlen2_ev = 0.0;
  hybrid->brlen1_t =  0.0;
  hybrid->brlen2_t =  0.0;

  // this is continuing lineage from the initial hybrid
  hybrid_conts_ind = (nt->node_id_ctr)++;

  nt->NodeArray[hybrid_conts_ind] = NewLineage(hybrid_conts_ind, h_nhc, h_nsphc, h_sl, NULL);

  // DEBUGGING
  //printf("NETGEN DEBUG: we just assigned the continuing hybrid lineage id of %ld\n", nt->NodeArray[hybrid_conts_ind]->id);
  // END DEBUGGING
      
  /************************************************************************/
  /* make the new node for the second parent lingeage to continue on      */
  /*            its original lineage will be "retired" later              */
  /************************************************************************/
  nhc = p2_ptr->num_hom_chromos;
  nsphc = p2_ptr->num_strands_per_hom_chrom;
  sl = p2_ptr->seq_length;

  p2_conts_ind = (nt->node_id_ctr)++;
  nt->NodeArray[p2_conts_ind] = NewLineage(p2_conts_ind, nhc, nsphc, sl, p2_ptr);


  /******************************************************************/
  /*      now get the lineages that will continue on, hooked        */
  /*               up in the network properly                       */
  /******************************************************************/

  // hook up the lineages such that the leftmost lineage will have
  // the hybrid as its right child and the rightmost lineage will
  // have the hybrid as its left child -- this lets us print out
  // NewickNetwork format correctly and is easiest to just set
  // correctly here from the creation of the hybrid so we don't
  // have to be swapping subtrees later

  leftmost = FindLeftmostNode(nt, chosen_ptr, p2_ptr);

  if (chosen_ptr == leftmost)
  {
    ConnectTwoChildren(chosen_ptr, nt->NodeArray[chosen_conts_ind], hybrid);
    ConnectTwoChildren(p2_ptr, hybrid, nt->NodeArray[p2_conts_ind]);
  }
  else if (p2_ptr == leftmost)
  {
    ConnectTwoChildren(chosen_ptr, hybrid, nt->NodeArray[chosen_conts_ind]);
    ConnectTwoChildren(p2_ptr, nt->NodeArray[p2_conts_ind], hybrid);
  }
  else
  {
    printf("NETGEN ERROR: hybrid parents (%ld and %ld) were not found in the network!!!!\n", chosen_ptr->id, p2_ptr->id);
    exit(0);
  }

  // now hook up the continuing lineage from the hybrid which is
  // always the left child of the hybrid which is also what we
  // want for our NewickNetwork printing
  ConnectOnlyChild(hybrid, nt->NodeArray[hybrid_conts_ind]);

  /******************************************************************/
  /*   deal with max number of hybrids local level -- if reached    */
  /******************************************************************/
  // if the max num of hybrids will be reached with this
  // hybrid creation routine, we need to reset the individual 
  // lineage hybrid rates to be 0 before their events are 
  // established -- global hybrid rates are reset back in
  // generate routine once this whole event is complete
  //
  // note:  the second parent can possibly inherit a hybrid 
  // event to do when hybrids are no longer allowed, but that 
  // will turn into an UnexecutedHybridEvent at the event 
  // time and will get a new event scheduled for it (and in order
  // for it not to ask for another hybrid event at that
  // rescheduling, we must also reset this lineage's hybrid 
  // rate to 0 here)
 
  if ( (nt->num_hybrids + 1) == (max_num_hybrids) )
  {
    nt->NodeArray[chosen_conts_ind]->hybrid_rate = 0.0;
    nt->NodeArray[hybrid_conts_ind]->hybrid_rate = 0.0;
    nt->NodeArray[p2_conts_ind]->hybrid_rate = 0.0;
  }

  /******************************************************************/
  /*      now get the events for the continuing lineages correct    */
  /******************************************************************/
  // the chosen_conts_ind and the hybrid_conts_ind both need new
  // events scheduled for them; however, the p2_conts_ind will
  // "inherit" the event originally scheduled for its parent, p2_lin

  EstablishEvent(&(nt->event_queue), nt->NodeArray[chosen_conts_ind], current_time);
  EstablishEvent(&(nt->event_queue), nt->NodeArray[hybrid_conts_ind], current_time);
  InheritParentalEvent(&(nt->event_queue), p2_lin, p2_conts_ind);

  /********************************/
  /*     more book-keeping        */
  /********************************/
  nt->num_act_lins += 3; // the two retired ones are subtracted later in CleanUp
  nt->num_nodes += 4; 
  nt->num_edges += 5;
  nt->num_hybrids += 1;

  return hybrid_index; // return the hybrid's index into nt->NodeArray
}


/******************************************************************************/
/*                            HybridCleanUp                                   */
/******************************************************************************/
void HybridCleanUp(NETWORK nt, long chosen_lin, long p2_lin)
{

  // set the chosen and second parent lineages as inactive, now that
  // they have done their work
  nt->NodeArray[chosen_lin]->active = 0;
  nt->NodeArray[p2_lin]->active = 0;

  // DEBUGGING
  //printf("NETGEN DEBUG:  we are about to reduce num_act_lins by 2, currently it is %ld\n", nt->num_act_lins);

  nt->num_act_lins -= 2;

  // DEBUGGING
  //printf("NETGEN DEBUG:  we just reduced num_act_lins by 2, now it is %ld\n", nt->num_act_lins);

  // parent 2 has an event waiting to happen, but now that it was
  // part of the hybridization, we need to cancel its event (its child
  // has already "inherited" this event as its own new event as part 
  // of HybridCreation routine)
  CancelLinEvent(&(nt->event_queue), &(nt->old_event_list), p2_lin);
}


/******************************************************************************/
/*                          UnexecutedHybridEvent                             */
/******************************************************************************/
void UnexecutedHybridEvent(NETWORK nt, long chosen_lin, double current_time)
{
  // we are here because the hybrid did not occur for some reason 
  // (i.e. no other active lineages, no acceptable lineage with which
  // to hybridize found, no more hybrids allowed)
  //
  // we need to find out what this lineage's next event will be as 
  // it is still an active lineage (if it's hybrid rate needed to be
  // reset to 0, this was taken care of in HybridCreation or the 
  // Generate routine)

  EstablishEvent(&(nt->event_queue), nt->NodeArray[chosen_lin], current_time);
}

/******************************************************************************/
/*                          CheckNodeEligibility                              */
/******************************************************************************/
int CheckNodeEligibility(NODE nd_1, NODE nd_2, int hybrid_type)
{
  // in this routine we check whether the criteria for the
  // type of hybrid being performed are met and return
  // either true (1) or false (0)

  int elig;


  // assume eligible until proved otherwise
  elig = 1;

  // criteria generic to all types of hybrids

  // both nodes need to be active
  if ((nd_1->active != 1) || (nd_2->active != 1))
    elig = 0;

  // the two nodes cannot be the same node!
  if ((nd_1->id) == (nd_2->id))
    elig = 0;

  // checks specific to type of hybrid

  // if diploid hybrid the number of homologous chromosomes and strands 
  // must match.....and there must be an even number of strands
  if (hybrid_type == 1)
  {
    if (nd_1->num_hom_chromos != nd_2->num_hom_chromos)
    {
      elig = 0;
    }

    if (nd_1->num_strands_per_hom_chrom != 
        nd_2->num_strands_per_hom_chrom)
    {
      elig = 0;
    }

    if ((nd_1->num_strands_per_hom_chrom%2) != 0)
    {
      elig = 0;
    }

  } // end of diploid hybrid checks

  // if polyploid hybrid we need to make sure the number of
  // unique homologous chromosomes matches and that the number of 
  // strands (per chromosome) are the same -- that is all we need
  // (without this the hamming distance will not be defined)
  if (hybrid_type == 2)
  {
    if (nd_1->num_unique_hom_chromos != nd_2->num_unique_hom_chromos)
    {
      elig = 0;
    }

    if (nd_1->num_strands_per_hom_chrom != 
        nd_2->num_strands_per_hom_chrom)
    {
      elig = 0;
    }
  }

  return elig;
}

/******************************************************************************/
/*                          CalcNumHybridHomChromos                           */
/******************************************************************************/
int CalcNumHybridHomChromos(NODE nd_1, NODE nd_2, int hybrid_type)
{
  int ret_value;

  // diploid hybrids end up with same number of homologous chromosomes that 
  // one parent has -- we have already checked the two parents have
  // the same number of homologous chromosomes
  if (hybrid_type == 1)
    ret_value = nd_1->num_hom_chromos;

  // polyploid hybrids get the sum of the parental chromosomes
  if (hybrid_type == 2)
    ret_value = nd_1->num_hom_chromos + nd_2->num_hom_chromos;


  if (hybrid_type > 2)
  {
    printf("NETGEN ERROR:  this type of hybridization (%d) is not defined\n", hybrid_type);
    exit(0);
  }
    
  return ret_value;
}

/******************************************************************************/
/*                          CalcNumHybridStrands                              */
/******************************************************************************/
int CalcNumHybridStrands(NODE nd_1, NODE nd_2, int hybrid_type)
{
  int ret_value;

  // both diploid and polyploid hybridizations keep the same
  // number of strands per homologous chromosome

  if ((hybrid_type == 1) || (hybrid_type == 2))
    ret_value = nd_1->num_strands_per_hom_chrom;

  if (hybrid_type > 2)
  {
    printf("NETGEN ERROR:  this type of hybridization (%d) is not defined\n", hybrid_type);
    exit(0);
  }

  return ret_value;
}

/******************************************************************************/
/*                             HDDebugging                                    */
/******************************************************************************/
void HDDebugging(NETWORK nt)
{
  // This routine is solely for testing the hamming distance routines;
  // it automatically exits NetGen at the end of this routine.

  NODE chosen_ptr;
  NODE nd;
  long h_dist;

  chosen_ptr = NewLineage(0,6,2,4,NULL);
  nd = NewLineage(1,4,2,4,NULL);

  chosen_ptr->num_hom_chromos = 6;
  chosen_ptr->num_unique_hom_chromos = 2;
  chosen_ptr->num_strands_per_hom_chrom = 2;

  chosen_ptr->hom_chromos[0]->id = 0;
  chosen_ptr->hom_chromos[0]->number = 0;
  chosen_ptr->hom_chromos[0]->strands[0]->id = 0;
  chosen_ptr->hom_chromos[0]->strands[0]->number = 0;
  chosen_ptr->hom_chromos[0]->strands[1]->id = 1;
  chosen_ptr->hom_chromos[0]->strands[1]->number = 1;

  chosen_ptr->hom_chromos[1]->id = 1;
  chosen_ptr->hom_chromos[1]->number = 1;
  chosen_ptr->hom_chromos[1]->strands[0]->id = 0;
  chosen_ptr->hom_chromos[1]->strands[0]->number = 0;
  chosen_ptr->hom_chromos[1]->strands[1]->id = 1;
  chosen_ptr->hom_chromos[1]->strands[1]->number = 1;
  
  /////more...
  chosen_ptr->hom_chromos[2]->id = 2;
  chosen_ptr->hom_chromos[2]->number = 0;
  chosen_ptr->hom_chromos[2]->strands[0]->id = 0;
  chosen_ptr->hom_chromos[2]->strands[0]->number = 0;
  chosen_ptr->hom_chromos[2]->strands[1]->id = 1;
  chosen_ptr->hom_chromos[2]->strands[1]->number = 1;

  chosen_ptr->hom_chromos[3]->id = 3;
  chosen_ptr->hom_chromos[3]->number = 1;
  chosen_ptr->hom_chromos[3]->strands[0]->id = 0;
  chosen_ptr->hom_chromos[3]->strands[0]->number = 0;
  chosen_ptr->hom_chromos[3]->strands[1]->id = 1;
  chosen_ptr->hom_chromos[3]->strands[1]->number = 1;
  
  chosen_ptr->hom_chromos[4]->id = 4;
  chosen_ptr->hom_chromos[4]->number = 0;
  chosen_ptr->hom_chromos[4]->strands[0]->id = 0;
  chosen_ptr->hom_chromos[4]->strands[0]->number = 0;
  chosen_ptr->hom_chromos[4]->strands[1]->id = 1;
  chosen_ptr->hom_chromos[4]->strands[1]->number = 1;

  chosen_ptr->hom_chromos[5]->id = 5;
  chosen_ptr->hom_chromos[5]->number = 1;
  chosen_ptr->hom_chromos[5]->strands[0]->id = 0;
  chosen_ptr->hom_chromos[5]->strands[0]->number = 0;
  chosen_ptr->hom_chromos[5]->strands[1]->id = 1;
  chosen_ptr->hom_chromos[5]->strands[1]->number = 1;
  /////

  strcpy(chosen_ptr->hom_chromos[0]->strands[0]->sequence, "ATTA");
  strcpy(chosen_ptr->hom_chromos[0]->strands[1]->sequence, "TAGG");

  strcpy(chosen_ptr->hom_chromos[1]->strands[0]->sequence, "TTCC");
  strcpy(chosen_ptr->hom_chromos[1]->strands[1]->sequence, "TCAG");

  //////////////more...
  strcpy(chosen_ptr->hom_chromos[2]->strands[0]->sequence, "GTAC");
  strcpy(chosen_ptr->hom_chromos[2]->strands[1]->sequence, "TTAC");

  strcpy(chosen_ptr->hom_chromos[3]->strands[0]->sequence, "CCCC");
  strcpy(chosen_ptr->hom_chromos[3]->strands[1]->sequence, "ATGG");

  strcpy(chosen_ptr->hom_chromos[4]->strands[0]->sequence, "TGAA");
  strcpy(chosen_ptr->hom_chromos[4]->strands[1]->sequence, "CCTT");

  strcpy(chosen_ptr->hom_chromos[5]->strands[0]->sequence, "AATC");
  strcpy(chosen_ptr->hom_chromos[5]->strands[1]->sequence, "TCTC");
  //////////////

  nd->num_hom_chromos = 4;
  nd->num_unique_hom_chromos = 2;
  nd->num_strands_per_hom_chrom = 2;

  nd->hom_chromos[0]->id = 0;
  nd->hom_chromos[0]->number = 0;
  nd->hom_chromos[0]->strands[0]->id = 0;
  nd->hom_chromos[0]->strands[0]->number = 0;
  nd->hom_chromos[0]->strands[1]->id = 1;
  nd->hom_chromos[0]->strands[1]->number = 1;

  /////////////////more....
  nd->hom_chromos[1]->id = 1;
  nd->hom_chromos[1]->number = 1;
  nd->hom_chromos[1]->strands[0]->id = 0;
  nd->hom_chromos[1]->strands[0]->number = 0;
  nd->hom_chromos[1]->strands[1]->id = 1;
  nd->hom_chromos[1]->strands[1]->number = 1;

  nd->hom_chromos[2]->id = 2;
  nd->hom_chromos[2]->number = 0;
  nd->hom_chromos[2]->strands[0]->id = 0;
  nd->hom_chromos[2]->strands[0]->number = 0;
  nd->hom_chromos[2]->strands[1]->id = 1;
  nd->hom_chromos[2]->strands[1]->number = 1;

  nd->hom_chromos[3]->id = 3;
  nd->hom_chromos[3]->number = 1;
  nd->hom_chromos[3]->strands[0]->id = 0;
  nd->hom_chromos[3]->strands[0]->number = 0;
  nd->hom_chromos[3]->strands[1]->id = 1;
  nd->hom_chromos[3]->strands[1]->number = 1;
  /////////////////

  strcpy(nd->hom_chromos[0]->strands[0]->sequence, "ATTA");
  strcpy(nd->hom_chromos[0]->strands[1]->sequence, "TAGG");

  ///////////more....
  strcpy(nd->hom_chromos[1]->strands[0]->sequence, "TTCC");
  strcpy(nd->hom_chromos[1]->strands[1]->sequence, "TCAG");

  strcpy(nd->hom_chromos[2]->strands[0]->sequence, "ATTA");
  strcpy(nd->hom_chromos[2]->strands[1]->sequence, "TAGG");

  strcpy(nd->hom_chromos[3]->strands[0]->sequence, "TTCC");
  strcpy(nd->hom_chromos[3]->strands[1]->sequence, "TCAG");
  ////////

  h_dist = HammingDistanceAvgNodes(chosen_ptr, nd);
  
  printf("NETGEN DEBUG:  avg hamming distance is %ld\n", h_dist);
  exit(0);

  return;
}

/******************************************************************************/
/*                          PolyploidDebugging                                */
/******************************************************************************/
void PolyploidDebugging(NODE nd)
{
  int hchromo_i;
  int strand_i;

  printf("NETGEN:  This was a polyploid hybrid:\n");
  printf("NETGEN:  The node id is %ld and its two parents ids are %ld and %ld\n", nd->id, nd->parent1->id, nd->parent2->id);
  printf("NETGEN:  hybrid sequences are:\n");
  for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nd->num_strands_per_hom_chrom; strand_i++)
    {
      printf("NETGEN: hchromo %d/strand %d: %s (ancestor node id %ld, hchromo number %d, strand number %d)\n", hchromo_i, strand_i, nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, nd->hom_chromos[hchromo_i]->strands[strand_i]->ancestor_node_id, nd->hom_chromos[hchromo_i]->number, nd->hom_chromos[hchromo_i]->strands[strand_i]->number); 
    }
  }

  printf("NETGEN:  parent1 sequences are:\n");
  for (hchromo_i = 0; hchromo_i < nd->parent1->num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; 
         strand_i < nd->parent1->num_strands_per_hom_chrom; 
	 strand_i++)
    {
      printf("NETGEN: hchromo %d/strand %d: %s (hchromo number %d, strand number %d)\n", hchromo_i, strand_i, nd->parent1->hom_chromos[hchromo_i]->strands[strand_i]->sequence, nd->parent1->hom_chromos[hchromo_i]->number, nd->parent1->hom_chromos[hchromo_i]->strands[strand_i]->number); 
    }
  }

  printf("NETGEN:  parent2 sequences are:\n");
  for (hchromo_i = 0; hchromo_i < nd->parent2->num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; 
         strand_i < nd->parent2->num_strands_per_hom_chrom; 
	 strand_i++)
    {
      printf("NETGEN: hchromo %d/strand %d: %s (hchromo number %d, strand number %d)\n", hchromo_i, strand_i, nd->parent2->hom_chromos[hchromo_i]->strands[strand_i]->sequence, nd->parent2->hom_chromos[hchromo_i]->number, nd->parent2->hom_chromos[hchromo_i]->strands[strand_i]->number); 
    }
  }
}
  
