/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include "globals.h"
#include "node.h"
#include "event.h"
#include "network.h"
#include "utilities.h"
#include "input.h"

/*****************************************************************************/
/*                         Read the input file!                              */
/*****************************************************************************/
void ReadParams() 
{
  char str[MAX_LINE_LENGTH];
  char * eol_ptr;

  eol_ptr = NULL;

  // set all default values -- alphabetical order
  // (the only exception is user_seed_netgen which is set in main.c
  //  as part of the input line before this is even called)
  
  desired_height = DEFAULT_DESIRED_HEIGHT;
  desired_height_given_yn = 0;
  dip_rate = DEFAULT_DIP_RATE;
  expected_height = -1;
  ex_rate = 0.0;
  ex_rate_var = DEFAULT_EX_RATE_VAR;
  fake_branch_lengths_yn = 0;
  gamma_scale_value = DEFAULT_GAMMA_SCALE_VALUE;
  gamma_shape_value = DEFAULT_GAMMA_SHAPE_VALUE;
  hamming_edge_report = 0;
  hhd_from_root_report = 0;
  height_scalar = 1; // if not changed by user input, brlen1/2_ev = brlen1/2_t
  hyb_ev_dist_thresh = HYB_EV_DIST_THRESH_DEFAULT;
  hyb_exp_fn_thresh_rate = HYB_EXP_FN_THRESH_RATE_DEFAULT;
  hyb_ham_thresh_rate = HYB_HAM_THRESH_RATE_DEFAULT;
  hyb_parent_selection = 0; // default of min hamming for second parent
  hyb_rate = 0.0;
  hyb_rate_var = DEFAULT_HYB_RATE_VAR;
  initial_root_sequences_yn = 0;
  init_sequence_length = 0;
  lambda_hd = DEFAULT_LAMBDA_HD;
  max_num_hybrids = -1;
  max_outg_diff_value = 0;
  max_outg_tries = 0;
  min_outg_diff_value = 0;
  modified_newick_report = 0;
  node_listing_report;
  nsgw_seeds = NULL;
  nsgw_seeds_index = 0;
  nsgw_seeds_node = NULL;
  nsgw_seeds_node_index = 0;
  num_extant_taxa_req = 0;
  num_root_hom_chromos = 0;
  num_root_strands_per_hc = 0;
  outgroup = 0;
  outgroup_done = 0;
  poly_rate = DEFAULT_POLY_RATE;
  print_opt_reports = 0;
  report_branch_hd = 0;
  seq_gen_options_yn = 0;
  sequences_yn = 1;
  sp_rate = 0.0;
  sp_rate_var = DEFAULT_SP_RATE_VAR;
  splitstree_report = 0;
  temp_perc_max_outg_diff = 0.0;
  temp_perc_min_outg_diff = 0.0;
  temp_sequence_length = 0;
  track_nsgw_seeds_yn = 0;
  update_seqs_for_all_events_yn = 0;
  ultrametric_yn = 1;
  value = 0.0;
  var_rate_yn = 0;

  while(fscanf(fp_in, "%s", str) != EOF) {
    if (strncmp(str,"#",1) == 0) {
      if (strchr(str,'\n') == NULL)
      {
        // no end of line found so do getline to ignore rest of line
	fgets(str, MAX_LINE_LENGTH, fp_in);
      }
      else
      {
	// end of line was found, so do nothing but proceed with processing
        ;
      }
    } else if (strcmp(str,"num_extant_taxa") == 0) {
      fscanf(fp_in, "%d", &num_extant_taxa_req);
    } else if (strcmp(str,"track_nsgw_seeds") == 0) {
      track_nsgw_seeds_yn = 1;
    } else if (strcmp(str,"desired_height") == 0) {
      desired_height_given_yn = 1;
      fscanf(fp_in, "%lf", &desired_height);
    } else if (strcmp(str,"sp_rate") == 0) {
      fscanf(fp_in, "%lf", &sp_rate);
    } else if (strcmp(str,"sp_rate_var") == 0) {
      fscanf(fp_in, "%lf", &sp_rate_var);
    } else if (strcmp(str,"hyb_rate") == 0) {
      fscanf(fp_in, "%lf", &hyb_rate);
    } else if (strcmp(str,"hyb_rate_var") == 0) {
      fscanf(fp_in, "%lf", &hyb_rate_var);
    } else if (strcmp(str,"ex_rate") == 0) {
      fscanf(fp_in, "%lf", &ex_rate);
    } else if (strcmp(str,"ex_rate_var") == 0) {
      fscanf(fp_in, "%lf", &ex_rate_var);
    } else if (strcmp(str,"dip_rate") == 0) {
      fscanf(fp_in, "%lf", &dip_rate);
    } else if (strcmp(str,"poly_rate") == 0) {
      fscanf(fp_in, "%lf", &poly_rate);
    } else if (strcmp(str,"max_num_hybrids") == 0) {
      fscanf(fp_in, "%d", &max_num_hybrids);
    } else if (strcmp(str,"hyb_ev_dist_threshold") == 0) {
      fscanf(fp_in, "%lf", &hyb_ev_dist_thresh);
    } else if (strcmp(str,"hyb_exp_fn_threshold_rate") == 0) {
      fscanf(fp_in, "%lf", &hyb_exp_fn_thresh_rate);
    } else if (strcmp(str,"hyb_ham_threshold_rate") == 0) {
      fscanf(fp_in, "%lf", &hyb_ham_thresh_rate);
    } else if (strcmp(str,"hyb_parent_min_hamming") == 0) {
      hyb_parent_selection = 0;
    } else if (strcmp(str,"hyb_parent_random") == 0) {
      hyb_parent_selection = 1;
    } else if (strcmp(str,"hyb_parent_min_bl") == 0) {
      hyb_parent_selection = 2;
    } else if (strcmp(str,"hyb_parent_max_bl") == 0) {
      hyb_parent_selection = 3;
    } else if (strcmp(str,"hyb_parent_min_ev_distance") == 0) {
      hyb_parent_selection = 4;
    } else if (strcmp(str,"hyb_parent_exp_fn") == 0) {
      fscanf(fp_in, "%ld", &lambda_hd);
      hyb_parent_selection = 5;
    } else if (strcmp(str,"init_sequence_length") == 0) {
      fscanf(fp_in, "%d", &init_sequence_length);
    } else if (strcmp(str,"num_root_homologous_chromosomes") == 0) {
      fscanf(fp_in, "%d", &num_root_hom_chromos);
    } else if (strcmp(str,"num_root_strands_per_homologous_chromosome") == 0) {
      fscanf(fp_in, "%d", &num_root_strands_per_hc);
    } else if (strcmp(str,"initial_root_sequences") == 0) {
      initial_root_sequences_yn = 1;
      ReadRootSequences();
    } else if (strcmp(str,"outgroup") == 0) {
      outgroup = 1;
    } else if (strcmp(str,"max_outgroup_diff_perc") == 0) {
      fscanf(fp_in, "%lf", &temp_perc_max_outg_diff);
    } else if (strcmp(str,"min_outgroup_diff_perc") == 0) {
      fscanf(fp_in, "%lf", &temp_perc_min_outg_diff);
    } else if (strcmp(str,"max_outgroup_diff_value") == 0) {
      fscanf(fp_in, "%d", &max_outg_diff_value);
    } else if (strcmp(str,"min_outgroup_diff_value") == 0) {
      fscanf(fp_in, "%d", &min_outg_diff_value);
    } else if (strcmp(str,"max_outgroup_tries") == 0) {
      fscanf(fp_in, "%d", &max_outg_tries);
    } else if (strcmp(str,"simultaneous_sequences") == 0) {
      sequences_yn = 1;
    } else if (strcmp(str,"no_simultaneous_sequences") == 0) {
      sequences_yn = 0;
    } else if (strcmp(str,"non_ultrametric") == 0) {
      ultrametric_yn = 0;
      fscanf(fp_in, "%d", &gamma_shape_value);
      fscanf(fp_in, "%lf", &gamma_scale_value);
    } else if (strcmp(str,"variable_rate") == 0) {
      var_rate_yn = 1;
    } else if (strcmp(str,"report_branch_hd") == 0) {
      report_branch_hd = 1;
      print_opt_reports = 1;
    } else if (strcmp(str,"update_seqs_for_all_events") == 0) {
      update_seqs_for_all_events_yn = 1;
      report_branch_hd = 1; // force the report even if not requested
      print_opt_reports = 1;
    } else if (strcmp(str,"hamming_edge_report") == 0) {
      hamming_edge_report = 1;
      print_opt_reports = 1;
    } else if (strcmp(str,"hhd_from_root_report") == 0) {
      hhd_from_root_report = 1;
      print_opt_reports = 1;
    } else if (strcmp(str,"modified_newick_report") == 0) {
      modified_newick_report = 1;
      print_opt_reports = 1;
    } else if (strcmp(str,"node_listing_report") == 0) {
      node_listing_report = 1;
      print_opt_reports = 1;
    } else if (strcmp(str,"splitstree_report") == 0) {
      splitstree_report = 1;
      print_opt_reports = 1;
    } else if (strcmp(str,"fake_branch_lengths") == 0) {
      fake_branch_lengths_yn = 1;
    } else if (strcmp(str,"seq_gen_options") == 0) {
      seq_gen_options_yn = 1;
      fgets(seq_gen_options, (SEQ_GEN_OPTIONS_STR_LENGTH-1), fp_in);
    } else {
      printf("NETGEN ERROR:  Unknown parameter in input file: %s\n", str);
      exit(0);
    }
  }


  /****************************************************************************/
  /*           Sanity check the input and do any necessary setup              */
  /****************************************************************************/

  ProcessSize();

  ProcessUltrametricity();

  ProcessHeight();

  ProcessRates();

  ProcessRandomSeeds();

  ProcessSeqGenOptions();

  ProcessSeqInfo();

  ProcessOutgroupInfo();

  ProcessRootSequences();

  ProcessHybridOption();

}


/**************************************************************************/
/*                 the helper functions broken out from                   */
/*                            ReadParams                                  */
/**************************************************************************/

void ReadRootSequences()
{
  char temp_root_sequence[MAX_SEQ_LENGTH+2];
  char junk_char[2];
  int seq_i, num_of_lines;

  // make sure we have the necessary info to do the reading
  if ( (num_root_hom_chromos == 0) || (init_sequence_length == 0) ||
       (num_root_strands_per_hc) == 0)
  {
    printf("NETGEN ERROR:  Cannot specify root sequences because num_root_homologous_chromosomes and/or init_sequence_length and/or num_root_strands_per_homologous_chromosome not yet specified\n");
    exit(0);
  }

  if (init_sequence_length > MAX_SEQ_LENGTH)
  {
    printf("NETGEN ERROR: Inputted sequence length is too long\n");
    printf("NETGEN ERROR: It exceeds the max default which is %d\n", MAX_SEQ_LENGTH);
    exit(0);
  }

  // reserve the space we need (note that the array of char pointers has
  // already been set up by globals.c -- it is the space for each individual
  // sequence that we need to do here -- the + 2 is for carriage return and
  // null termination done automatically by fscanf when a %s is specified)
  for (seq_i = 0; seq_i < (num_root_hom_chromos * num_root_strands_per_hc); seq_i++)
  {
    root_sequences[seq_i] = (char *) calloc((init_sequence_length+2),sizeof(char));

    if (root_sequences[seq_i] == NULL)
    {
      printf("NETGEN ERROR:  Could not allocate space for initial root sequences\n");
      exit(0);
    }
  }

  // read the carriage return of the initial_root_sequences line so
  // we don't get messed up reading the root sequences
  fgets(&(junk_char[0]), 2, fp_in);

  // read the correct number of lines
  num_of_lines = num_root_hom_chromos * num_root_strands_per_hc;

  // now get all the root sequences which must be on the subsequent lines
  // note that fgets includes the carriage return and a null string 
  // terminator for the %s, so do length + 2
  for (seq_i = 0; seq_i < num_of_lines; seq_i++)
  {
    // read the correct length, but remember:
    //     * we want to read the newline (it gets replaced later)
    //       (but fgets stops when it sees a new line so be careful!)
    //     * we have room for the null termination which occurs with fgets %s
    //     * fgets copies n-1, so bump up length by 2
    fgets(&(temp_root_sequence[0]), (init_sequence_length + 2), fp_in);

    strcpy(root_sequences[seq_i], &(temp_root_sequence[0]));     

    // be sure to add the null string to the end of the sequence (this will
    // replace the carriage return we read in and that is what we want)
    root_sequences[seq_i][init_sequence_length] = '\0';
  }
}

void ProcessSize()
{

  // make sure we will have enough room to do what the user requested
  if (num_extant_taxa_req > (0.5 * MAX_NODES))
  {
    printf("NETGEN ERROR:  num_extant_taxa is too large for MAX_NODES default of %d\n", MAX_NODES);
    exit(0);
  }
}

void ProcessUltrametricity()
{
  if (ultrametric_yn == 0)
  {
    if ( (gamma_shape_value == DEFAULT_GAMMA_SHAPE_VALUE) ||
	 (gamma_scale_value == DEFAULT_GAMMA_SCALE_VALUE) )
    {
      printf("NETGEN ERROR:  Gamma shape/scale value(s) not properly specified for non-ultrametric option\n");
      exit(0);
    }
  }
}

void ProcessHeight()
{

  int    event_i;
  double net_birth_rate = sp_rate + hyb_rate - ex_rate;

  // only do all this work and recalculate the height scalar if the
  // user specified a desired height (otherwise the scalar was set to
  // 1 up above)

  if (desired_height_given_yn == 1)
  {

    // if the user specified a desired height, sanity check it 
    if ( (desired_height > 1.0) || (desired_height < 0) )
    {
      printf("NETGEN ERROR:  the desired height of a tree should be between 0 and 1 in order to make evolutionary sense\n");
      exit(0);
    }

    // now calculate the expected height of the tree depending on the minimum
    // number of (net) birth events needed to reach the desired number of extant 
    // taxa 
    //
    // NOTE:  the expected (average) time to reach the next event is 1/lambda
    //        where lambda is the (instantaneous rate) * (number active lineages)
    //        which is the event_i here -- this is because as there are more
    //        and more lineages, the time to the next event is less and less
    //        on average

    expected_height = 0;

    for (event_i = 2; event_i < num_extant_taxa_req; event_i++)
    {
      expected_height = expected_height + (1/(event_i * net_birth_rate));
    }

    // now calculate the height_scalar which will allow us to convert the
    // internal branch length times to meaningful evolutionary branch lengths
    // that in the end will yield the desired height on average (not always)
  
    height_scalar = desired_height / expected_height;

    // if doing a non-ultrametric tree, also need to multiply the height scalar
    // by 1/average gamma value because each ev branch will be getting a
    // gamma value scaling itself -- expected value of a gamma distribution
    // is eta/lambda which is shape/scale parameters

    if (ultrametric_yn == 0)
    {
      height_scalar = height_scalar * (1/(gamma_shape_value/gamma_scale_value));
    }

  }

}

void ProcessRates()
{
  double sum_of_rates;

  // With independent events, there is no longer a requirement that the 
  // speciation, extinction and hybrid rates sum to something less than
  // or equal to 1.0 -- but we do want something to happen, so make
  // sure speciation is greater than 0.0 (the default) at least
  
  if ( (sp_rate <= 0.0) && (hyb_rate <= 0.0) )
  {
    printf("NETGEN ERROR:  Need a sp_rate or hybrid rate greater than 0.0 for a simulation\n");
    exit(0);
  }

  // make sure if the variances are not the defaults, the user specified
  // variable_rate parameter
  if ( (sp_rate_var != DEFAULT_SP_RATE_VAR) || 
       (ex_rate_var != DEFAULT_EX_RATE_VAR) || 
       (hyb_rate_var != DEFAULT_HYB_RATE_VAR) )
  {
    if (var_rate_yn == 0) 
    {
      printf("NETGEN ERROR:  Variance values should not be specified unless running with variable_rate option");
      exit(0);
    }
  }

  // make sure all the variances are positive
  if ( (sp_rate_var <= 0.0) || (ex_rate_var <= 0.0) || (hyb_rate_var <= 0.0) )
  {
    printf("NETGEN ERROR:  The variance values should be greater than 0.0\n");
    exit(0);
  }

  // If the user requested 0 for the max number of hybrids, 
  // set overall hybrid rate to 0 as well as the the sub rates
  // for diploid and polyploid hybridization types to 0
  if (max_num_hybrids == 0) 
  {
    hyb_rate = 0.0;
    dip_rate = 0.0;
    poly_rate = 0.0;
  }

  // we currently do diploid and polyploid hybridizations -- thus
  // these two subrates should sum to 1.0

  sum_of_rates = (dip_rate + poly_rate);

  // the rates of all hybrids (if present at all) should add up to 1
  if ( (hyb_rate != 0.0) && ((sum_of_rates < 1.0) || (sum_of_rates > 1.0)))
  {
    printf("NETGEN ERROR:  The sum of diploid and polyploid rates should be 1.0 (not %f) \n", dip_rate + poly_rate);
    exit(0);
  }
}

void ProcessRandomSeeds()
{
  // check that random number seed for twister is a legit value -- we 
  // need an unsigned long which is the same as a positive integer 
  //
  // there is no real way to check this because when we read it in as an
  // unsigned long that is how it gets cast (so even if the user enters
  // a negative number it gets converted to positive -- but at least the
  // print routine says what the seed used was, so the user has to note
  // that it is not the seed they requested)
  //
  // (the max value is 4294967295, but do not test against this
  // value because the computer does not know anything 'bigger' than this)
  

  // check and see if we will be tracking NSGW seeds (and the related nodes)
  // if so set up the arrays to store them here
  if (track_nsgw_seeds_yn == 1)
  {
    nsgw_seeds = (unsigned long *) calloc(MAX_NSGW_SEEDS, sizeof(unsigned long));
    nsgw_seeds_node = (int *) calloc(MAX_NSGW_SEEDS, sizeof(int));

    if ( (nsgw_seeds == NULL) || (nsgw_seeds_node == NULL) )
    {
      printf("NETGEN ERROR: Could not allocate space NSGW seed tracking as requested\n");
      exit(0);
    }
  }
}

void ProcessSeqGenOptions()
{
  int char_i, start_search_i;

  // deal with seq_gen_options -- the new version of SeqGen requires a model
  // to be specified, and according to their manual, the HKY model with
  // none of the additional parameters set is equivalent to the Jukes-Cantor
  // which appears to be the most 'simple' so we use that here
  if (seq_gen_options_yn == 0)
  {
    strcpy(seq_gen_options, "\"-mHKY\""); 
  }

  // if seq gen options are specified -- make sure we have the quotes
  // and remove the newline at the end -- the quotes are needed by
  // NSGW and also let us sanity check that the user did not forget
  // to include the options after specifying seq_gen_options!
  // NOTE:  the quotes will be printed as part of the call to NSGW, but
  // when it reads them in, it signals the string is a single command
  // line argument (which we want it to be) and the quotes are removed,
  // which is what we want for when Seq-Gen is called, (and the new Seq-Gen
  // needs it such that there is no space between -m and the model name)
  // so it all works out!
  if (seq_gen_options_yn)
  {
    if ((seq_gen_options[0] != '"') && (seq_gen_options[1] != '"'))
    {
      printf("NETGEN ERROR: seq_gen_options command line did not start with \"\n");
      printf("NETGEN ERROR: Specifying seq_gen_options is required to be followed by the options enclosed in quotes\n");
      exit(0);
    }

    if (seq_gen_options[0] == '"') start_search_i = 1;
    if (seq_gen_options[1] == '"') start_search_i = 2;

    // look for end "
    for (char_i = start_search_i; 
         char_i < (SEQ_GEN_OPTIONS_STR_LENGTH-1); 
	 char_i++)
    {
      if (seq_gen_options[char_i] == '"')
        break;
    }

    // add null to end of string
    seq_gen_options[char_i+1] = '\0';

    // check to see if the ending quote was ever found
    // (if the above loop went all the way to the end, it was not found,
    //  and it was either forgotten or the string is too long)
    if (char_i == (SEQ_GEN_OPTIONS_STR_LENGTH-1))
    {
      printf("NETGEN ERROR:  seq_gen_options command line did not end with \" (or the string itself exceeds the limit of %d)\n", SEQ_GEN_OPTIONS_STR_LENGTH);
      exit(0);
    }
  } // end of seq_gen_options string checking
 
  if ( (seq_gen_options_yn == 1) && (sequences_yn == 0) )
  {
    printf("NETGEN ERROR: Should not specify seq_gen_options unless running simultaneous sequence mode\n");
    exit(0);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: done processing seq_gen_options in input.c and we have %s\n", seq_gen_options);
}

void ProcessSeqInfo()
{
  // if no number of root homologous chromosomes given, set the defaults 
  if (num_root_hom_chromos == 0)
  {
    num_root_hom_chromos = DEFAULT_NUM_HOM_CHROMOS;
  }

  if (num_root_strands_per_hc == 0)
  {
    num_root_strands_per_hc = DEFAULT_NUM_STRANDS_PER_HC;
  }

  if (init_sequence_length == 0) 
  {
    init_sequence_length = DEFAULT_SEQ_LENGTH;
  }

  // sanity check sizes of inputted values
  if (num_root_hom_chromos > MAX_NUM_HOM_CHROMOS)
  {
    printf("NETGEN ERROR: Inputted number of root homoologous chromosomes is too large\n");
    printf("NETGEN ERROR: It exceeds the max default which is %d\n", MAX_NUM_HOM_CHROMOS);
    exit(0);
  }

  if (num_root_strands_per_hc > MAX_NUM_STRANDS_PER_HC)
  {
    printf("NETGEN ERROR: Inputted number of strands per homologous chromosome is too large\n");
    printf("NETGEN ERROR: It exceeds the max default which is %d\n", MAX_NUM_STRANDS_PER_HC);
    exit(0);
  }

}

void ProcessOutgroupInfo()
{
  // if we have an outgroup set up the desired differing range which 
  // dictates how different (but not too different) 
  //
  // note that we don't necessarily have to be running with simultaneous
  // sequences because one can always take the topology (and no ss) from
  // here into the NSG_wrapper for 'after' processing
  if (outgroup)
  {
    // depending on whether init_sequence_length was given or not, 
    // set up a sequence length value to be used in the following calcs
    if (init_sequence_length == 0)
      temp_sequence_length = MAX_SEQ_LENGTH;
    else
      temp_sequence_length = init_sequence_length;
    
    // set the defaults if no values given
    if ( (temp_perc_max_outg_diff == 0.0) && 
         (temp_perc_min_outg_diff == 0.0) &&
         (max_outg_diff_value == 0) && 
         (min_outg_diff_value == 0))
    {
      // note that the defaults are established as:
      //
      // (MIN_OUTG_SCALAR * 2 * desired height)
      // (MAX_OUTG_SCALAR * 2 * desired_height)
      //
      // This is because any two extant taxa will be on average 2 * height
      // apart;  and although distance is an upper bound on average hamming
      // distance (because of saturation and reversals), we want to make sure
      // the outgroup is noticeably different; finally we must also remember, 
      // we cannot go over 100% change in either case
      
      min_outg_diff_value = (int) ( (MIN_OUTG_SCALAR * 2.0 * desired_height) * 
	                            (temp_sequence_length) );

      max_outg_diff_value = (int) ( (MAX_OUTG_SCALAR * 2.0 * desired_height) * 
	                            (temp_sequence_length) );

      min_outg_diff_value = MinInt(min_outg_diff_value, temp_sequence_length);
      max_outg_diff_value = MinInt(max_outg_diff_value, temp_sequence_length);

    }
   
    // set the values if percentages were given
    if ( (max_outg_diff_value == 0) && 
         (min_outg_diff_value == 0) &&
	 (temp_perc_max_outg_diff != 0.0) && 
	 (temp_perc_min_outg_diff != 0.0))
    {
      max_outg_diff_value = (int)(temp_perc_max_outg_diff*temp_sequence_length);
      min_outg_diff_value = (int)(temp_perc_min_outg_diff*temp_sequence_length);
    }

    if ( (min_outg_diff_value > temp_sequence_length) ||
	 (min_outg_diff_value > max_outg_diff_value)  ||
	 (max_outg_diff_value > temp_sequence_length) ||
	 (max_outg_diff_value < min_outg_diff_value))
    {
      printf("NETGEN ERROR: Inconsistent values for min/max outgroup differences (%d, %d) (temp seq length: %d)\n", min_outg_diff_value, max_outg_diff_value, temp_sequence_length);
      exit(0);
    }

    if (max_outg_tries == 0)
      max_outg_tries = MAX_OUTGROUP_TRIES;     

    if (max_outg_tries > MAX_OUTGROUP_TRIES)
    {
      printf("NETGEN ERROR: Number of max tries for the outgroup is too large\n");
      printf("NETGEN ERROR: The max tries permitted is %d\n", MAX_OUTGROUP_TRIES);
      exit(0);
    }
  }
}

void ProcessRootSequences()
{
  int seq_i, char_i;

  // make sure the root sequences (if provided) are legitimate
  // note that the length check would be artificial as a \0 has already been
  // added to what "should" be the end of the string according to
  // init_sequence_length; if the user gave a sequence length too long as
  // compared to provided sequences it should show up here as there will be
  // invalid entries in the string; if the user gave a sequence length too
  // small, the smaller value will be used, but the user should notice that
  // shorter than desired strings were printed in the report
  if (initial_root_sequences_yn == 1)
  {
    for (seq_i = 0; 
	 seq_i < (num_root_hom_chromos * num_root_strands_per_hc); 
	 seq_i++)
    {
      for (char_i = 0; char_i < init_sequence_length; char_i++)
      {
        if ( (root_sequences[seq_i][char_i] != 'A') && 
             (root_sequences[seq_i][char_i] != 'G') &&
             (root_sequences[seq_i][char_i] != 'C') &&
             (root_sequences[seq_i][char_i] != 'T') )
        {
          printf("NETGEN ERROR: Root sequences are not legitimate\n");
	  printf("NETGEN ERROR: Found char is %c\n", root_sequences[seq_i][char_i]);
	  exit(0);
        }
      }
    }
  } // end checking root sequences
}

void ProcessHybridOption()
{
  long max_hd;

  max_hd = MaxHammingDistanceAvgNodes(NULL, NULL);

  if (hyb_parent_selection == 5)
  {
    if ((lambda_hd < 0) || (lambda_hd > max_hd))
    {
      printf("NETGEN ERROR:  parameter specifying the exponential function (lambda_hd) must be at least 0 and less than or equal to the max hamming distance possible (%ld) \n", max_hd);
      exit(0);
    }
  }

  if ( (sequences_yn == 0) && (hyb_rate > 0.0) )
  {
    if ( (hyb_parent_selection == 0) ||
	 (hyb_parent_selection == 5) )
    {
      printf("NETGEN ERROR:  Method for selecting second hybrid parent is not appropriate as there are no sequences requested with this run (either request simultaneous sequences or change second parent selection method)\n");
    }
  }

  // check appropriate threshold 

  if (hyb_parent_selection == 0)
  {
    if ( (hyb_ham_thresh_rate > 1.0) || (hyb_ham_thresh_rate < 0.0) )
    {
      printf("NETGEN ERROR:  hyb_ham_threshold_rate must be between 0 and 1\n");
      exit(0);
    }
  }

  if (hyb_parent_selection == 4)
  {
    if (hyb_ev_dist_thresh < 0.0)
    {
      printf("NETGEN ERROR:  hyb_ev_dist_threshold cannot be less than 0\n");
      exit(0);
    }
  }

  if (hyb_parent_selection == 5)
  {
    if ((hyb_exp_fn_thresh_rate < 0.0) || (hyb_exp_fn_thresh_rate > 1.0))
    {
      printf("NETGEN ERROR:  hyb_exp_fn_threshold_rate must be between 0.0 and 1.0\n");
      exit(0);
    }
  }


}
