/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<limits.h>
#include<float.h>
#include"globals.h"
#include"node.h"
#include"event.h"
#include"sequences.h"
#include"utilities.h"
#include"twister.h"
#include"network.h"

/******************************************************************************/
/*                              NewNetwork                                    */
/******************************************************************************/

NETWORK NewNetwork() 
{
  NETWORK nt = (NETWORK)malloc(sizeof(NETWORKTYPE));
 
  if (nt == NULL)
  {
    printf("NETGEN ERROR: Could not allocate space for network\n");
    exit(0);
  }
  
  nt->event_queue = NULL;
  nt->old_event_list = NULL;

  nt->node_id_ctr = 0;
  nt->num_act_lins = 0;
  nt->num_nodes = 0;
  nt->num_edges = 0;
  nt->num_leaves = 0;
  nt->num_extant_taxa = 0;
  nt->num_hybrids = 0;
  nt->parsimony_score = -1;
  nt->clock_height = 0.0;
  nt->ev_height = 0.0;
  nt->root = NULL;
  return nt;
}


/******************************************************************************/
/*                              CleanUpMemory                                 */
/******************************************************************************/
void CleanUpMemory(NETWORK nt)
{
  long node_i;
  int hchromo_i, strand_i, seq_i;
  EVENT temp_ev_ptr;


  // first clean up the network stuff
  for (node_i = 0; node_i < (nt->num_nodes); node_i++)
  {
    for (hchromo_i = 0;
         hchromo_i < nt->NodeArray[node_i]->num_hom_chromos;
	 hchromo_i++)
    {
      for (strand_i = 0;
           strand_i < nt->NodeArray[node_i]->num_strands_per_hom_chrom;
	   strand_i++)
      {
        free(nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
        free(nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]);
      }
      free(nt->NodeArray[node_i]->hom_chromos[hchromo_i]);
    }
    free(nt->NodeArray[node_i]);
  }

  // now clean up the event queue and the old event list
  while (nt->event_queue != NULL)
  {
    temp_ev_ptr = nt->event_queue;
    nt->event_queue = nt->event_queue->next_event;
    free(temp_ev_ptr);
  }

  while (nt->old_event_list != NULL)
  {
    temp_ev_ptr = nt->old_event_list;
    nt->old_event_list = nt->old_event_list->next_event;
    free(temp_ev_ptr);
  }

  // now we are finally free to free the network -- and it has no dynamically
  // allocated structures left so just go ahead and free it!
  free(nt);

  // now go ahead and clean up root sequences and nsgw_seeds if
  // they were used on this run

  if (initial_root_sequences_yn)
  {
    for (seq_i = 0; seq_i < (num_root_hom_chromos * num_root_strands_per_hc); seq_i++)
    {
      free(root_sequences[seq_i]);
    }
  }

  if (track_nsgw_seeds_yn)
  {
    free(nsgw_seeds);
  }

}

/******************************************************************************/
/*                          InitializeCandidatePool                           */
/******************************************************************************/
void InitializeCandidatePool(NETWORK nt)
{
  long node_i;

  for (node_i = 0; node_i < MAX_NODES; node_i++)
  {
    nt->Candidates[node_i] = -1;
  }

  return;
}


/******************************************************************************/
/*                         Connection Functions                               */
/******************************************************************************/

void ConnectOnlyChild(NODE parent, NODE child) 
{
  parent->left = child;
  parent->right = NULL;

  child->parent1 = parent;

  return;
}

void ConnectTwoChildren(NODE parent, NODE left_child, NODE right_child) 
{
  /**********************************/
  /* connect the kids to the parent */
  /**********************************/
  parent->left  = left_child;
  parent->right = right_child;
  
  
  /**********************************/
  /* connect the parent to the kids */
  /**********************************/
  // if the left child does not already have a parent, do parent1
  if (left_child->parent1 == NULL)  
  {
    left_child->parent1 = parent;
  }
  else // otherwise set parent up as parent2 
  {
    left_child->parent2 = parent;
  }
 
  // if the right child does not already have a parent, do parent2
  if (right_child->parent1 == NULL) 
  {
    right_child->parent1 = parent;
  }
  else 
  {
    right_child->parent2 = parent;
  }
  return;
}


/******************************************************************************/
/*                            MakeRootNode                                    */
/******************************************************************************/
void MakeRootNode(NETWORK nt, double current_time)
{
  int hchromo_i, strand_i, seq_i;
  long root_id = (nt->node_id_ctr)++;

  nt->root = NewLineage(root_id, num_root_hom_chromos, num_root_strands_per_hc, init_sequence_length, NULL);
  nt->root->time_created = current_time;
  nt->root->active = 0;
  nt->NodeArray[root_id] = nt->root;
  (nt->num_nodes)++;

  // DEBUGGING
  //printf("NETGEN DEBUG: init seq length from GenerateNetwork is %d\n", init_sequence_length);
  //printf("NETGEN DEBUG: root seq length from GenerateNetwork is %d\n", nt->root->seq_length);
  //printf("NETGEN DEBUG: root id from GenerateNetwork is %ld \n", nt->root->id);
  //printf("NETGEN DEBUG: initial_root_sequences_yn from GenerateNetwork is %d \n", initial_root_sequences_yn);
  // END DEBUGGING

  // sequence processing for root lineage if needed
  if (sequences_yn)
  {
    // if root sequences are given as input, copy them into where they belong;
    if (initial_root_sequences_yn == 1)
    {
      seq_i = 0;

      for (hchromo_i = 0; hchromo_i < nt->root->num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; 
	     strand_i < nt->root->num_strands_per_hom_chrom; 
	     strand_i++)
        {
          strcpy(nt->root->hom_chromos[hchromo_i]->strands[strand_i]->sequence, root_sequences[seq_i]);
	  seq_i++;
        }
      }

    }
    // otherwise get NSGW to come up with a root sequence randomly --
    // note that AssignSequence makes sure that within each homologous
    // chromosome, all sequences will match for the root case!
    else
    {
      AssignSequences(0, nt->NodeArray, nt->root->id, 0, current_time);
    }
  }
  else // we will be running with artifical sequences so assign a fake root
  {
    // note that if running with no sequences -- the user will not have
    // had the option to input a set of root sequences, so automatically
    // in this case go to artificial assignment
    AssignArtificialSequences(0, nt->NodeArray, nt->root->id, 0, current_time);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG:  seqs now stored in root as:\n");
  //for (hchromo_i = 0; hchromo_i < nt->root->num_hom_chromos; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nt->root->num_strands_per_hom_chrom; strand_i++)
    //{
      //printf("NETGEN DEBUG: hchromo_i %d, strand_i %d has seq:%s\n", strand_i, strand_i, nt->root->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
  // END DEBUGGING
}


/******************************************************************************/
/*                           MakeSubRootNode                                  */
/******************************************************************************/
void MakeSubRootNode(NETWORK nt, double current_time)
{
  int nhc, nsphc, sl;
  long na_index;

  // if we have an outgroup we essentially need another root since the first
  // root will act as a 'super root' which will have two kids -- the outgroup
  // and the 'sub root' which will be the root of everything else; so set
  // up that 'sub root' here if needed (note that the assumption is that
  // the edge between the two 'roots' is assumed to be zero, so most likely
  // they will end up with the same sequences if assigning sequences with
  // Seq-Gen); the outgroup leaf is generated at the end of processing
  //

  nhc = nt->root->num_hom_chromos;
  nsphc = nt->root->num_strands_per_hom_chrom;
  sl = nt->root->seq_length;

  na_index = nt->node_id_ctr++;
  nt->NodeArray[na_index] = NewLineage(na_index, nhc, nsphc, sl, nt->root);
  nt->NodeArray[na_index]->time_created = current_time;
  nt->NodeArray[na_index]->active = 0;
  nt->NodeArray[na_index]->parent1 = nt->root;
  nt->NodeArray[na_index]->brlen1_ev = 0.0; //branch between sub and super roots
  nt->NodeArray[na_index]->brlen1_t = 0.0; // evolutionary time is 0
  nt->root->left = nt->NodeArray[na_index]; // clock time here is arbitrary at 0
  (nt->num_nodes)++;


  if (sequences_yn)
  {
    AssignSequences(1, nt->NodeArray, na_index, 0, current_time);
  }
  else
  {
    AssignArtificialSequences(1, nt->NodeArray, na_index, 0, current_time);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG:  done with MakeSubRootNode, we now have %ld nodes\n", nt->num_nodes);
}


/******************************************************************************/
/*                          MakeFirstActLins                                  */
/******************************************************************************/
void MakeFirstActLins(NETWORK nt, double current_time)
{
  int nhc = nt->root->num_hom_chromos;
  int nsphc = nt->root->num_strands_per_hom_chrom;
  int sl = nt->root->seq_length;

  long first_index = (nt->node_id_ctr)++;
  long second_index = (nt->node_id_ctr)++;
  nt->num_nodes += 2;
  
  // make the lineages
  if (outgroup)
  {
    nt->NodeArray[first_index] = 
         NewLineage(first_index, nhc, nsphc, sl, nt->root->left);
    nt->NodeArray[second_index] = 
         NewLineage(second_index, nhc, nsphc, sl, nt->root->left);
  }
  else
  {
    nt->NodeArray[first_index] = 
         NewLineage(first_index, nhc, nsphc, sl, nt->root);
    nt->NodeArray[second_index] = 
         NewLineage(second_index, nhc, nsphc, sl, nt->root);
  }

  // DEBUGGING
  //printf("NETGEN DEBUG:  from MakeFirstActLins we now have %ld nodes\n", nt->num_nodes);
  //printf("NETGEN DEBUG:  from MakeFirstActLins NodeArray[first_index] has id %ld\n", nt->NodeArray[first_index]->id);
  //printf("NETGEN DEBUG:  from MakeFirstActLins NodeArray[second_index] has id %ld\n", nt->NodeArray[second_index]->id);
  // END DEBUGGING

  // now the parent of the two lineages and some of the startup info 
  // is dependent on whether there is an outgroup or not
  if (outgroup)
  {
    nt->NodeArray[first_index]->parent1 = nt->root->left;
    nt->root->left->left = nt->NodeArray[first_index];

    nt->NodeArray[second_index]->parent1 = nt->root->left;
    nt->root->left->right = nt->NodeArray[second_index];

    nt->num_edges = 3;
  }
  else
  {
    nt->NodeArray[first_index]->parent1 = nt->root;
    nt->root->left = nt->NodeArray[first_index];

    nt->NodeArray[second_index]->parent1 = nt->root;
    nt->root->right = nt->NodeArray[second_index];

    nt->num_edges = 2;
  }

  // DEBUGGING
  //printf("NETGEN DEBUG:  done with MakeFirstActLins we now have %ld nodes\n", nt->num_nodes);
  // END DEBUGGING
}


/******************************************************************************/
/*                           MakeFinalLeaves                                  */
/******************************************************************************/
void MakeExtantLeaves(NETWORK nt, double current_time) 
{
  int lin_i, prev_update_yn;
  double time_distance;

  nt->num_extant_taxa = nt->num_act_lins;
  nt->num_leaves += nt->num_act_lins;

  for(lin_i = 0; lin_i < nt->num_nodes; lin_i++) 
  { 
    if (nt->NodeArray[lin_i]->active)
    {
      nt->NodeArray[lin_i]->time_created = current_time; 
      nt->NodeArray[lin_i]->leaf = 1;
      nt->NodeArray[lin_i]->extant = 1;

      // assign branch lengths
      if (fake_branch_lengths_yn)
      {
        nt->NodeArray[lin_i]->brlen1_ev = FakeBranchLength();
        nt->NodeArray[lin_i]->brlen1_t = nt->NodeArray[lin_i]->brlen1_ev;
      }
      else if (ultrametric_yn == 0) // doing an non-ultrametric network
      {
        // evolutionary time can deviate, clock time does not
        time_distance = (nt->NodeArray[lin_i]->time_created - 
	    	         nt->NodeArray[lin_i]->parent1->time_created);

        nt->NodeArray[lin_i]->brlen1_t = time_distance;

	// take tiem and scale according to desired size and finally mulitply
	// by lineage's gamma value so we will have a non-ultrametric network
        nt->NodeArray[lin_i]->brlen1_ev = time_distance * 
		                          height_scalar *
					  nt->NodeArray[lin_i]->gamma_value;

      }
      else
      {
        // evolutionary time can deviate, clock time does not
        time_distance = (nt->NodeArray[lin_i]->time_created - 
	    	         nt->NodeArray[lin_i]->parent1->time_created);

        nt->NodeArray[lin_i]->brlen1_t = time_distance;

        nt->NodeArray[lin_i]->brlen1_ev = time_distance * height_scalar;
      }

      // DEBUGGING
      //printf("NETGEN DEBUG: we have just done a final leaf, the node's id was %ld and the current time is %f with its parent's time being %f giving us a clock branch length of %f and an ev branch length of %f\n", nt->NodeArray[lin_i]->id, current_time, nt->NodeArray[lin_i]->parent1->time_created, nt->NodeArray[lin_i]->brlen1_t, nt->NodeArray[lin_i]->brlen1_ev);
      // END DEBUGGING
    }
  }

  // if we are doing sequences assign the leaves that are still active
  // note that extinct leaves were assigned their sequences when the 
  // lineage died
  if (sequences_yn)
  {
    for(lin_i = 0; lin_i < nt->num_nodes; lin_i++) 
    { 
      if (nt->NodeArray[lin_i]->active)
      {
        if (nt->NodeArray[lin_i]->prev_update_time == PREV_UPDATE_TIME_DEFAULT)
          prev_update_yn = 0;
        else
          prev_update_yn = 1;

        AssignSequences(1, nt->NodeArray, lin_i, prev_update_yn, current_time);
	nt->NodeArray[lin_i]->active = 0;
      }
    }
  }
  else // need to assign artificial sequences in this case
  {
    for(lin_i = 0; lin_i < nt->num_nodes; lin_i++) 
    { 
      if (nt->NodeArray[lin_i]->active)
      {
        if (nt->NodeArray[lin_i]->prev_update_time == PREV_UPDATE_TIME_DEFAULT)
          prev_update_yn = 0;
        else
          prev_update_yn = 1;

        AssignArtificialSequences(1, nt->NodeArray, lin_i, prev_update_yn, 
			          current_time);
	nt->NodeArray[lin_i]->active = 0;
      }
    }
  }

}


/******************************************************************************/
/*                          MakeOutgroupLeaf                                  */
/******************************************************************************/
void MakeOutgroupLeaf(NETWORK nt, double current_time) 
{
  NODE outg;
  double time_distance;

  int nhc = nt->root->num_hom_chromos;
  int nsphc = nt->root->num_strands_per_hom_chrom;
  int sl = nt->root->seq_length;

  // note that we count this outgroup leaf as a leaf,
  // but not as an extant taxa;  it is a contemporary
  // of the extant taxa, but we don't want to label
  // it as such for processing purposes

  outg = NewLineage(nt->node_id_ctr, nhc, nsphc, sl, nt->root);
  nt->NodeArray[nt->node_id_ctr] = outg;
  outg->leaf = 1;
  outg->outgroup = 1;
  outg->parent1 = nt->root;
  nt->root->right = outg;
  outg->seq_length = nt->root->seq_length;
  outg->time_created = current_time;

  if (fake_branch_lengths_yn)
  {
    outg->brlen1_t = FakeBranchLength();
    outg->brlen1_ev = outg->brlen1_t;
  }
  else if (ultrametric_yn == 0) // non-ultrametric network
  {
    time_distance = (outg->time_created - nt->root->time_created);

    outg->brlen1_t = time_distance;
    outg->brlen1_ev = time_distance * height_scalar * outg->gamma_value;
  }
  else // ultrametric network
  {
    time_distance = (outg->time_created - nt->root->time_created);
    outg->brlen1_t = time_distance;
    outg->brlen1_ev = time_distance * height_scalar;
  }
  (nt->node_id_ctr)++;
  (nt->num_nodes)++;
  (nt->num_leaves)++;
  (nt->num_edges)++;
  
  // if we are running without simultaneous sequences, don't need to
  // do the work to find a good sequence, but still need to assign
  // a fake sequence for code integrity sake
  if (sequences_yn)
  {
    AssignOutgroupSequences(nt->NodeArray, outg->id, nt->num_nodes, 
                            nt->num_extant_taxa, current_time);
  }
  else
  {
    AssignArtificialSequences(1, nt->NodeArray, outg->id, 0, current_time);
  }

  // note that the outgroup leaf was made -- to help with reporting
  outgroup_done = 1;
} 

/******************************************************************************/
/*                        Tree Hunting Related Functions                      */
/******************************************************************************/
int IsTree(NETWORK nt)
{
  int node_i;
  int result;

  result = 1;

  // go through all the nodes and see if we have any hybrid nodes
  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->hybrid != 0)
      result = 0;
  }

  return result;
}

// Resets the 'marker variable in all the nodes of the network rooted at 'nd'
void ResetMarkers(NODE nd)
{
  nd->marker = 0;

  if (nd->left != NULL)
    ResetMarkers(nd->left);

  if (nd->right != NULL)
    ResetMarkers(nd->right);
}

NODE FindLeftmostNode(NETWORK nt, NODE nd1, NODE nd2)
{
  // do a preorder traversal looking for nd1 and n2 -- return as
  // soon as one of them is found -- the "found_node" will be the 
  // leftmost node (lineage) by definition of a preorder traversal
  // (if neither found, NULL is returned)

  NODE found_node = NULL;

  found_node = TreeTraverseSearch(found_node, nd1, nd2, nt->root);

  return found_node;
}

NODE TreeTraverseSearch(NODE found_node, NODE nd1, NODE nd2, NODE treenode)
{
  // this is a simplistic preorder search for a tree...since we have
  // a network, we may repeat paths, but we will not enter an infinite
  // cycle as we have directed edges and only go "down" (ie a DAG)

  if (treenode != NULL)
  {

    // see if we have found one of our nodes
    if (treenode == nd1)
    {
      found_node = nd1;

      // DEBUGGING
      //printf("NETGEN DEBUG: we found nd1, in treetraversesearch\n");

      return (found_node);
    }

    if (treenode == nd2)
    {
      found_node = nd2;

      // DEBUGGING
      //printf("NETGEN DEBUG: we found nd2, in treetraversesearch\n");

      return (found_node);
    }

    if ( (found_node == NULL) && (treenode->left != NULL) )
    {
      found_node = TreeTraverseSearch(found_node, nd1, nd2, treenode->left);
    }

    if ( (found_node == NULL) && (treenode->right != NULL) )
    {
      found_node = TreeTraverseSearch(found_node, nd1, nd2, treenode->right);
    }

    // leaf or found_node case
    return (found_node); 

  }
  else
    return (NULL);  // should never happen, but want to keep compiler happy

}

/*****************************************************************************/
/*                           FakeBranchLength                                */
/*****************************************************************************/
double FakeBranchLength()
{
  double choice[3];
  int choice_index;

  // possible fake branch lengths
  choice[0] = 50.0;
  choice[1] = 50.0;
  choice[2] = 50.0;

  choice_index = (int) (genrand_real2()*3);

  return(choice[choice_index]);

}


/******************************************************************************/
/*                         UpdateAllBranchLengths                             */
/******************************************************************************/
void UpdateAllBranchLengths(NETWORK nt, double current_time)
{
  /**********************************************************************/
  /* go through all active lineages and update the branch lengths       */
  /* (both evolutionary and time clock) to make sure we have up-to-date */
  /* branch lengths before doing other functions which rely on the      */
  /* branch lengths (all active lineages have only one parent so only   */
  /* need to check brlen1_ev and brlen1_t)                              */
  /**********************************************************************/
  
  long lin_i;
  double time_distance;

  for (lin_i = 0; lin_i < (nt->num_nodes); lin_i++) 
  {
    if (nt->NodeArray[lin_i]->active)
    {
      if (fake_branch_lengths_yn)
      {
        nt->NodeArray[lin_i]->brlen1_ev = FakeBranchLength();
        nt->NodeArray[lin_i]->brlen1_t =  nt->NodeArray[lin_i]->brlen1_ev; 
      } 
      else if (ultrametric_yn == 0) // non-ultrametric network
      {
        time_distance = (current_time - 
			 nt->NodeArray[lin_i]->parent1->time_created);

        nt->NodeArray[lin_i]->brlen1_t =  time_distance;
        nt->NodeArray[lin_i]->brlen1_ev = time_distance * height_scalar *
		                          nt->NodeArray[lin_i]->gamma_value;
        
      }
      else // ultrametric network
      {
        time_distance = (current_time - 
			 nt->NodeArray[lin_i]->parent1->time_created);

        nt->NodeArray[lin_i]->brlen1_t =  time_distance;
        nt->NodeArray[lin_i]->brlen1_ev = time_distance * height_scalar;
      }
    } // end of active check
  } // next lineage
}

/******************************************************************************/
/*                           UpdateAllSequences                               */
/******************************************************************************/
void UpdateAllSequences(NETWORK nt, double current_time)
{
  /********************************************************************/
  /* go through and update sequences for all active lineages -- all   */
  /* active lineages only have one parent so don't need to deal with  */
  /* any of the 2 stuff                                               */
  /********************************************************************/

  long lin_i;
  int prev_update_yn;

  for (lin_i = 0; lin_i < (nt->num_nodes); lin_i++) 
  {
    if (nt->NodeArray[lin_i]->active)
    {
      // see if seqs have been previously updated in which case AssignSequences
      // needs to know about it so the previous sequence and time are used
      // to generate the current sequence (ie not based on their parent)
      if (nt->NodeArray[lin_i]->prev_update_time == PREV_UPDATE_TIME_DEFAULT)
        prev_update_yn = 0;
      else
        prev_update_yn = 1;

      // note that if we are running with "no simultaneous sequences" (which
      // MAY be the case if we are choosing second parents based on criteria
      // other than hamming distance), then we need to artifically assign 
      // sequences so the code still works, but we don't take the 
      // time/effort to call NSGW/SG
      if (sequences_yn)
        AssignSequences(1, nt->NodeArray, lin_i, prev_update_yn, current_time);
      else
        AssignArtificialSequences(1, nt->NodeArray, lin_i, prev_update_yn, current_time);

      // now that the lineage had its sequences updated, record this time
      nt->NodeArray[lin_i]->prev_update_time = current_time;
    }
  }
}


/******************************************************************************/
/*                              Height Functions                              */
/******************************************************************************/

// returns the longest path (according to internal clock time)
// from a node to a leaf
double ClockHeight(NODE nd)
{
  double br1, br2;

  // nowhere to go
  if (nd->leaf == 1)
  {
    return 0;
  }
  else
  {
    // only have a right subtree to look at
    if (nd->left == NULL)
    {
      if (nd->right->parent1 == nd)
        br1 = nd->right->brlen1_t;
      else
	br1 = nd->right->brlen2_t;

      return (br1 + ClockHeight(nd->right));
    }
    // only have a left subtree to look at
    else if (nd->right == NULL)
    {
      if (nd->left->parent1 == nd)
        br1 = nd->left->brlen1_t;
      else
        br1 = nd->left->brlen2_t;

      return (br1 + ClockHeight(nd->left));
    }
    // have both subtrees to look at
    else
    {
      if (nd->left->parent1 == nd)
        br1 = nd->left->brlen1_t;
      else
        br1 = nd->left->brlen2_t;

      if (nd->right->parent1 == nd)
        br2 = nd->right->brlen1_t;
      else
	br2 = nd->right->brlen2_t;

      return (MaxDouble( (br1 + ClockHeight(nd->left)), 
			 (br2 + ClockHeight(nd->right)) ));
    }
  }
}

// returns the longest path (according to evolutionary time) from
// the provided node node to a leaf of the network
double EvHeight(NODE nd)
{
  double br1, br2;

  // nowhere to go
  if (nd->leaf == 1)
  {
    return 0;
  }
  else
  {
    // only have a right subtree to look at
    if (nd->left == NULL)
    {
      if (nd->right->parent1 == nd)
        br1 = nd->right->brlen1_ev;
      else
	br1 = nd->right->brlen2_ev;

      return (br1 + EvHeight(nd->right));
    }
    // only have a left subtree to look at
    else if (nd->right == NULL)
    {
      if (nd->left->parent1 == nd)
        br1 = nd->left->brlen1_ev;
      else
        br1 = nd->left->brlen2_ev;

      return (br1 + EvHeight(nd->left));
    }
    // have both subtrees to look at
    else
    {
      if (nd->left->parent1 == nd)
        br1 = nd->left->brlen1_ev;
      else
        br1 = nd->left->brlen2_ev;

      if (nd->right->parent1 == nd)
        br2 = nd->right->brlen1_ev;
      else
	br2 = nd->right->brlen2_ev;

      return (MaxDouble( (br1+EvHeight(nd->left)), (br2+EvHeight(nd->right))));
    }
  }
}

// returns the average height (Hamming distance defined!) from the extant
// taxa to the root -- note that Hamming distance used is total not average
// for two nodes (as we use elsewhere at times) -- this routine assumes
// the work of setting the HD from the root has already been done
// (with the UpdateAllHDsFromRoot below)
double AvgHDHeight(NETWORK nt)
{
  NODE nd;
  int node_i;
  long hd_total;
  long hd;
  double hd_result;
  int  denom_count;

  denom_count = 0;
  hd = 0;
  hd_total = 0;

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    if ( ((nt->NodeArray[node_i]->extant) && (nt->NodeArray[node_i]->leaf)) ||
          (nt->NodeArray[node_i]->outgroup == 1) )
    {
      nd = nt->NodeArray[node_i];
      denom_count++;

      hd = nd->hd_from_root;

      hd_total = hd_total + hd;
    }
  }

  hd_result = (double) (((double) hd_total)/((double) denom_count));


  // FIX LATER
  // DEBUGGING
  //printf("NG DEBUG:  hd_total is %ld, denom_count is %ld, and hd_result is %f\n", hd_total, denom_count, hd_result);
  
  return hd_result;

}

// go through all the nodes and calculate the total HD between it and the root
void UpdateAllHDsFromRoot(NETWORK nt)
{
  NODE nd;
  int node_i;

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    nd = nt->NodeArray[node_i];

    nd->hd_from_root = HammingDistanceNodes(nd, nt->root);

    // FIX LATER
    // DEBUGGING
    //printf("NG DEBUG:  dist bt %ld and root is %ld\n", nd->id, nd->hd_from_root);
  }

  return;

}


/******************************************************************************/
/*                             CalcParsScore                                  */
/******************************************************************************/
long CalcParsScore(NETWORK nt)
{
  // This routine is where we go through the network topology and calculate
  // the hamming distance for every strand sequence over every parent
  // child that is goes along.  Note that for diploid hybrids, not all 
  // sequences continue on through the hybrid, but with polyploid they
  // do, but in both cases, the seqs can change their location without
  // an evolutionary change.  Therefore the logic works as performing the
  // work from the root and every other hybrid node so that we keep
  // all this straight. (By definition there is no change along the
  // "vertical" edges from parents to hybrid node, so we do not need
  // to examine those.)  We also store these results on a per branch
  // basis in case we want to report the individual results.

  long hd_sum_strand, total_hd_sum;
  int node_i, hchromo_i, strand_i;

  total_hd_sum = 0;

  // start by doing all the work for the root
  for (hchromo_i = 0; hchromo_i < nt->root->num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nt->root->num_strands_per_hom_chrom; strand_i++)
    {
      hd_sum_strand = 0;
      ResetMarkers(nt->root);

      hd_sum_strand = CalcStrandHD(nt->root, hchromo_i, strand_i);

      //printf("NG DEBUG: hd sum strand for root of hchromo %i strand %i is %ld\n", hchromo_i, strand_i, hd_sum_strand);

      total_hd_sum = total_hd_sum + hd_sum_strand;

    }
  }

  // now repeat the process for all hybrid nodes -- but be careful in that
  // the first call is done with the hybrid node as the parent, not the
  // child -- there is no change from parent to hybrid node, so don't worry
  // about that either -- do be careful here as the hybrid node may have
  // a new number of homologous chromosomes and strands if it is a polyploid

  for (node_i = 0; node_i < (nt->num_nodes); node_i++)
  {
    if (nt->NodeArray[node_i]->hybrid != 0)
    {
      for (hchromo_i = 0; hchromo_i < nt->NodeArray[node_i]->num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; strand_i < nt->NodeArray[node_i]->num_strands_per_hom_chrom; strand_i++)
        {
	  hd_sum_strand = 0;
	  ResetMarkers(nt->root);

	  hd_sum_strand = CalcStrandHD(nt->NodeArray[node_i], hchromo_i, strand_i);

          //printf("NG DEBUG: hd sum strand for hybrid (id %ld) of hchromo %i strand %i is %ld\n", nt->NodeArray[node_i]->id, hchromo_i, strand_i, hd_sum_strand);

	  total_hd_sum = total_hd_sum + hd_sum_strand;
	}
      }
    }
  }

  return total_hd_sum; 
} 

/******************************************************************************/ 
/*                           CalcStrandHD                                     */
/******************************************************************************/
long CalcStrandHD(NODE rt, int hchromo_i, int strand_i)
{
  int orig_call;

  // This routine starts at the given node and figures out how many changes
  // occur for the given strand from the node down to where it shows
  // up in the extant taxa or hits a hybrid node.  This routine is called
  // for the real root and any hybrid node which acts as a root in this
  // case.
  
  orig_call = 1;
  strand_summed_hd = 0; // this is a global

  if (rt->hybrid != 0)
  {
    // if this is a hybrid node, the hybrid node is the parent
    HDDFS(rt, rt->left, hchromo_i, strand_i, orig_call);
  }
  else
  {
    // with the real root, there is no parent to pass here
    HDDFS(NULL, rt, hchromo_i, strand_i, orig_call);
  }

  // this variable is global which is "cheating" in the world of
  // recursion, but it sure makes our life easier, so, so be it!
  return strand_summed_hd;
}

/******************************************************************************/
/*                 HDDFS (hamming distance depth first search)                */
/******************************************************************************/
void HDDFS(NODE parent, NODE child, int hchromo_i, int strand_i, int orig_call)
{
  NODE tmp_nd;
  long hd;

  // first do the work for this pair of nodes if eligible
  if ( ((parent != NULL) && (parent->hybrid != 0) && (orig_call != 1)) ||
       ((child != NULL) && (child->hybrid != 0)) )
  {
    printf("NG DEBUG:  Not proceeding down this lineage in HDDFS as hybrid node hit\n");
    return;
  }
  else
  {
    if ( (child != NULL) && (child->marker == 0) && (parent != NULL) )
    {
      hd = HammingDistanceStrings(parent->hom_chromos[hchromo_i]->strands[strand_i]->sequence, child->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

      //printf("NG DEBUG:  HD for %ld to %ld is %ld (the seqs were %s %s)\n", parent->id, child->id, hd, parent->hom_chromos[hchromo_i]->strands[strand_i]->sequence, child->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

      // update this node's attributes appropriately
      child->marker = 1;
      
      // only parent1 case should really happen, but check to be careful...
      if (child->parent1 == parent)
        child->brlen1_hd = child->brlen1_hd + hd;
      else
      {
        printf("NG ERROR:  Unexpected branch traversal HDDFS (parent: %ld child: %ld) -- program aborted\n", parent->id, child->id);
	exit(0);
      }

      // update the globlal
      strand_summed_hd = hd + strand_summed_hd;
    }
  }

  // deal with the traversal
  if ( (child != NULL) && (child->left != NULL) )
  {
    HDDFS(child, child->left, hchromo_i, strand_i, 0);

    if (child->right != NULL)
      HDDFS(child, child->right, hchromo_i, strand_i, 0);
  }

  return;
}

/******************************************************************************/
/*                             DijkstraDist                                   */
/******************************************************************************/
double DijkstraDist(NETWORK nt, long source_id)
{
  long node_i;
  long node_to_process;

  // first we set the distance from the source to the source as zero and
  // we initialize all other dijkstra_dist attributes to DBL_MAX
  //
  // (we also initialize the markers and predecessors accordingly which 
  // indicates whether the lineage/node has been processed yet or not
  // and if so, which lineage/node is the predecessor)
  
  for (node_i = 0; node_i < (nt->num_nodes); node_i++)
  {
    if (node_i != source_id)
    {
      nt->NodeArray[node_i]->dijkstra_dist = DBL_MAX;
      nt->NodeArray[node_i]->dijkstra_pred = -1;
      nt->NodeArray[node_i]->marker = 0;
    }
    else
    {
      nt->NodeArray[node_i]->dijkstra_dist = 0;
      nt->NodeArray[node_i]->dijkstra_pred = -1;
      nt->NodeArray[node_i]->marker = 1;
    }
  }

  // this code is based upon the Dijkstra pseudo-code presented in
  // Cormen, Leiserson, and Rivest 1990 -- page 527
  //
  // this code is not as efficient though as we are not keeping a
  // priority queue, but this routine is used only for hybrid events
  // and hence should not be called that often to warrant the effort
  // of coding it more efficiently

  // always start with the source
  node_to_process = source_id;
  
  while (node_to_process != -1)
  {
    // for each of the possible 4 edges/branches, do relaxation
    RelaxNeighbors(nt, node_to_process);
    
    // mark this node as processed
    nt->NodeArray[node_to_process]->marker = 1;

    // set up the next node to process
    node_to_process = NextToProcess(nt);
  }

}

/******************************************************************************/
/*                 RelaxNeighbors(helper function for DijkstraDist)           */
/******************************************************************************/
void RelaxNeighbors(NETWORK nt, long node_to_process)
{

  if (nt->NodeArray[node_to_process]->left != NULL)
  {
    RelaxRoutine(nt,
		 node_to_process, 
	         nt->NodeArray[node_to_process]->left->id, 
	         nt->NodeArray[node_to_process]->left->brlen1_ev); 
  }

  if (nt->NodeArray[node_to_process]->right != NULL)
  {
    RelaxRoutine(nt,
		 node_to_process, 
                 nt->NodeArray[node_to_process]->right->id, 
	         nt->NodeArray[node_to_process]->right->brlen1_ev); 
  }

  if (nt->NodeArray[node_to_process]->parent1 != NULL)
  {
    RelaxRoutine(nt,
		 node_to_process, 
	         nt->NodeArray[node_to_process]->parent1->id, 
	         nt->NodeArray[node_to_process]->brlen1_ev); 
  }

  if (nt->NodeArray[node_to_process]->parent2 != NULL)
  {
    RelaxRoutine(nt,
		 node_to_process, 
	         nt->NodeArray[node_to_process]->parent2->id, 
	         nt->NodeArray[node_to_process]->brlen2_ev); 
  }

}

/******************************************************************************/
/*         RelaxRoutine(helper function for RelaxNeighbors/DijkstraDist)      */
/******************************************************************************/
void RelaxRoutine(NETWORK nt, long next_to_process_id, long neighbor_id, 
		  double bl)
{
  double new_dist;

  new_dist = nt->NodeArray[next_to_process_id]->dijkstra_dist + bl;

  if ( (nt->NodeArray[neighbor_id]->dijkstra_dist) > (new_dist) )
  {
     nt->NodeArray[neighbor_id]->dijkstra_dist = new_dist;
     nt->NodeArray[neighbor_id]->dijkstra_pred = next_to_process_id;
  }

}

/******************************************************************************/
/*               NextToProcess (helper function for DijkstraDist)             */
/******************************************************************************/
long NextToProcess(NETWORK nt)
{
  long node_i;
  long return_node;
  double min_dijkstra_dist;

  return_node = -1;
  min_dijkstra_dist = DBL_MAX;


  for (node_i = 0; node_i < (nt->num_nodes); node_i++)
  {
    if ((nt->NodeArray[node_i]->marker == 0) &&
        (nt->NodeArray[node_i]->dijkstra_dist < min_dijkstra_dist) )
    {
      min_dijkstra_dist = nt->NodeArray[node_i]->dijkstra_dist;
      return_node = node_i;
    }
  
  }

  return return_node;
}


