/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<limits.h>
#include<float.h>
#include"globals.h"
#include"node.h"
#include"event.h"
#include"network.h"
#include"sequences.h"
#include"utilities.h"
#include"twister.h"

/******************************************************************************/
/*                            NewSingleStrand(int, int)                       */
/******************************************************************************/
SINGLESTRAND NewSingleStrand(int strand_id, int sl)
{
  char *temp_ptr;
  char strand_id_str[10];
  SINGLESTRAND strand;

  strand = (SINGLESTRAND)malloc(sizeof(SINGLESTRANDTYPE));

  if (strand == NULL)
  {
    printf("NETGEN ERROR: Could not allocate space for strand\n");
    exit(0);
  }

  // add + 2 to seq length because will have a end of string character (\0)
  // for sure and possibly a carriage return we store, but never act on
  // (it comes back from NSGW in the sequence work)
  temp_ptr = (char *) calloc((sl+2),sizeof(char));

  if (temp_ptr == NULL)
  {
    printf("NETGEN ERROR: Cannot allocate memory for a strand's sequence\n");
    printf("NETGEN ERROR: a seq length of %d was requested\n", sl);
    exit(0);
  }

  strand->sequence = temp_ptr;

  // if the sequence length is long enough put in a temp name
  if (sl >= 25)
  {
    strcpy(strand->sequence, "init strand");
    sprintf(strand_id_str, "%d", strand_id);
    strcat(strand->sequence, strand_id_str);
  }

  // put in place holders for number and ancestor node id 
  // until the NewLineage routine rewrites
  strand->number = -1;
  strand->ancestor_node_id = -1;

  strand->id = strand_id;

  return strand;
}


/******************************************************************************/
/*                            NewHomChrom(int, int)                           */
/******************************************************************************/
HCHROMOSOME NewHomChrom(int hchromo_id, int nsphc,  int sl)
{
  int strand_i;
  HCHROMOSOME hchromo;

  hchromo = (HCHROMOSOME)malloc(sizeof(HCHROMOSOMETYPE));

  if (hchromo == NULL)
  {
    printf("NETGEN ERROR: Could not allocate space for homologous chromosome\n");
    exit(0);
  }

  for (strand_i = 0; strand_i < nsphc; strand_i++)
  {
    hchromo->strands[strand_i] = NewSingleStrand(strand_i, sl);
  }

  // put in place holder for number until NewLineage rewrites it
  hchromo->number = -1;
  hchromo->id = hchromo_id;

  return hchromo;
}


/******************************************************************************/
/*                            NewLineagae(int, int, int)                      */
/******************************************************************************/
LINEAGE NewLineage(long lin_id, int nhc, int nsphc, int sl, NODE parent_ptr) 
{
  int hchromo_i;
  int strand_i;

  // DEBUGGING
  //printf("NETGEN DEBUG: making new lineage with lin_id %ld, nhc %d, nsphc %d, and sl %d\n", lin_id, nhc, nsphc, sl);

  // make sure this lineage will not exceed our limit
  if (lin_id >= (MAX_NODES))
  {
    printf("NETGEN ERROR: Simulation could not proceed as MAX_NODES (%d) is exceeded\n", MAX_NODES);
    exit(0);
  }

  // make the new lineage
  LINEAGE lin;
  lin = (LINEAGE)malloc(sizeof(LINEAGETYPE));

  if (lin == NULL)
  {
    printf("NETGEN ERROR: Could not allocate space for lineage\n");
    exit(0);
  }

  // use the values we were given
  lin->id = lin_id;
  lin->num_hom_chromos = nhc;
  lin->num_strands_per_hom_chrom = nsphc; 
  lin->seq_length = sl;

  // set the rates for this lineage
  if (var_rate_yn == 0)
  {
    lin->birth_rate = sp_rate;
    lin->death_rate = ex_rate;
    lin->hybrid_rate = hyb_rate;
  }
  else
  {
    lin->birth_rate = NormRandHalfDist(sp_rate, sp_rate_var); 
    lin->death_rate = NormRandHalfDist(ex_rate, ex_rate_var); 
    lin->hybrid_rate = NormRandHalfDist(hyb_rate, hyb_rate_var); 
  }

  if (ultrametric_yn)
  {
    lin->gamma_value = DEFAULT_GAMMA_VALUE;
  }
  else
  {
    // this will be the value we use to scale the time branch lengths
    // to make non-ultrametric evolutionary branch lengths for this lineage
    lin->gamma_value = GammaDist(gamma_shape_value, gamma_scale_value);
  }

  // set all others to default values
  lin->left = NULL;
  lin->right = NULL;
  lin->parent1 = NULL;
  lin->parent2 = NULL;
  lin->active = 1;  // assume every new lineage is active
  lin->leaf = 0;   
  lin->extant = 0;
  lin->outgroup = 0;
  lin->hybrid = 0;
  lin->time_created = 0.0;
  lin->prev_update_time = PREV_UPDATE_TIME_DEFAULT;
  lin->brlen1_t = 0.0;
  lin->brlen2_t = 0.0;
  lin->brlen1_ev = 0.0;
  lin->brlen2_ev = 0.0;
  lin->brlen1_hd = 0.0;
  lin->brlen2_hd = 0.0;
  lin->marker = 0;
  lin->dijkstra_dist = DBL_MAX;
  lin->dijkstra_pred = -1;

  for (hchromo_i = 0; hchromo_i < lin->num_hom_chromos; hchromo_i++)
  {
    lin->hom_chromos[hchromo_i] = NewHomChrom(hchromo_i, lin->num_strands_per_hom_chrom, lin->seq_length);
  }

  // now set the numbers, ancestor ids, and num unique homologous
  // chromosomes according to the parent (remember that number is for 
  // biological tracking purposes -- it is not the same as the index!)
  // if these values are going to be reset (as in the hybrid case) 
  // or when there is no parent (as in the root case) use defaults

  if (parent_ptr == NULL)
  {
    lin->num_unique_hom_chromos = nhc;  

    for (hchromo_i = 0; hchromo_i < lin->num_hom_chromos; hchromo_i++)
    {
      lin->hom_chromos[hchromo_i]->number = hchromo_i;

      for (strand_i = 0; 
           strand_i < lin->num_strands_per_hom_chrom; 
	   strand_i++)
      {
        lin->hom_chromos[hchromo_i]->strands[strand_i]->number = strand_i;
        lin->hom_chromos[hchromo_i]->strands[strand_i]->ancestor_node_id = -1;
      }
    }
  }
  else
  {
    lin->num_unique_hom_chromos = parent_ptr->num_unique_hom_chromos;  

    for (hchromo_i = 0; hchromo_i < lin->num_hom_chromos; hchromo_i++)
    {
      lin->hom_chromos[hchromo_i]->number = 
        parent_ptr->hom_chromos[hchromo_i]->number;

      for (strand_i = 0; 
           strand_i < lin->num_strands_per_hom_chrom; 
	   strand_i++)
      {
        lin->hom_chromos[hchromo_i]->strands[strand_i]->number = 
	  parent_ptr->hom_chromos[hchromo_i]->strands[strand_i]->number;

        lin->hom_chromos[hchromo_i]->strands[strand_i]->ancestor_node_id = 
	  parent_ptr->id;
      }
    }
  }

  return lin;

}  // end of NewLineage()

/******************************************************************************/
/*                         HammingDistanceNodes                               */
/******************************************************************************/
long HammingDistanceNodes(NODE nd_1, NODE nd_2)
{
  // here we define that hamming distance between two nodes does
  // not care about the number of homologous chromosomes/strands -- simply
  // the number of sequences to compare (so if one node has one
  // chromosome with two strands and the other node has two 
  // chromosomes each with one strand, that is fine)
  //
  // here, if the number of sequences do not match, the number of 
  // nucleotides which are differing is added to the hamming distance
  // (e.g. if one node has two sequences of 10 each and the other
  // node has 1 sequence of 5 then the starting hamming distance
  // is 10 (as there is one extra string of length 10) -- the difference
  // in string length is captured and added to the hamming distance in
  // the HammingDistanceStrings function

  long num_seqs_1;
  long num_seqs_2;
  long num_seqs_to_compare;

  long hd;
  long seq_pair_i;

  int nd_1_hchromo_index;
  int nd_1_strand_index;
  int nd_2_hchromo_index;
  int nd_2_strand_index;

  char str_1[nd_1->seq_length];
  char str_2[nd_2->seq_length];

  hd = 0;
  num_seqs_1 = nd_1->num_hom_chromos * nd_1->num_strands_per_hom_chrom;
  num_seqs_2 = nd_2->num_hom_chromos * nd_2->num_strands_per_hom_chrom;

  if (num_seqs_1 == num_seqs_2)
  {
    num_seqs_to_compare = num_seqs_1;
  }
  else if (num_seqs_1 > num_seqs_2)
  {
    num_seqs_to_compare = num_seqs_2;
    hd = (num_seqs_1 - num_seqs_2) * nd_1->seq_length;
  }
  else
  {
    num_seqs_to_compare = num_seqs_1;
    hd = (num_seqs_2 - num_seqs_1) * nd_2->seq_length;
  }

  // FIX LATER
  // DEBUGGING
  //printf("NETGEN DEBUG:  hd before string comp is %ld\n", hd);

  // FIX LATER
  // DEBUGGING
  //printf("NETGEN DEBUG:  num seqs to compare is %ld, our node ids are %ld %ld\n", num_seqs_to_compare, nd_1->id, nd_2->id);

  // now call HammingDistanceStrings for each pair of strings
  // we have available for comparing
  nd_1_hchromo_index = 0;
  nd_1_strand_index = 0;
  nd_2_hchromo_index = 0;
  nd_2_strand_index = 0;

  for (seq_pair_i = 0; seq_pair_i < num_seqs_to_compare; seq_pair_i++)
  {
    // figure out which strings we will be comparing and copy them into 
    // str_1 and str_2 -- note that since the number of homologous 
    // chromosomes/strands may not match, need to track the index of the 
    // next string for each node and basically do mod arithmetic by hand 
    // -- there is probably a more elegant way to do this, but this should 
    // be straightforward and work
    
    strcpy(str_1, nd_1->hom_chromos[nd_1_hchromo_index]->strands[nd_1_strand_index]->sequence);

    nd_1_strand_index++;
    if (nd_1_strand_index == nd_1->num_strands_per_hom_chrom)
    {
      nd_1_hchromo_index++;
      nd_1_strand_index = 0;
    }

    strcpy(str_2, nd_2->hom_chromos[nd_2_hchromo_index]->strands[nd_2_strand_index]->sequence);

    nd_2_strand_index++;
    if (nd_2_strand_index == nd_2->num_strands_per_hom_chrom)
    {
      nd_2_hchromo_index++;
      nd_2_strand_index = 0;
    }

    // FIX LATER 
    // DEBUGGING
    //printf("NETGEN DEBUG:  num hchromo/strands indexes for node 2 during seq %d after copying sequence -- %d %d\n", seq_pair_i, nd_2_hchromo_index, nd_2_strand_index); 

    hd = HammingDistanceStrings(str_1, str_2) + hd;

    // FIX LATER
    // DEBUGGING
    //printf("NETGEN DEBUG:  hd from hamming distance nodes is %ld\n", hd);
    
  }

  return hd;
}

/******************************************************************************/
/*                        HammingDistanceAvgNodes                             */
/******************************************************************************/
long HammingDistanceAvgNodes(NODE nd_1, NODE nd_2)
{
  // in this routine we define hamming distance as an average

  long hd;
  long hd_sum_strand_level;
  long hd_sum_hchromo_level;
  long hd_sum_strands;
  long hd_sum_hchromos;
  long hd_avg_strand;
  long hd_avg_strands;
  long hd_avg_hchromos;
  long hd_avg;

  int num_unique_hcs;
  int num_strands_per_hc;
  int strand_i;
  int hchromo_i;
  int hchromo_j;
  int hchromo_k;
  int counter;

  // first we make sure that the number of unique homologous chromosomes
  // and the number of strands per chromosome match
  // for the two nodes -- doing an average does not
  // make sense unless this is true -- this should really be
  // checked before this routine is called, so this is an
  // extra check just in case

  if ( (nd_1->num_unique_hom_chromos != nd_2->num_unique_hom_chromos) ||
       (nd_1->num_strands_per_hom_chrom != nd_2->num_strands_per_hom_chrom) )
  {
    hd_avg = -1;
    return hd_avg;
  }

  num_unique_hcs = nd_1->num_unique_hom_chromos;
  num_strands_per_hc = nd_1->num_strands_per_hom_chrom;

  // for each unique chromosome and for each strand, go through and 
  // find the possible pairs between the two nodes and keep a running 
  // sum so that there is first an average calculated for each strand 
  // and then average across all strands for that chromosome and finally
  // across all unique homologous chromosomes

  hd_sum_hchromo_level = 0;

  for (hchromo_i = 0; hchromo_i < num_unique_hcs; hchromo_i++)
  {
    hd_sum_strands = 0;

    for (strand_i = 0; strand_i < num_strands_per_hc; strand_i++)
    {
      hd = 0;
      hd_sum_strand_level = 0;
      counter = 0;

      for (hchromo_j = 0; hchromo_j < nd_1->num_hom_chromos; hchromo_j++)
      {
        if (nd_1->hom_chromos[hchromo_j]->number == hchromo_i)
	{
          for (hchromo_k = 0; hchromo_k < nd_2->num_hom_chromos; hchromo_k++)
	  {
	    if (nd_1->hom_chromos[hchromo_j]->number == 
	        nd_2->hom_chromos[hchromo_k]->number)
	    {
              hd = HammingDistanceStrings(nd_1->hom_chromos[hchromo_j]->strands[strand_i]->sequence, nd_2->hom_chromos[hchromo_k]->strands[strand_i]->sequence); 
              // DEBUGGING
	      //printf("NETGEN DEBUG:  simple hd for two seqs is %ld (%s, %s)\n", hd, nd_1->hom_chromos[hchromo_j]->strands[strand_i]->sequence, nd_2->hom_chromos[hchromo_k]->strands[strand_i]->sequence);

	      hd_sum_strand_level = hd + hd_sum_strand_level;

	      counter++;
	    }
	  }
	}
      }
      hd_avg_strand = hd_sum_strand_level / counter;

      // DEBUGGING
      //printf("NETGEN DEBUG: hd avg for strand is %ld\n", hd_avg_strand);

      hd_sum_strands = hd_avg_strand + hd_sum_strands;
    }
    hd_avg_strands = hd_sum_strands / num_strands_per_hc;

    // DEBUGGING
    //printf("NETGEN DEBUG: hd avg for strandS is %ld\n", hd_avg_strands);

    hd_sum_hchromo_level = hd_avg_strands + hd_sum_hchromo_level;
  }

  hd_avg_hchromos = hd_sum_hchromo_level / num_unique_hcs;

  // DEBUGGING
  //printf("NETGEN DEBUG: hd avg for hchromOs is %ld\n", hd_avg_hchromos);

  hd_avg = hd_avg_hchromos;

  return hd_avg;
}

/******************************************************************************/
/*                         MaxHammingDistanceNodes                            */
/******************************************************************************/
long MaxHammingDistanceNodes(NODE nd_1, NODE nd_2)
{
  // Here we define max hamming distance to be the SUM of all sequences
  // lengths -- thus num_hom chromo * num_strands * seq_length which corresponds
  // to the hamming distance calculated in the HammingDistanceNodes function

  long ret_value;
  long nd_1_length;
  long nd_2_length;

  if ((nd_1 == NULL) && (nd_2 == NULL))
  {
    ret_value = (long) num_root_hom_chromos * 
                       num_root_strands_per_hc * 
		       init_sequence_length;
  }
  else
  {
    nd_1_length = (long) (nd_1->num_hom_chromos *
                          nd_1->num_strands_per_hom_chrom *
		          nd_1->seq_length);

    nd_2_length = (long) (nd_2->num_hom_chromos *
                          nd_2->num_strands_per_hom_chrom *
		          nd_2->seq_length);

    ret_value = (long) (MaxLong(nd_1_length, nd_2_length));
  }

  return ret_value;
  
}

/******************************************************************************/
/*                        MaxHammingDistanceAvgNodes                          */
/******************************************************************************/
long MaxHammingDistanceAvgNodes(NODE nd_1, NODE nd_2)
{
  // Here we define max hamming distance between nodes to be the 
  // sequence length of each strand sequence -- this corresponds to
  // the hamming distance being calculated as an average (as in the
  // function HammingDistanceAvgNodes above)

  long ret_value;

  if ((nd_1 == NULL) && (nd_2 == NULL))
    ret_value = (long) (init_sequence_length);
  else
    ret_value = (long) (MaxInt(nd_1->seq_length, nd_2->seq_length));

  return ret_value;
}
