/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#ifndef _NODE_H
#define _NODE_H


/**************************************************/
/*     A single strand of nulceotide bases        */
/**************************************************/
struct single_strand {
  char* sequence;
  int   number;  // this is the strand number biologically -- not the index!
  int   id;      // this is the index of the strand for processing purposes
  long  ancestor_node_id;  
};

typedef struct single_strand SINGLESTRANDTYPE;
typedef struct single_strand* SINGLESTRAND;


/*********************************************************************/
/*   A homologous chromosome is a grouping of one or more strands    */
/*********************************************************************/
struct homologous_chromosome {
  SINGLESTRAND strands[MAX_NUM_STRANDS_PER_HC];
  int number;  // this is the number biologically -- not the index!
  int id;      // this is the index for processing purposes
};

typedef struct homologous_chromosome HCHROMOSOMETYPE;
typedef struct homologous_chromosome* HCHROMOSOME;



/**********************************************************************/
/*    A node is an element of our network.                            */
/*                                                                    */
/*    However at times it is convenient to think of lineages within   */
/*    our network as well. Both lineages and nodes have the same      */
/*    elements, but it is conceptually easier to think about lineages */
/*    changing and evolving over time and eventually terminating in   */
/*    a node of the network.                                          */
/*                                                                    */
/*    Hence, this single struct has pointers called both NODE and     */
/*    LINEAGE.  The struct is created during a birth-like event       */
/*    (either birth or hybridization).                                */
/*                                                                    */
/*    In general, one thinks of nodes when processing "all the nodes" */
/*    of the network (NodeArray) and one thinks of lineages when      */
/*    working with an active lineage undergoing evolutionary change.  */
/*    (The attribute "active" indicates whether the lineage is        */
/*    active or not.                                                  */
/*                                                                    */
/*    In summary, one can think of it as each lineage ends in a node, */
/*    and the types NODE and LINEAGE can be used interchangeably in   */
/*    the code depending on which terminology makes the most sense.   */
/**********************************************************************/

struct node {
  struct node *left, *right, *parent1, *parent2;
  long id;
  int active;   // 0 false, 1 true (whether a lineage is finished or active)
  int leaf;     // 0 false, 1 true
  int extant;   // 0 false, 1 true (not every leaf is extant!)
  int outgroup; // 0 false, 1 true 
  int hybrid;   // 0: tree node 1-6: newtork node (value indicates hybrid type)
  long hd_from_root;  // Hamming distance (total not avg) between this node
                      // and the root of the network
  int num_hom_chromos;
  int num_strands_per_hom_chrom;;
  int num_unique_hom_chromos; // refers to number of biological chromosomes
                              // which is needed when comparing two
			      // lineages with differing number of 
			      // chromosomes but want to do an average
			      // hamming distance
  int seq_length; // this could be moved down to the chromosome level if
                  // not all chromosomes had the same length but not needed now
  double birth_rate;
  double death_rate;
  double hybrid_rate;
  double time_created;
  double prev_update_time;
  double gamma_value;
  double brlen1_t, brlen2_t;    // branch lengths per time
  double brlen1_ev, brlen2_ev;  // branch lengths per evolution
  double brlen1_hd, brlen2_hd;  // branch hamming distance from parent to node 
  HCHROMOSOME hom_chromos[MAX_NUM_HOM_CHROMOS];
  int marker;             // marker variable for generic use
  double dijkstra_dist;   // for use in DijkstraDist routine
  long dijkstra_pred;     // for use in DijkstraDist routine
};
typedef struct node NODETYPE;
typedef struct node* NODE;
typedef struct node LINEAGETYPE;
typedef struct node* LINEAGE;

// the "constructors"
SINGLESTRAND  NewSingleStrand(int strand_id, int sl);
HCHROMOSOME NewHomChrom(int hc_id, int pl, int sl);
LINEAGE    NewLineage(long lin_id, int nhc, int nsphc, int sl, NODE parent_ptr);

/**************************************************************************/
/*                      utility functions for nodes                       */
/**************************************************************************/

// NOTE:  We currently have two ways of calculating/defining the hamming
//        distance of two nodes.  
//
//        (1) HammingDistanceNodes interprets differing number
//            of sequences (num_hom_chromos * num_strands_per_hom_chrom) 
//            as contributing to the hamming distance.  Hamming
//            distance across all string pairs are summed.
//
//        (2) HammingDistanceAvgNodes requires the number of unique
//            homologous chromosomes to the be same (number of strands
//            per homologous chromosome currently cannot differ as 
//            this value is specifed at the node level).  
//
//            If numbers of unique homologous chromosomes are the same 
//            then an "average across sequences can be done" -- cycling 
//            through the sequences of the node with the least 
//            amount of sequences multiple times as the 
//            hom_chrom/strand will match up.
//
//        Both routines use the HammingDistanceStrings function located
//        in the utilities code -- that code interprets strings of 
//        differing lengths to be contributing to the hamming distance.
//
//        However, it is not expected that this will occur currently as
//        we define sequence length at the node level and all strands
//        are assumed to be the same length.
//
long HammingDistanceNodes(NODE nd_1, NODE nd_2);    
long HammingDistanceAvgNodes(NODE nd_1, NODE nd_2); 

// Here are the corresponding two approaches for defining max hamming
// distance -- either assumed to be 
//      num_hom_chromos * num_strands_per_hom_chrom * seq_length
// OR just seq_length.
//
long MaxHammingDistanceNodes(NODE nd_1, NODE nd_2);
long MaxHammingDistanceAvgNodes(NODE nd_1, NODE nd_2);

#endif


