/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include"globals.h"
#include"node.h"
#include"event.h"
#include"network.h"
#include"utilities.h"
#include"output.h"


/******************************************************************************/
/*                              NetworkReport                                 */
/******************************************************************************/
void NetworkReport(NETWORK nt) 
{

  /*******************************************/
  /*  print the standard run parameter stuff */
  /*******************************************/
  PrintRunParams(nt);


  /***************************************/
  /*    print the rate parameter stuff   */
  /***************************************/
  fprintf(fp_out, "c The rates that were used to generate this network:\n");

  //fprintf(fp_out, "c sp_rate=%lf, ex_rate=%lf, hyb_rate=%lf\n",
  // 	  sp_rate, ex_rate, hyb_rate); 

  fprintf(fp_out, "c sp_rate=%lf, ex_rate=%lf, hyb_rate=%lf, dip_rate=%lf, poly_rate=%lf\n", sp_rate, ex_rate, hyb_rate, dip_rate, poly_rate); 

  if (var_rate_yn)
    fprintf(fp_out, "c sp_rate_var=%lf, ex_rate_var=%lf, hyb_rate_var=%lf\n",
	    sp_rate_var, ex_rate_var, hyb_rate_var); 


  /*******************************************/
  /*    print the hybrid related parameters  */
  /*******************************************/
  PrintHybridParams(); 


  /*******************************************/
  /*      print the height related info      */
  /*******************************************/
  fprintf(fp_out, "c \n");

  if (var_rate_yn)
    fprintf(fp_out, "c The user requested a variable rate network\n");
  else
    fprintf(fp_out, "c The user requested a constant rate network\n");

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The clock height of the network was: %lf\n", nt->clock_height);
  fprintf(fp_out, "c The evolutionary height of the network was: %lf\n", nt->ev_height);
  if (sequences_yn)
  {
    fprintf(fp_out, "c The average Hamming distance height of the network was: %lf\n", nt->avg_hd_height);
  }
  fprintf(fp_out, "c \n");

  if (desired_height_given_yn)
  {
    fprintf(fp_out, "c The desired height of the network was: %lf\n", desired_height);
    fprintf(fp_out, "c The expected height of the network was: %lf\n", expected_height);
  }

  fprintf(fp_out, "c The height scalar, used to convert internal time, was: %lf\n", height_scalar);
  fprintf(fp_out, "c \n");

  /************************************************/
  /*      print the ultrametric related info      */
  /************************************************/
  fprintf(fp_out, "c \n");

  if (ultrametric_yn)
    fprintf(fp_out, "c The user requested an ultrametric network\n");
  else
  {
    fprintf(fp_out, "c The user requested a non-ultrametric network\n");
    fprintf(fp_out, "c The gamma shape value was: %d and the gamma scale value was: %lf\n", gamma_shape_value, gamma_scale_value);
  }

  fprintf(fp_out, "c \n");


  /*******************************************/
  /*    print stuff related to sequences     */
  /*******************************************/
  if (sequences_yn)
  {
    PrintSeqParams();
  }
  else
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Topology only (no sequences) generated in this simulation.\n");
  }


  /***************************************/
  /*     print network summary stuff     */
  /***************************************/
  
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The following is the network summary line:\n");
  
  // if we made an outgroup, there are extra 
  // parameters to be printed here
 
  if (outgroup_done)
  {
    // we know if there is an outgroup it was added last, so use
    // nt->num_nodes - 1 as a proxy for the id of the outgroup
    // (because we number nodes/lineages starting at 0)

    fprintf(fp_out, "c    # nodes, # arcs, # leaves, # hybrids, outgroup node id, ev height,\n");
    fprintf(fp_out, "c    min outgroup diff, max outgroup diff, max outgroup tries\n");

    fprintf(fp_out, "s %ld %ld %ld %ld %ld %lf %d %d %d\n", 
	    nt->num_nodes, nt->num_edges, nt->num_leaves, 
	    nt->num_hybrids, (nt->num_nodes)-1, nt->ev_height, 
	    min_outg_diff_value, max_outg_diff_value, max_outg_tries);
  }
  else
  {
    fprintf(fp_out, "c    # nodes, # arcs, # leaves, # hybrids, outgroup (-1), ev height\n");

    fprintf(fp_out, "s %ld %ld %ld %ld -1 %lf\n", 
	    nt->num_nodes, nt->num_edges, nt->num_leaves, 
	    nt->num_hybrids, nt->ev_height);
  }


  /*********************************/
  /* print nsgw seeds if requested */
  /*********************************/
  if ( (track_nsgw_seeds_yn) && (sequences_yn) )
  {
    PrintNSGWSeedInfo();
  }


  /*********************/
  /*       nodes       */
  /*********************/
  PrintNodes(nt);


  /********************/
  /*       leaves     */
  /********************/
  PrintLeaves(nt);
  

  /********************/
  /*       edges      */
  /**************** ***/
  PrintEdges(nt);

  /***********************************************************/
  /*       parsimony score -- only if sequences though       */
  /***********************************************************/
  if (sequences_yn)
  {
    PrintParsimonyScore(nt);
  }


  /**********************************************/
  /*     do optional printing if requested      */
  /**********************************************/
  if (print_opt_reports)
  {
    PrintOptReports(nt);
  }

}  // end of NetworkReport


/******************************************************************************/
/*                          PrintRunParams                                    */
/******************************************************************************/
void PrintRunParams(NETWORK nt)
{
  fprintf(fp_out, "c The process id for this run was: %d\n", pid);
  fprintf(fp_out, "c \n");

  fprintf(fp_out, "c The input file for this run was: %s\n", infile_name);
  fprintf(fp_out, "c \n");

  fprintf(fp_out, "c This was a run for %d extant taxa\n", num_extant_taxa_req);

  if (nt->num_extant_taxa < num_extant_taxa_req)
  {
    fprintf(fp_out, "c The simulation ended prematurely due to all lineages dying out\n");
    if (outgroup)
    {
      fprintf(fp_out, "c (hence the requested outgroup was not made)\n");
    }
    fprintf(fp_out, "c \n");
  }

  fprintf(fp_out, "c \n");
  if (user_seed_netgen_yn == 1)
  {
    fprintf(fp_out, "c The user specified %lu as the random number seed for this program\n", user_seed_netgen);
  }

  fprintf(fp_out, "c The random number seed used by this run was: %lu\n", seed_netgen);
  fprintf(fp_out, "c \n");

  fprintf(fp_out, "c \n");
  if (update_seqs_for_all_events_yn)
  {
    fprintf(fp_out, "c The user requested that all branch lengths and sequences be updated prior to every event\n");
    fprintf(fp_out, "c \n");
  }

  fprintf(fp_out, "c \n");
}


/******************************************************************************/
/*                          PrintHybridParams                                 */
/******************************************************************************/
void PrintHybridParams() 
{
  fprintf(fp_out, "c \n");
  if (max_num_hybrids != -1)
  {
    fprintf(fp_out, "c The max number of hybrids requested by user was: %d\n", 
	    max_num_hybrids);
  }
  else
  {
    fprintf(fp_out, "c The max number of hybrids was not limited by the user \n");
  }

  if (rates_redone_yn == 1)
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Some of the above rates were reset as the max number of hybrids was reached.\n");
    fprintf(fp_out, "c The original rates (used before reaching max_num_hybrids) were:\n");
    //fprintf(fp_out, "c hyb_rate_orig=%lf\n", hyb_rate_orig); 
    
    fprintf(fp_out, "c hyb_rate_orig=%lf dip_rate_orig=%lf poly_rate=%lf\n", hyb_rate_orig, dip_rate_orig, poly_rate_orig); 

    if (var_rate_yn == 1)
    {
      fprintf(fp_out, "c hyb_rate_var_orig=%lf\n", hyb_rate_var_orig); 
    }
    fprintf(fp_out, "c \n");
  }

  // print hybrid threshold  -- depends on type of hybrid parent selection
  if (hyb_parent_selection == 0)
  {
    fprintf(fp_out, "c hyb_ham_threshold_rate = %lf\n", hyb_ham_thresh_rate);
  }

  if (hyb_parent_selection == 4)
  {
    fprintf(fp_out, "c hyb_ev_dist_threshold = %lf\n", hyb_ev_dist_thresh);
  }

  if (hyb_parent_selection == 5)
  {
    fprintf(fp_out, "c hyb_exp_fn_thresh_rate = %lf\n", hyb_exp_fn_thresh_rate);
  }

  fprintf(fp_out, "c \n");

  // print the method for choosing the second parent of hybrids
  if (hyb_parent_selection == 0)
  {
    fprintf(fp_out, "c Second parent for hybrids based on min hamming distance\n");
  }
  else if (hyb_parent_selection == 1)
  {
    fprintf(fp_out, "c Second parent for hybrids chosen randomly\n");
  }
  else if (hyb_parent_selection == 2)
  {
    fprintf(fp_out, "c Second parent for hybrids based on min branch length\n");
  }
  else if (hyb_parent_selection == 3)
  {
    fprintf(fp_out, "c Second parent for hybrids based on max branch length\n");
  }
  else if (hyb_parent_selection == 4)
  {
    fprintf(fp_out, "c Second parent for hybrids based on min evolutionary distance\n");
  }
  else if (hyb_parent_selection == 5)
  {
    fprintf(fp_out, "c Second parent for hybrids based on exponential function\n");
    fprintf(fp_out, "c with %ld as the hamming distance having 1/e probability of\n", lambda_hd);
    fprintf(fp_out, "c being the distance for the second parent from the first\n");
  }
  else 
  {
    fprintf(fp_out, "c NETGEN ERROR: invalid parent selection for hybrids\n");
    exit(0);
  }

}


/******************************************************************************/
/*                           PrintSeqParams                                   */
/******************************************************************************/
void PrintSeqParams()
{
  int hchromo_i, strand_i;
  NODE root_node;

  fprintf(fp_out, "c Sequences were generated at each node as the topology progressed.\n");

  if (initial_root_sequences_yn == 0)
  {
    fprintf(fp_out, "c The root sequences were not specified -- NSGW generated them randomly  \n");
  }
  else
  {
    fprintf(fp_out, "c The root sequences specified were:\n");

    for (hchromo_i = 0; hchromo_i < root_node->num_hom_chromos; hchromo_i++)
    {
      for (strand_i = 0; 
           strand_i < root_node->num_strands_per_hom_chrom; 
	   strand_i++)
      {
        fprintf(fp_out, "c %s\n", root_node->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      }
    }
  }
  if (track_nsgw_seeds_yn)
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c The seeds for NSGW were tracked -- see below\n");
  }
  else
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c The seeds for NSGW were NOT tracked\n");
  }
}


/******************************************************************************/
/*                           PrintNSGWSeedInfo                                */
/******************************************************************************/
void PrintNSGWSeedInfo() 
{
  int seed_i;

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The seeds for NSGW (and related node) were tracked and reported below\n");

  if (nsgw_seeds_index >= MAX_NSGW_SEEDS)
  {
    fprintf(fp_out, "c Only the first %d NSGW seeds were recorded\n", MAX_NSGW_SEEDS);
  }

  for (seed_i = 0; seed_i < nsgw_seeds_index; seed_i++)
  {
    fprintf(fp_out, "c Seed %3d to NSGW was: %14lu  For node: %3d\n", seed_i, nsgw_seeds[seed_i], nsgw_seeds_node[seed_i]);
  }
}


/******************************************************************************/
/*                                PrintNodes                                  */
/******************************************************************************/
void PrintNodes(NETWORK nt)
{
  NODE root_node;
  NODE nd;
  int node_i;

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The following are all the internal nodes of the network\n");
  fprintf(fp_out, "c \n");

  if (sequences_yn)
  {
    fprintf(fp_out, "c For each node we list the:\n");
    fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n"); 
    fprintf(fp_out, "c      sequence length, time created, speciation rate, death rate,\n");
    fprintf(fp_out, "c      hybrid rate, and sequences\n");
  }
  else
  {
    fprintf(fp_out, "c For each node we list the:\n");
    fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n");
    fprintf(fp_out, "c      sequence length, time created, speciation rate, death rate, hybrid rate\n");
  }

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c Types are: r = root, n = internal node\n");
  
  fprintf(fp_out, "c \n");

  /***************************/
  /*      print root         */
  /***************************/
  root_node = nt->root;

  fprintf(fp_out, "r %ld %d %d %d %d %.4f %.4f %.4f %.4f\n", root_node->id, root_node->hybrid, root_node->num_hom_chromos, root_node->num_strands_per_hom_chrom, root_node->seq_length, root_node->time_created, root_node->birth_rate, root_node->death_rate, root_node->hybrid_rate); 

  if (sequences_yn)
    PrintNodeSequences(root_node);


  /***************************************************************/
  /*  print internal nodes (but we have already done the root)   */
  /***************************************************************/
  for (node_i = 1; node_i < nt->num_nodes; node_i++)
  {
    if (!(nt->NodeArray[node_i]->leaf))
    {
      nd = nt->NodeArray[node_i];

      fprintf(fp_out, "n %ld %d %d %d %d %.4f %.4f %.4f %.4f\n", nd->id, nd->hybrid, nd->num_hom_chromos, nd->num_strands_per_hom_chrom, nd->seq_length, nd->time_created, nd->birth_rate, nd->death_rate, nd->hybrid_rate); 

      if (sequences_yn)
        PrintNodeSequences(nd);
    }
  }

}
  

/******************************************************************************/
/*                                PrintLeaves                                 */
/******************************************************************************/
void PrintLeaves(NETWORK nt)
{
  int node_i;
  long end;
  NODE nd;

  /**********************************/
  /*       non-extant header       */
  /**********************************/
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The following are all the non-extant leaves of the network\n");
  fprintf(fp_out, "c \n");

  if (sequences_yn)
  {
    fprintf(fp_out, "c For each node we list the:\n");
    fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n"); 
    fprintf(fp_out, "c      sequence length, time created, and sequences\n");
  }
  else
  {
    fprintf(fp_out, "c For each node we list the:\n");
    fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n");
    fprintf(fp_out, "c      sequence length, time created\n");
  }

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c Types are:  l = non-extant leaf\n");
  
  fprintf(fp_out, "c \n");


  /***************************/
  /* print non-extant leaves */
  /***************************/

  // we don't want to do the outgroup leaf -- we print it out separately
  // and it is not labeled as extant, so it would qualify here, so don't
  // check the last node if we made an outgroup!
  
  if (outgroup_done)
    end = (nt->num_nodes) - 1;
  else
    end = nt->num_nodes;

  for (node_i = 0; node_i < end; node_i++)
  {
    if (!(nt->NodeArray[node_i]->extant) && (nt->NodeArray[node_i]->leaf))
    {
      nd = nt->NodeArray[node_i];

      fprintf(fp_out, "l %ld %d %d %d %d %.4f\n", nd->id, nd->hybrid, nd->num_hom_chromos, nd->num_strands_per_hom_chrom, nd->seq_length, nd->time_created); 

      if (sequences_yn)
        PrintNodeSequences(nd);
    }
  }

  /**********************************/
  /*         extant header          */
  /**********************************/
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The following are all the extant leaves of the network\n");
  fprintf(fp_out, "c \n");

  if (sequences_yn)
  {
    fprintf(fp_out, "c For each node we list the:\n");
    fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n"); 
    fprintf(fp_out, "c      sequence length, time created, and sequences\n");
  }
  else
  {
    fprintf(fp_out, "c For each node we list the:\n");
    fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n");
    fprintf(fp_out, "c      sequence length, time created\n");
  }

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c Types are:  e = extant leaf\n");
  
  fprintf(fp_out, "c \n");


  /***************************/
  /*   print extant leaves   */
  /***************************/
  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    if ((nt->NodeArray[node_i]->extant) && (nt->NodeArray[node_i]->leaf))
    {
      nd = nt->NodeArray[node_i];

      fprintf(fp_out, "e %ld %d %d %d %d %.4f\n", nd->id, nd->hybrid, nd->num_hom_chromos, nd->num_strands_per_hom_chrom, nd->seq_length, nd->time_created); 

      if (sequences_yn)
        PrintNodeSequences(nd);
    }
  }

  /************************************/
  /*         outgroup header          */
  /************************************/
  if (outgroup_done)
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c The following is the outgroup of the network\n");
    fprintf(fp_out, "c \n");

    if (sequences_yn)
    {
      fprintf(fp_out, "c For each node we list the:\n");
      fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n"); 
      fprintf(fp_out, "c      sequence length, time created, and sequences\n");
    }
    else
    {
      fprintf(fp_out, "c For each node we list the:\n");
      fprintf(fp_out, "c      type, id, hybrid status, #hchromosomes, #strands,\n");
      fprintf(fp_out, "c      sequence length, time created\n");
    }

    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Types are:  o = outgroup leaf\n");
  
    fprintf(fp_out, "c \n");

    /***************************/
    /*   print outgroup leaf   */
    /***************************/
    nd = nt->NodeArray[(nt->num_nodes)-1];

    fprintf(fp_out, "o %ld %d %d %d %d %.4f\n", nd->id, nd->hybrid, nd->num_hom_chromos, nd->num_strands_per_hom_chrom, nd->seq_length, nd->time_created); 

    if (sequences_yn)
      PrintNodeSequences(nd);
  }

}


/******************************************************************************/
/*                          PrintNodeSequences                                */
/******************************************************************************/
void PrintNodeSequences(NODE nd)
{
  int hchromo_i, strand_i;

  int nhc;
  int nsphc;

  nhc = nd->num_hom_chromos;
  nsphc = nd->num_strands_per_hom_chrom;

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {
      fprintf(fp_out, "%s\n", nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    }

  }
}


/******************************************************************************/
/*                               PrintEdges                                   */
/******************************************************************************/
void PrintEdges(NETWORK nt)
{
  NODE p_node;
  NODE c_node;
  int node_i;

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The following are all the edges of the network\n");
  fprintf(fp_out, "c \n");

  fprintf(fp_out, "c For each edge we list the:\n");
  fprintf(fp_out, "c      type, parent, child, evolutionary branch length, clock branch length\n");

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c Types are:\n");
  fprintf(fp_out, "c     t = \"tip\" (edge from internal node to extant taxon),\n");
  fprintf(fp_out, "c     b = all other edges (between two internal nodes, or internal node and\n");
  fprintf(fp_out, "c         non-extant leaf)\n");
  fprintf(fp_out, "c \n");

  /***************************/
  /*       print edges       */
  /***************************/
  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    // print the edges associated with this node
    p_node = nt->NodeArray[node_i];

    // first print the edge from it to the left child
    c_node = nt->NodeArray[node_i]->left;

    if (c_node != NULL)
    {
      if (c_node->extant)
        fprintf(fp_out, "t ");
      else
        fprintf(fp_out, "b ");

      if (c_node->parent1 == p_node)
      {
        fprintf(fp_out, "%ld %ld %lf %lf\n", p_node->id, c_node->id, c_node->brlen1_ev, c_node->brlen1_t);
      }
      else
      {
        fprintf(fp_out, "%ld %ld %lf %lf\n", p_node->id, c_node->id, c_node->brlen2_ev, c_node->brlen2_t);
      }
    }
    
    // now print the edge from it to the right child
    c_node = nt->NodeArray[node_i]->right;

    if (c_node != NULL)
    {
      if (c_node->extant)
        fprintf(fp_out, "t ");
      else
        fprintf(fp_out, "b ");

      if (c_node->parent1 == p_node)
      {
        fprintf(fp_out, "%ld %ld %lf %lf\n", p_node->id, c_node->id, c_node->brlen1_ev, c_node->brlen1_t);
      }
      else
      {
        fprintf(fp_out, "%ld %ld %lf %lf\n", p_node->id, c_node->id, c_node->brlen2_ev, c_node->brlen2_t);
      }

    }
  }
}

/******************************************************************************/
/*                        PrintParsimonyScore                                 */
/******************************************************************************/
void PrintParsimonyScore(NETWORK nt)
{
  fprintf(fp_out, "c\n");
  fprintf(fp_out, "c The parsimony score for this network is %ld\n", nt->parsimony_score);
  fprintf(fp_out, "c\n");
}


/******************************************************************************/
/*                            PrintOptReports                                 */
/******************************************************************************/
void PrintOptReports(NETWORK nt)
{
  // these are extra reports if user requested
  //
  //    * each is requested individually in the input file by:
  //
  //          * hamming_edge_report (warning: LOTS of output -- rarely used!)
  //          * modified_newick_report
  //          * node_listing_report
  //          * hd_from_root_report
  //          * splitstree_report
  //
  //    * each line starts with a "c" as they are officially comment lines
  //
 
  NODE root_node = nt->root;

  /*************************************************************************/
  /***************************** Hamming Edge Report ***********************/
  /*************************************************************************/
  if (hamming_edge_report)
  {
    // print the arcs again and their hamming info if there were seqs
    // too much output, unless you really want it!!!!
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Hamming Edge Report\n");

    if (sequences_yn)
    {
      fprintf(fp_out, "c The following are all the edges and their hamming info\n");
      ResetMarkers(root_node);
      PrintHammingEdges(root_node); 
      ResetMarkers(root_node);
    }
    else
    {
      fprintf(fp_out, "c Hamming edge report was requested but no sequences to report\n");
    }
  }

  /*************************************************************************/
  /************************** (Modified) Newick Report *********************/
  /*************************************************************************/

  if (modified_newick_report)
  {
    fprintf(fp_out, "c \n");

    // if tree, print regular newick format
    if (IsTree(nt))
    {
      fprintf(fp_out, "c Newick Report -- clock / time branch lengths\n");
      ResetMarkers(root_node);
      fprintf(fp_out,"c");       // comment line start
      PrintNewick(root_node, 0); // for time/clock branch lengths
      fprintf(fp_out,");\n");    // termination for newick format
      ResetMarkers(root_node);

      fprintf(fp_out, "c \n");
      fprintf(fp_out, "c Newick Report -- evolutionary branch lengths\n");

      ResetMarkers(root_node);
      fprintf(fp_out,"c");       // comment line start
      PrintNewick(root_node, 1); // for evolutionary branch lengths
      fprintf(fp_out,");\n");    // termination for newick format
      ResetMarkers(root_node);

      fprintf(fp_out, "c \n");
      fprintf(fp_out, "c Newick Report -- Hamming distance branch lengths\n");

      ResetMarkers(root_node);
      fprintf(fp_out,"c");       // comment line start
      PrintNewick(root_node, 2); // for hamming distance branch lengths
      fprintf(fp_out,");\n");    // termination for newick format
      ResetMarkers(root_node);
    }


    // if network, print our customized newick network format
    if (!IsTree(nt))
    {
      fprintf(fp_out, "c (Modified) Newick Report -- clock / time branch lengths\n");
      ResetMarkers(root_node);
      fprintf(fp_out,"c");              // comment line start
      PrintNewickNetwork(root_node, 0); // for time/clock branch lengths
      fprintf(fp_out,");\n");           // termination for newick format
      ResetMarkers(root_node);

      fprintf(fp_out, "c \n");
      fprintf(fp_out, "c (Modified) Newick Report -- evolutionary branch lengths\n");

      ResetMarkers(root_node);
      fprintf(fp_out,"c");              // comment line start
      PrintNewickNetwork(root_node, 1); // for evolutionary branch lengths
      fprintf(fp_out,");\n");           // termination for newick format
      ResetMarkers(root_node);

      fprintf(fp_out, "c \n");
      fprintf(fp_out, "c (Modified) Newick Report -- Hamming distance branch lengths\n");

      ResetMarkers(root_node);
      fprintf(fp_out,"c");              // comment line start
      PrintNewickNetwork(root_node, 2); // for hamming distance branch lengths
      fprintf(fp_out,");\n");           // termination for newick format
      ResetMarkers(root_node);
    }
  }

  /*************************************************************************/
  /*************************** Node Listing Report *************************/
  /*************************************************************************/

  if (node_listing_report)
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Node Listing Report\n");
    PrintAllNodes(nt);
  }

  /*************************************************************************/
  /*************************** HD from Root Report *************************/
  /*************************************************************************/

  // FIX LATER
  //if ( (hhd_from_root_report) && (nt->num_hybrids > 0) )
  if ( (hhd_from_root_report) )
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Hybrid Hamming Distance from Root Report\n");
    fprintf(fp_out, "c (provides id and Hamming distance between it and root)\n");
    PrintHHDFromRoot(nt);
  }

  /*************************************************************************/
  /********************** Detailed Edge Listing Report *********************/
  /*************************************************************************/

  if (report_branch_hd)
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c Detailed Edge Listing Report\n");
    PrintDetailedEdgeListing(nt);
  }
  
  /*************************************************************************/
  /*************************** SplitsTree Report ***************************/
  /*************************************************************************/
  if (splitstree_report)
  {
    fprintf(fp_out, "c \n");
    fprintf(fp_out, "c SplitsTree Report\n");
    PrintSplitsTreeFormat(nt);
  }
}


/******************************************************************************/
/*                 Print Special Hamming Report on the Edges                  */
/******************************************************************************/
void PrintHammingEdges(NODE nd)
{
  // left subtree
  if (nd->marker==0 &&nd->left!=NULL) 
  {
    if (nd->left->parent1 == nd) 
    {
      fprintf(fp_out, "c a %ld %ld %lf %lf\n", 
	      nd->id, nd->left->id, nd->left->brlen1_ev, nd->left->brlen1_t);

      HammingReport(nd, 0);
    } 
    else 
    {
      fprintf(fp_out, "c a %ld %ld %lf %lf\n", 
	      nd->id, nd->left->id, nd->left->brlen2_ev, nd->left->brlen2_t);

      HammingReport(nd, 0);
    }

    PrintHammingEdges(nd->left);
  }

  // right subtree
  if (nd->marker==0 && nd->right!=NULL) 
  {
    if (nd->right->parent1 == nd) 
    {
      fprintf(fp_out, "c a %ld %ld %lf %lf\n", 
	      nd->id, nd->right->id, nd->right->brlen1_ev, nd->right->brlen1_t);

      HammingReport(nd, 1);
    } 
    else 
    {
      fprintf(fp_out, "c a %ld %ld %lf %lf\n", 
	      nd->id, nd->right->id, nd->right->brlen2_ev, nd->right->brlen2_t);

      HammingReport(nd, 1);
    };

    PrintHammingEdges(nd->right);
  }

  // recurse
  if (nd->marker==1 && nd->left!=NULL)
	PrintHammingEdges(nd->left);

  if (nd->marker==1 && nd->right!=NULL)
	PrintHammingEdges(nd->right);

  // mark node as done
  nd->marker = 1;
}


void HammingReport(LINEAGE nd, int direction)
{
  int hchromo_i, strand_i;

  if (direction == 0)
  {
    // print hamming distance with left child
    for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
    {
      for (strand_i = 0; 
           strand_i < nd->num_strands_per_hom_chrom; 
	   strand_i++)
      {
        PrintHamming(hchromo_i, strand_i,  
		      nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		      nd->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		      1);
      }
    }

    // if grandchildren through the left child exist, print
    // the hamming distances between the grandparent (our current node)
    // and the grandchild; also see if the grandkids
    // have kids -- if so print the hamming distance between the 
    // great grandparent (our current node) and the great grandkids
    if (nd->left->left != NULL)
    {
      // left->left grandchild (#A)
      for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; 
	     strand_i < nd->num_strands_per_hom_chrom; 
	     strand_i++)
        {
          PrintHamming(hchromo_i, strand_i,  
		       nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		       nd->left->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		       2);
        }
      }

      // left->left->left great grand child (#1)
      if (nd->left->left->left != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
		         nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		          nd->left->left->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		          3);
          }
        }
      }

      // left->left->right great grand child (#2)
      if (nd->left->left->right != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
		          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		          nd->left->left->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		          3);
          }
        }
      }
    }

    // now do the same for the left->right descendants (grandkid #B)
    if (nd->left->right != NULL)
    {
      for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; 
	     strand_i < nd->num_strands_per_hom_chrom; 
	     strand_i++)
        {
          PrintHamming(hchromo_i, strand_i,  
                        nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                nd->left->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                2);
        }
      }

      // left->right->left great grand child (#3)
      if (nd->left->right->left != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
                          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                  nd->left->right->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                  3);
          }
        }
      }

      // left->right->right great grand child (#4)
      if (nd->left->right->right != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
                          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                  nd->left->right->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                  3);
          }
        }
      }
    }
  } // end of left child processing
  else
  {
    // print hamming distance with right child
    for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
    {
      for (strand_i = 0; 
           strand_i < nd->num_strands_per_hom_chrom; 
	   strand_i++)
      {
        PrintHamming(hchromo_i, strand_i,  
                      nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                      nd->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	              1);
      }
    }

    // go through and do grandkid and great grand kid hamming distances
    // with current node (the ancestor) for the right kid's side
    // just like we did for the left stuff above

    if (nd->right->left != NULL)
    {
      // grandkid #C
      for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; 
	     strand_i < nd->num_strands_per_hom_chrom; 
	     strand_i++)
        {
          PrintHamming(hchromo_i, strand_i,  
                        nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                        nd->right->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                2);
        }
      }
      // great grandkid #5
      if (nd->right->left->left != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
                          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                          nd->right->left->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                  3);
          }
        }
      }
      // great grandkid #6
      if (nd->right->left->right != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
                          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                          nd->right->left->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	                  3);
          }
        }
      }
    }

    // grandkid #D
    if (nd->right->right != NULL)
    {
      for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; 
	     strand_i < nd->num_strands_per_hom_chrom; 
	     strand_i++)
        {
          PrintHamming(hchromo_i, strand_i,  
                        nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                        nd->right->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                        2);
        }
      }
      // great grandkid #7
      if (nd->right->right->left != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
                          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                          nd->right->right->left->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                          3);
          }
        }
      }
      // great grandkid #8
      if (nd->right->right->right != NULL)
      {
        for (hchromo_i = 0; hchromo_i < nd->num_hom_chromos; hchromo_i++)
        {
          for (strand_i = 0; 
	       strand_i < nd->num_strands_per_hom_chrom; 
	       strand_i++)
          {
            PrintHamming(hchromo_i, strand_i,  
                          nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                          nd->right->right->right->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
                          3);
          }
        }
      }
    } // end of right grandkid processing
  } // end of right child processing
}

void PrintHamming(int hchromo_i, int strand_i, char * seqA, 
		   char * seqB, int num_levels)
{
  fprintf(fp_out, "c Level %d HofHChromo%dStrand%d is %d (between %s and %s)\n", num_levels, hchromo_i, strand_i, HammingDistanceStrings(seqA, seqB), seqA, seqB);

}
/******************************************************************************/



/******************************************************************************/
/*                           PrintNewick                                      */
/******************************************************************************/
void PrintNewick(LINEAGE treenode, int bl_type)
{
  /* this routine follows a postfix approach */
  /* the last characters ); must be put in by the calling routine */

  if (treenode != NULL)
  {
    /* left subtree */
    if (treenode->left != NULL)
    {
      fprintf(fp_out, "(");
      PrintNewick(treenode->left, bl_type);
    }
    /* left leaf node */
    else if ( (treenode->leaf == 1) && (treenode->parent1->left == treenode))
    {
      if (bl_type == 0)
        fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_t);
      else if (bl_type == 1) 
        fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_ev);
      else 
        fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_hd);
    }

    /* right subtree */
    if (treenode->right != NULL)
    {
      fprintf(fp_out, ",");
      PrintNewick(treenode->right, bl_type);
    }
    /* right leaf node */
    else if ( (treenode->leaf == 1) && (treenode->parent1->right == treenode))
    {
      if (bl_type == 0)
        fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_t);
      else if (bl_type == 1)
        fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_ev);
      else
        fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_hd);
    }

    /* root of (sub)tree */
    if ((treenode->left != NULL) && (treenode->right != NULL) &&
	(treenode->parent1 != NULL))
    {
      if (bl_type == 0)
        fprintf(fp_out,")%ld:%f", treenode->id, treenode->brlen1_t);
      else if (bl_type == 1)
        fprintf(fp_out,")%ld:%f", treenode->id, treenode->brlen1_ev);
      else
        fprintf(fp_out,")%ld:%f", treenode->id, treenode->brlen1_hd);
    }
  }
}

/******************************************************************************/
/*                           PrintNewickNetwork                               */
/******************************************************************************/
void PrintNewickNetwork(LINEAGE treenode, int bl_type)
{
  /* this routine follows a postfix approach */
  /* the last ); must be put in by the calling routine */

  if (treenode != NULL)
  {
    /**********************************/
    /*         left subtree           */
    /**********************************/
    if ( (treenode->left != NULL) && (treenode->marker == 0) )
    {
      fprintf(fp_out, "(");
      PrintNewickNetwork(treenode->left, bl_type);
    }
    // if we have a true left leaf node, or a hybrid node acting 
    // as a left leaf because it has been printed once already, 
    // we want to print out the "leaf"
    else if ( ((treenode->leaf == 1) && (treenode->parent1->left == treenode))||
	      ((treenode->hybrid != 0) && (treenode->marker == 1) &&
	         ((treenode->parent1->left == treenode) ||
		  (treenode->parent2->left == treenode))) )
    {
      if (treenode->hybrid != 0)
      {
	if (bl_type == 0)
          fprintf(fp_out, "%ld#H:%f", treenode->id, treenode->brlen1_t);
	else if (bl_type == 1)
          fprintf(fp_out, "%ld#H:%f", treenode->id, treenode->brlen1_ev);
	else
          fprintf(fp_out, "%ld#H:%f", treenode->id, treenode->brlen1_hd);
      }
      else
      {
	// for a true leaf, label it extant (#E) if it is or if it is
	// the outgroup leaf, along with the correct type of branch length
	
	if (bl_type == 0)
	{
	  if (treenode->extant == 1 || treenode->outgroup == 1)
            fprintf(fp_out, "%ld#E:%f", treenode->id, treenode->brlen1_t);
	  else
            fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_t);
	}
	else if (bl_type == 1) 
	{
	  if (treenode->extant == 1 || treenode->outgroup == 1)
            fprintf(fp_out, "%ld#E:%f", treenode->id, treenode->brlen1_ev);
	  else
            fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_ev);
	}
	else 
	{
	  if (treenode->extant == 1 || treenode->outgroup == 1)
            fprintf(fp_out, "%ld#E:%f", treenode->id, treenode->brlen1_hd);
	  else
            fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_hd);
	}
      }

      // mark this leaf/node as printed
      // (this marker goes up to 2 hybrid nodes get printed twice)
      treenode->marker = treenode->marker + 1;
    }


    /**********************************/
    /*         right subtree          */
    /**********************************/
    if ( (treenode->right != NULL) && (treenode->marker == 0) )
    {
      fprintf(fp_out, ",");
      PrintNewickNetwork(treenode->right, bl_type);
    }
    // if we have a true right leaf node, or a hybrid node acting 
    // as a right leaf because it has been printed once already, 
    // we want to print out the "leaf"
    else if ( ((treenode->leaf == 1) &&(treenode->parent1->right == treenode))||
	      ((treenode->hybrid != 0) && (treenode->marker == 1) &&
	         ((treenode->parent1->right == treenode) ||
		  (treenode->parent2->right == treenode))) )
    {
      if (treenode->hybrid != 0)
      {
	if (bl_type == 0)
          fprintf(fp_out, "%ld#H:%f", treenode->id, treenode->brlen1_t);
	else if (bl_type == 1)
          fprintf(fp_out, "%ld#H:%f", treenode->id, treenode->brlen1_ev);
	else
          fprintf(fp_out, "%ld#H:%f", treenode->id, treenode->brlen1_hd);
      }
      else
      {
	// for a true leaf, label it extant (#E) if it is or if it is
	// the outgroup leaf, along with the correct type of branch length
	
	if (bl_type == 0)
        {
	  if (treenode->extant == 1 || treenode->outgroup == 1)
            fprintf(fp_out, "%ld#E:%f", treenode->id, treenode->brlen1_t);
	  else
            fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_t);
	}
	else if (bl_type == 1)
        {
	  if (treenode->extant == 1 || treenode->outgroup == 1)
            fprintf(fp_out, "%ld#E:%f", treenode->id, treenode->brlen1_ev);
	  else
            fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_ev);
	}
	else
        {
	  if (treenode->extant == 1 || treenode->outgroup == 1)
            fprintf(fp_out, "%ld#E:%f", treenode->id, treenode->brlen1_hd);
	  else
            fprintf(fp_out, "%ld:%f", treenode->id, treenode->brlen1_hd);
	}
      }

      // mark this leaf/node as printed
      // (this marker goes up to 2 hybrid nodes get printed twice)
      treenode->marker = treenode->marker + 1;
    }


    /**********************************/
    /*             root               */
    /**********************************/
    /* root of (sub)tree */
    if ( (treenode->leaf != 1) && (treenode->parent1 != NULL) &&
	 (treenode->marker == 0) ) 
    {
      if (treenode->hybrid != 0)
      {
	if (bl_type == 0)
          fprintf(fp_out,")%ld#H:%f",treenode->id, treenode->brlen1_t);
	else if (bl_type == 1)
          fprintf(fp_out,")%ld#H:%f",treenode->id, treenode->brlen1_ev);
	else
          fprintf(fp_out,")%ld#H:%f",treenode->id, treenode->brlen1_hd);
      }
      else
      {
	if (bl_type == 0)
          fprintf(fp_out,")%ld:%f",treenode->id, treenode->brlen1_t);
	else if (bl_type == 1)
          fprintf(fp_out,")%ld:%f",treenode->id, treenode->brlen1_ev);
	else
          fprintf(fp_out,")%ld:%f",treenode->id, treenode->brlen1_hd);
      }

      // mark this leaf/node as printed
      // (this marker goes up to 2 as hybrid nodes get printed twice)
      treenode->marker = treenode->marker + 1;
    }
  }
}

/******************************************************************************/
/*                            PrintSplitsTreeFormat                           */
/******************************************************************************/
void PrintSplitsTreeFormat(NETWORK nt)
{
  long node_i, counter;

  // print out the required header stuff
  fprintf(fp_out, "c #nexus\n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c BEGIN Taxa;\n");
  fprintf(fp_out, "c DIMENSIONS ntax=0;\n");
  fprintf(fp_out, "c END; [Taxa]\n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c BEGIN Network;\n");
  fprintf(fp_out, "c DIMENSIONS ntax=0 nvertices=%ld nedges=%ld;\n", 
	  nt->num_nodes, nt->num_edges);

  // print out the vertice stuff -- note that SplitsTree does not
  // like the first node to be zero, so inflate all ids by 1!
  fprintf(fp_out, "c VERTICES\n");
  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    fprintf(fp_out, "c %ld 0 0,\n", (nt->NodeArray[node_i]->id) + 1);
  }
  fprintf(fp_out, "c ;\n");
  fprintf(fp_out, "c recompute runs=100;\n");
  fprintf(fp_out, "c VLABELS\n");
  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    fprintf(fp_out, "c %ld 'v%ld',\n", 
	    (nt->NodeArray[node_i]->id) + 1,
	    (nt->NodeArray[node_i]->id) + 1);
  }
  fprintf(fp_out, "c ;\n");

  // print out the edge stuff -- and once again we must add 1 to all
  // ids because SplitsTree does not cope with node id of 0!
  fprintf(fp_out, "c EDGES\n");
  counter = 1;

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->parent1 != NULL)
    {
      fprintf(fp_out, "c %ld %ld %ld,\n",
	      counter++,
	      (nt->NodeArray[node_i]->parent1->id) + 1,
	      (nt->NodeArray[node_i]->id) + 1);
    }

    if (nt->NodeArray[node_i]->parent2 != NULL)
    {
      fprintf(fp_out, "c %ld %ld %ld,\n",
	      counter++,
     	      (nt->NodeArray[node_i]->parent2->id) + 1,
	      (nt->NodeArray[node_i]->id) + 1);
    }
  }
  fprintf(fp_out, "c ;\n");

  // print out ending syntax
  fprintf(fp_out, "c END; [Network]\n");
    

}

/******************************************************************************/
/*                                PrintAllNodes                             */
/******************************************************************************/
void PrintAllNodes(NETWORK nt)
{
  int node_i, hchromo_i, strand_i;
  int nhc, nsphc;

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    fprintf(fp_out, "c n %ld\n", 
            nt->NodeArray[node_i]->id);
    
    nhc = nt->NodeArray[node_i]->num_hom_chromos;
    nsphc = nt->NodeArray[node_i]->num_strands_per_hom_chrom;

    if (sequences_yn)
    {
      for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
      {
        for (strand_i = 0; strand_i < nsphc; strand_i++)
        {
          fprintf(fp_out, "c %s\n",
		  nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
        }
      }
    }

  }
}

/******************************************************************************/
/*             PrintHHDFromRoot (Hybrid Hamming Distance From Root)           */
/******************************************************************************/
void PrintHHDFromRoot(NETWORK nt)
{
  int node_i;
  NODE nd;

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    nd = nt->NodeArray[node_i];

    if (nd->hybrid != 0)
    {
      fprintf(fp_out, "c %ld %ld\n", nd->id, nd->hd_from_root);
    }
  }

  return;
}


/******************************************************************************/
/*                        PrintDetailedEdgeListing                            */
/******************************************************************************/
void PrintDetailedEdgeListing(NETWORK nt)
{
  // this routine is very similar to the edge report, but wanted something
  // separate as to not muck up that routine any more with more details which
  // may or may not need to be reported.....and to keep this report simple

  NODE p_node;
  NODE c_node;
  int node_i;

  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c \n");
  fprintf(fp_out, "c The following are DETAILS about all the edges\n");
  fprintf(fp_out, "c   This report was generated due to report_branch_hd being invoked\n");
  fprintf(fp_out, "c \n");

  fprintf(fp_out, "c For each edge we list the:\n");
  fprintf(fp_out, "c      parent, child, evolutionary branch length, clock branch length and hamming distance from parent to child\n");

  fprintf(fp_out, "c \n");

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    // print the edges associated with this node
    p_node = nt->NodeArray[node_i];

    // first print the edge from it to the left child
    c_node = nt->NodeArray[node_i]->left;

    if (c_node != NULL)
    {
      if (c_node->parent1 == p_node)
      {
        fprintf(fp_out, "%ld %ld %lf %lf %lf\n", p_node->id, c_node->id, c_node->brlen1_ev, c_node->brlen1_t, c_node->brlen1_hd);
      }
      else
      {
        fprintf(fp_out, "%ld %ld %lf %lf %lf\n", p_node->id, c_node->id, c_node->brlen2_ev, c_node->brlen2_t, c_node->brlen2_hd);
      }
    }
    
    // now print the edge from it to the right child
    c_node = nt->NodeArray[node_i]->right;

    if (c_node != NULL)
    {
      if (c_node->parent1 == p_node)
      {
        fprintf(fp_out, "%ld %ld %lf %lf %lf\n", p_node->id, c_node->id, c_node->brlen1_ev, c_node->brlen1_t, c_node->brlen1_hd);
      }
      else
      {
        fprintf(fp_out, "%ld %ld %lf %lf %lf\n", p_node->id, c_node->id, c_node->brlen2_ev, c_node->brlen2_t, c_node->brlen2_hd);
      }

    }
  }
}

/******************************************************************************/
/*                            PrintFutureEvents                               */
/******************************************************************************/
void PrintFutureEvents(FILE *fp, EVENT queue)
{
  EVENT temp_ev_ptr;

  if (queue != NULL)
  { 
    // these we print in order
    fprintf(fp, "These are the events scheduled for the future\n");
  
    temp_ev_ptr = queue;

    while (temp_ev_ptr != NULL)
    {
      fprintf(fp, "Time: %lf\tType: %d\tSchedule Status: %d\tLineage: %ld\tTime Created: %lf\n", temp_ev_ptr->event_time, temp_ev_ptr->event_type, temp_ev_ptr->sched_status, temp_ev_ptr->lin_id, temp_ev_ptr->time_created);

      temp_ev_ptr = temp_ev_ptr->next_event;
    }
  }
  else
  {
    fprintf(fp, "There are NO events scheduled for the future\n");
  }
}

/******************************************************************************/
/*                             PrintOldEvents                                 */
/******************************************************************************/
void PrintOldEvents(FILE *fp, EVENT list)
{
  EVENT temp_ev_ptr;
  EVENT temp_prev_ev_ptr;

  if (list == NULL)
  {
    fprintf(fp, "There are NO old events to report\n");
  }
  else
  {
    // we print in reverse order because the list is made in stack
    // fashion and we want to print it in the order in which they were
    // logged, so go backwards!
    
    fprintf(fp, "These are old events\n");
  
    temp_prev_ev_ptr = NULL;
    temp_ev_ptr = list;

    // first get to the end of the list
    while (temp_ev_ptr != NULL)
    {
      temp_prev_ev_ptr = temp_ev_ptr;
      temp_ev_ptr = temp_ev_ptr->next_event;
    }

    // now work backwards through the list
    while (temp_prev_ev_ptr != NULL)
    {
      fprintf(fp, "Time: %lf\tType: %d\tSchedule Status: %d\tLineage: %ld\tTime Created: %lf\tSecond Lineage: %ld\tSecond Parent Time: %lf\n", temp_prev_ev_ptr->event_time, temp_prev_ev_ptr->event_type, temp_prev_ev_ptr->sched_status, temp_prev_ev_ptr->lin_id, temp_prev_ev_ptr->time_created, temp_prev_ev_ptr->second_lin_id, temp_prev_ev_ptr->second_parent_time);

      temp_prev_ev_ptr = temp_prev_ev_ptr->previous_event;
    }

  }// we had events to report
}


/******************************************************************************/
/*                         PrintInternalBranches                              */
/******************************************************************************/
void PrintInternalEdgeLengths(FILE *fp, NETWORK nt)
{
  long node_i;

  // go through the node array and print the edge length
  // from the node to its parent providing the node is not
  // an extant taxa

  fprintf(fp, "These are the internal (non-extant taxa) branch lengths\n");

  for (node_i = 0; node_i < nt->num_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->extant != 1)
    {
	if (nt->NodeArray[node_i]->parent1 != NULL)
	{
          fprintf(fp, "c a %ld %ld %lf\n",
		  nt->NodeArray[node_i]->parent1->id,
		  nt->NodeArray[node_i]->id,
		  nt->NodeArray[node_i]->brlen1_t);
	}

	if (nt->NodeArray[node_i]->parent2 != NULL)
	{
          fprintf(fp, "c a %ld %ld %lf\n",
		  nt->NodeArray[node_i]->parent2->id,
		  nt->NodeArray[node_i]->id,
		  nt->NodeArray[node_i]->brlen2_t);
	}
    }
  }
}
