/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include "globals.h"
#include "node.h"
#include "event.h"
#include "network.h"
#include "utilities.h"
#include "twister.h"
#include "sequences.h"

/******************************************************************************/
/*                      Assign Artificial Sequences                           */
/******************************************************************************/
void AssignArtificialSequences(int num_parents, LINEAGE process_array[], 
		               int node_index, int prev_update_yn, 
			       double current_time)
{
  /**********************************************************************/
  /*   This routine gives "token" sequences to nodes so that diploid    */
  /*   hybrid processing can take place without calling NSGW/SG for     */
  /*   sequences.  (This is the case when we choose second parents      */
  /*   for hybrids based on something other than hamming distance, and  */
  /*   no_simultaneous_sequences have been specified.)  So we want the  */
  /*   the code to proceed along and we just give fake sequences and    */
  /*   never report them.                                               */
  /*                                                                    */
  /* note: we don't use current_time or prev_update_yn here but want    */
  /*       to keep the calling of this routine as much the same as      */
  /*       the AssignSequences call                                     */
  /**********************************************************************/

  /**********************************************************************/
  /*      num_parents = 0 for the ancestor give "AAAA"                  */
  /*      num_parents = 1 for all others, copy parental sequences       */
  /*                                                                    */
  /*   (don't need to be concerned about hybrids here because hybrid    */
  /*	routine handles sequence assignment there)                      */
  /**********************************************************************/

  int num_of_seqs, hchromo_i, strand_i, nhc, nsphc;
  LINEAGE temp_ptr;

  temp_ptr = process_array[node_index];
  nhc = temp_ptr->num_hom_chromos;
  nsphc = temp_ptr->num_strands_per_hom_chrom;
  num_of_seqs = nhc * nsphc;

  /* root case */
  if (num_parents == 0)
  {
    for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
    {
      for (strand_i = 0; strand_i < nsphc; strand_i++)
      {
        strcpy(temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	       "AAAA");
      }
    }
  }

  /* simple child case */
  if (num_parents == 1)
  {
    for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
    {
      for (strand_i = 0; strand_i < nsphc; strand_i++)
      {
        strcpy(temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	       temp_ptr->parent1->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      }
    }
  }
} // end AssignArtificialSequences

/******************************************************************************/
/*                              Assign Sequences                              */
/******************************************************************************/
void AssignSequences(int num_parents, LINEAGE process_array[], int node_index, 
		     int prev_update_yn, double current_time)
{
  /**********************************************************************/
  /*      num_parents = 0 for the ancestor                              */
  /*      num_parents = 1 for all others                                */
  /*                                                                    */
  /* note: hybrids are handled in the hybrid events and feed their      */
  /*       requests for sequences here one at a time so no hybrid       */
  /*       or 2 parent stuff here                                       */
  /**********************************************************************/
  
  FILE *pipe_ptr;
  char exec_NSGW[500];
  char temp_sequence[MAX_SEQ_LENGTH+2]; // end of string (\0) and carriage ret
  char * return_sequences[MAX_NUM_HOM_CHROMOS * MAX_NUM_STRANDS_PER_HC];
  int seq_i;
  int rm_error;
  int num_of_seqs;
  int nhc;
  int nsphc;

  int hchromo_i, strand_i;
  LINEAGE temp_ptr;

  temp_ptr = process_array[node_index];
  num_of_seqs = temp_ptr->num_hom_chromos * 
                temp_ptr->num_strands_per_hom_chrom;
  
  //  NOTE: do not specify output file for NSGW because the pipe wants
  //  to read it back from stdout which is the default for NSGW

  char *exec_nsgw_0 = "./NSGW -d -l ";
  char *temp_file_partial = "nsgw_temp_in";
  char temp_file[250] = "\0";
  char pid_str[11] = "\0";

  if (strlen(temp_file_partial) > 199)
  {
    printf("NETGEN ERROR: File name to pass to NSGW is too long\n");
    exit(0);
  }

  // add the pid so we get a unique file name for this simulation
  strcat(temp_file, temp_file_partial);
  strcat(temp_file, "_");
  sprintf(pid_str, "%d", pid); 
  strcat(temp_file, pid_str);
  
  /*************************************************************************/
  /*                        setup/admin work                               */
  /*************************************************************************/

  // set up memory for the return sequences we will get -- note that
  // this memory is freed at the end of this function as it is no longer
  // needed once the sequences are copied into the proper node's sequences
  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    return_sequences[seq_i] = (char *) calloc((MAX_SEQ_LENGTH+2),sizeof(char));

    if (return_sequences[seq_i] == NULL)
    {
      printf("NETGEN ERROR: Could not allocate space for return sequences\n");
      exit(0);
    }
  }


  /*************************************************************************/
  /*                  write the input file for NSGW                        */
  /*************************************************************************/

  MakeNSGWInputFile(num_parents, process_array, node_index, prev_update_yn,
        	    current_time, temp_file);


  /***********************************************************************/
  /*                  construct the call for NSGW                        */
  /***********************************************************************/
  // construct the command line for NSGW -- is done once per node as NSGW
  // knows to do a separate call to SG for each sequence
  MakeCall(exec_NSGW, exec_nsgw_0, temp_ptr->seq_length,
	   temp_ptr->id, temp_file);

  //DEBUGGING
  //printf("NETGEN DEBUG: command line for NSGW is: %s\n", exec_NSGW);
  //printf("NETGEN DEBUG: the above is a call to NSGW for node id %ld\n", process_array[node_index]->id);
  //END DEBUGGING

  /***********************************************************************/
  /*             execute the call for NSGW  & get the sequences          */
  /***********************************************************************/
  if ((pipe_ptr = popen(exec_NSGW,"r")) != NULL)
  {
    // NSGW gives the sequences -- so loop through the reading process
    for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
    {
      fgets(temp_sequence, (temp_ptr->seq_length+2), pipe_ptr);
      strcpy(return_sequences[seq_i], temp_sequence);
    }  // done reading the sequences in

    // close the pipe!
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NETGEN ERROR: Unexpected issue with pipe call to NSGW\n");
    exit(0);
  }

  /***********************************************************************/
  /*        assign the sequences to the lineage here in NETGEN           */
  /***********************************************************************/

  // DEBUGGING
  //printf("NETGEN DEBUG: the node's seqs BEFORE reading back results\n");
  //seq_i = 0;
  //nhc = temp_ptr->num_hom_chromos;
  //nsphc = temp_ptr->num_strands_per_hom_chrom;
  //for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nsphc; strand_i++)
    //{
      //printf("NETGEN DEBUG: %s\n", temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      //seq_i++;
    //}
  //}
  // END DEBUGGING


  seq_i = 0;
  nhc = temp_ptr->num_hom_chromos;
  nsphc = temp_ptr->num_strands_per_hom_chrom;

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {
      strcpy(temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence,return_sequences[seq_i]);
      seq_i++;
    }
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: the node's seqs AFTER reading back results\n");
  //seq_i = 0;
  //nhc = temp_ptr->num_hom_chromos;
  //nsphc = temp_ptr->num_strands_per_hom_chrom;
  //for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nsphc; strand_i++)
    //{
      //printf("NETGEN DEBUG: %s\n", temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      //seq_i++;
    //}
  //}
  // END DEBUGGING

  /***********************************************************************/
  /*         additional processing for root seqs                         */
  /***********************************************************************/
  // note that if we are dealing with the root, within 
  // each homologous chromosome, all chromatids should be the same, so redo   
  // the strands here so that within each chromosome, all strands get 
  // strand's 0 sequence
  if (num_parents == 0)
  {
    seq_i = 0;
    nhc = temp_ptr->num_hom_chromos;
    nsphc = temp_ptr->num_strands_per_hom_chrom;

    for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
    {
      for (strand_i = 0; strand_i < nsphc; strand_i++)
      {
        strcpy(temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence,temp_ptr->hom_chromos[hchromo_i]->strands[0]->sequence);

        seq_i++;
      }
    }
  }

  /***********************************************************************/
  /*    clean up the newlines which signal the end of the sequence       */
  /*    but we do not want to keep as part of the node's sequence        */
  /***********************************************************************/
  CleanUpSequences(process_array[node_index]);
  
  /***********************************************************************/
  /* remove the temp file so it will be written fresh next time around   */
  /***********************************************************************/

  rm_error = remove(temp_file);

  if (rm_error != 0)
  {
    printf("NETGEN ERROR: There was a problem removing the temp input file for NSGW\n");
    exit(0);
  }

  /**************************************/
  /* be good and clean up memory used   */
  /**************************************/
  // be good and clean up memory used
  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    free(return_sequences[seq_i]);
  }

  // DEBUGGING
  //for (hchromo_i = 0; hchromo_i < temp_ptr->num_hom_chromos; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < temp_ptr->num_strands_per_hom_chrom; strand_i++)
    //{
      //printf("NETGEN DEUBG: Before leaving assign seqs for hchromo %d strand %d:%s \n", hchromo_i, strand_i, temp_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
  // END DEBUGGING
  
} // end AssignSequences

/******************************************************************************/
/*                            MakeNSGWInputFile                               */
/******************************************************************************/
void MakeNSGWInputFile(int num_parents, LINEAGE process_array[], int node_index,
	               int prev_update_yn, double current_time, 
		       char *temp_file)
{
  FILE *fp_temp_out;
  int hchromo_i, strand_i;
  double time_distance, ev_distance;
  LINEAGE temp_prev_ptr;

  int num_of_seqs = process_array[node_index]->num_hom_chromos * 
	            process_array[node_index]->num_strands_per_hom_chrom;
 
  // first open the file that will be used for NSGW input
  fp_temp_out = fopen(temp_file,"w");
  if (fp_temp_out == NULL)
  {
    printf("NETGEN ERROR:  Cannot open the temp output file for simultaneous sequence work.\n");
    exit(0);
  }

  // now write to the file
  if (num_parents == 0)
  {
    // note:  for the case of 0 parents -- no sequences or network info to list
    //        just 0 parents and number of seqs needed back -- this is 
    //        the root case where NSGW will get the root sequences
    fprintf(fp_temp_out, "%d     %d     %d\n", pid, num_parents, num_of_seqs);
  } 
  else if (num_parents == 1) 
  {
    // give the header line with number of parents and number of sequences
    fprintf(fp_temp_out, "%d     %d     %d\n", pid, num_parents, num_of_seqs);

    // decide which node to which we should be referring for history stuff
    // if the node (lineage) was previously updated, we need to use 
    // this node (lineage); if not, then we can use the parent node (lineage)
    // as the reference
    if (prev_update_yn == 1)
    {
      temp_prev_ptr = process_array[node_index];

      // DEBUGGING -- keep debugging or not?? -- hopefully no....big time sink
      //printf("NETGEN DEBUG: we previously updated this lineage id  %ld\n", process_array[node_index]->id);
      // END DEBUGGING

    }
    else
    {
      temp_prev_ptr = process_array[node_index]->parent1;
    }

    // give the sequences of the appropriate node
    for (hchromo_i = 0; hchromo_i < temp_prev_ptr->num_hom_chromos; hchromo_i++)
    {
      for (strand_i = 0; 
           strand_i < temp_prev_ptr->num_strands_per_hom_chrom; 
	   strand_i++)
      {
        fprintf(fp_temp_out, "%s\n", temp_prev_ptr->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      }
    }

    // now print the 'network' stuff; however since this is only one piece
    // going to NSGW all the values are pre-determined for that side of things
    //
    // #nodes #arcs #leaves #hybrids
    fprintf(fp_temp_out,"2 1 1 0 \n");

    // node line: n node# nodetype
    fprintf(fp_temp_out,"n 1 0 \n");

    // leaf line: node# name nodetype
    fprintf(fp_temp_out,"l 2 temp 0 \n");

    // arc line: srcnode# destnode# length
    if (prev_update_yn == 1)
    {
      // if previously updated, we need a modified branch length which
      // reflects the amount of evolutionary time since the last sequence
      // update; must be careful to make sure this modified branch length
      // reflects ev time difference not clock time difference
      // NOTE:  gamma value is only used if doing a non-ultramtretic topology!

      time_distance = current_time - process_array[node_index]->prev_update_time;
      if (ultrametric_yn == 0) 
      {
        ev_distance = time_distance * 
	              height_scalar *
	              process_array[node_index]->gamma_value;
      
      }
      else // ultrametric
      {
        ev_distance = time_distance * 
	              height_scalar; 
      }

      fprintf(fp_temp_out,"a 1 2 %lf\n", ev_distance);


      // DEBUGGING
      //printf("NETGEN DEBUG: prev update situation -- inter bl to give NSGW is %lf (ct %lf put %lf gv %lf)\n", ev_distance,  current_time, process_array[node_index]->prev_update_time, process_array[node_index]->gamma_value); 
    }
    else 
    {
      // note that we know it is brlen1_ev because there is always only one 
      // parent when we are here and brlen1 refers to the length TO parent1 
      // from the current node (and we are concerned with evolutionary time, 
      // not clock time)
      
      fprintf(fp_temp_out,"a 1 2 %lf\n", process_array[node_index]->brlen1_ev);

      // DEBUGGING
      //printf("NETGEN DEBUG: NOT a prev update situation -- bl to give NSGW (from c %ld to p %ld) is %lf\n", process_array[node_index]->id, process_array[node_index]->parent1->id,  process_array[node_index]->brlen1_ev); 
    }
  }

  // now close the input file to NSGW (otherwise it won't be readable to NSGW)
  fclose(fp_temp_out);
}


/******************************************************************************/
/*                             MakeCall(...)                                  */
/******************************************************************************/
void MakeCall(char * exec_NSGW, char * exec_nsgw_0, int seq_length, 
	      int node_id, char * temp_file) 
{
  int error_line_length;
  int exec_line_length;
  char NSGW_error_file[500] = "\0"; // be good with memory
  char pid_str[11] = "\0";
  char seq_length_str[5] = "\0";
  char rand_num_seed_str[11] = "\0";
  unsigned long rnd_nsgw;

  // construct the call line for NSGW -- need to piece in sequence length,
  // random number seed, input file, and seqgenoptions (usually the model
  // we want Seq-Gen to use, but can be anything that would be legit for
  // Seq-Gen -- the wrapper just adds this string to the command line when
  // it calls Seq-Gen itself)
  
  strcpy(exec_NSGW, exec_nsgw_0);
  exec_line_length = strlen(exec_nsgw_0); 

  sprintf(seq_length_str, "%d", seq_length); 
  strcat(exec_NSGW, seq_length_str);
  exec_line_length = exec_line_length + strlen(seq_length_str); 

  strcat(exec_NSGW, " -z ");
  exec_line_length = exec_line_length + 4; 

  // make and track the random number seed and the node for which we are 
  // using this random number we give to NSGW, if requested
  rnd_nsgw = genrand_int32();
  if (track_nsgw_seeds_yn)
  {
    if (nsgw_seeds_index >= MAX_NSGW_SEEDS)
    {
      printf("NETGEN WARNING: Exceeded max size for storing NSGW seed info\n");
      printf("NETGEN WARNING: continuing to run, but will no longer store NSGW seed info\n");
    }
    else
    {
      nsgw_seeds[nsgw_seeds_index++] = rnd_nsgw;
      nsgw_seeds_node[nsgw_seeds_node_index++] = node_id;
    }
  }

  sprintf(rand_num_seed_str, "%lu", rnd_nsgw); 
  strcat(exec_NSGW, rand_num_seed_str);
  exec_line_length = exec_line_length + strlen(rand_num_seed_str); 

  strcat(exec_NSGW, " ");
  strcat(exec_NSGW, temp_file);
  exec_line_length = exec_line_length + strlen(temp_file) + 1; 

  strcat(exec_NSGW, " ");
  exec_line_length = exec_line_length + 1; 

  // seq_gen_options comes in with the required quotes already
  strcat(exec_NSGW, seq_gen_options);
  exec_line_length = exec_line_length + strlen(seq_gen_options) + 2; 

  // NOTE:  we use the single > for overwrite (>> would be append but
  //        that is way too much data for operational mode!)
  strcat(NSGW_error_file, " 2> se_out_nsgw");
  strcat(NSGW_error_file, "_");
  sprintf(pid_str, "%d", pid); 
  strcat(NSGW_error_file, pid_str);
  error_line_length = strlen(NSGW_error_file); 
  strcat(exec_NSGW, NSGW_error_file);
  exec_line_length = exec_line_length + error_line_length;

  if ((exec_line_length > 500) || (error_line_length > 200))
  {
    printf("NETGEN ERROR: Command line (or error file name) for NSG_wrapper is too long %d, %d\n", exec_line_length, error_line_length);
    exit(0);
  }
  else
  {
   // DEBUGGING
   //printf("NETGEN DEBUG: Command line from MakeCall is: %s\n", exec_NSGW);
  }

} // end MakeCall


/******************************************************************************/
/*                          CleanUpSequences                                  */
/******************************************************************************/
void CleanUpSequences(LINEAGE nd)
{
  int hchromo_i, strand_i;
  int nhc, nsphc;

  nhc = nd->num_hom_chromos;
  nsphc = nd->num_strands_per_hom_chrom;

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {
      nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[nd->seq_length] = '\0';
    }
  }

}


/******************************************************************************/
/*                           AssignOutgroupSequences                          */
/******************************************************************************/
void AssignOutgroupSequences(LINEAGE NodeArray[], int node_index, 
		             int num_nodes, int num_extant_taxa, 
			     double current_time)
{
  // here we do the work of assigning sequences to the outgroup
  // the big issue is that we want a leaf that is somewhat similar to the
  // leaves of the phylogeny, but not too similar
  //
  // max_tries tells us the maximum number of sequences to try in assigning
  // each sequence; while max_outg_diff_value and min_outg_diff_value give
  // us the range for how dissimilar/similar the new sequence should be

  int seq_i, try_i;                 // loop variables
  int hchromo_i, strand_i;          // more loop variables
  int outg_found_yn;                // found flag
  int best_avg_h_dist, best_try;    // tracking variables

  // to store the tried sequences
  char * outg_try[MAX_OUTGROUP_TRIES]; 

  // to store the final sequences decided upon
  char * final_sequences[MAX_NUM_HOM_CHROMOS * MAX_NUM_STRANDS_PER_HC];

  // to store avg dist of tried seqs to all other extant taxa
  int * avg_h_dist_value; 

  // to track whether the tried sequence under study meets the range for
  // each extant taxon (-1/1 entry)
  int * meets_goal;       
  
  // to make things shorter and clearer
  LINEAGE outg = NodeArray[node_index];
  int nhc = outg->num_hom_chromos;
  int nsphc = outg->num_strands_per_hom_chrom;
  int num_of_seqs = nhc * nsphc;

  // setup dynamically allocated stuff
  avg_h_dist_value = (int *) calloc(MAX_OUTGROUP_TRIES, sizeof(int));
  meets_goal = (int *) calloc(MAX_LINEAGES, sizeof(int));

  if ( (avg_h_dist_value == NULL) || (meets_goal == NULL))
  {
    printf("NETGEN ERROR: Could not allocate space for outgroup processing\n");
    exit(0);
  }

  // set up memory for storing the tries
  for (try_i = 0; try_i < max_outg_tries; try_i++)
  {
    outg_try[try_i] = (char *) calloc((MAX_SEQ_LENGTH+2),sizeof(char));
    if (outg_try[try_i] == NULL)
    {
      printf("NETGEN ERROR: Could not allocate space for tried sequences as part of the outgroup work\n");
      exit(0);
    }
  }

  // set up the memory for the final sequences
  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    final_sequences[seq_i] = (char *) calloc((MAX_SEQ_LENGTH+2),sizeof(char));
    
    if (final_sequences[seq_i] == NULL)
    {
      printf("NETGEN ERROR: Could not allocate space for final sequences as part of the outgroup work\n");
      exit(0);
    }
  }

  // DEBUGGING
  //printf ("NETGEN DEBUG: Starting Outgroup Sequence Work\n");
  // END DEBUGGING
  
  seq_i = 0;

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {

      // DEBUGGING
      //printf("NETGEN DEBUG: Working out outgroup hchromo %d, strand %d\n", hchromo_i, strand_i);
      // END DEBUGGING

      outg_found_yn = 0;

      for (try_i = 0; try_i < max_outg_tries; try_i++)
      {
	outg_found_yn = MakeOutgroupTry(try_i, &best_try, hchromo_i, strand_i,
			                num_nodes, num_extant_taxa, seq_i,
					NodeArray, node_index, current_time,
					outg_try, meets_goal, avg_h_dist_value);
        if (outg_found_yn)
          break;

        // DEBUGGING
        //printf("NETGEN DEBUG: Need another try for hchromo_i %d, strand_i %d, which corresponds to seq_i %d \n", hchromo_i, strand_i, seq_i);
	// END DEBUGGING

      } // do next try

      // if outgroup has not been found, choose the 'best' one of the tries

      // first see if one of the tries meets the parameters when
      // using the average
      if (outg_found_yn == 0)
      {
        for (try_i = 0; try_i < max_outg_tries; try_i++)
        {
          // cast avg_h_dist_value as an int so we can compare properly
          if ( (((int) avg_h_dist_value[try_i]) >= (min_outg_diff_value)) &&
               (((int) avg_h_dist_value[try_i]) <= (max_outg_diff_value)))
          {
            outg_found_yn = 1;
            best_try = try_i;
	    break;
	  }
        }
      }

      // if still have not found outgroup, pick the one with the highest
      // average hamming distance (this assumes that the difficulty is
      // getting a sequences 'different')
      if (outg_found_yn == 0)
      {
        best_avg_h_dist = 0;
        best_try = 0;
        for (try_i = 0; try_i < max_outg_tries; try_i++)
        {

	  // DEBUGGING
	  //printf("NETGEN DEBUG:  Looking at try %d whose avg ham dist is %d\n", try_i, avg_h_dist_value[try_i]);

	  // END DEBUGGING

          if (avg_h_dist_value[try_i] > best_avg_h_dist)
	  {
            best_try = try_i;
	    best_avg_h_dist = avg_h_dist_value[try_i];
	  }
        }
	// we are guaranteed to have found something!
	outg_found_yn = 1;
      }

      // now take our best found (from any of the above methods) and
      // put it into the final sequences array

      strcpy(final_sequences[seq_i], outg_try[best_try]);

      // DEBUGGING
      //printf("NETGEN DEBUG:  We have our final %d seq as:%s\n", seq_i, final_sequences[seq_i]);
      //printf("NETGEN DEBUG:  Our best try was %d and its avg hamming distance was %d \n", best_try, avg_h_dist_value[best_try]);
      //printf("NETGEN DEBUG:  The length of outg_try[best_try] was %d \n", strlen(outg_try[best_try]));
      //printf("NETGEN DEBUG:  max_outg_tries is %d \n", max_outg_tries);
      //for (try_i = 0; try_i < max_outg_tries; try_i++)
      //{
        //printf("NETGEN DEBUG:  The length of outg_try[%d] was %d \n", try_i, strlen(outg_try[try_i]));
      //}
      // END DEBUGGING

      seq_i++;

    } // next strand
  } // next homologous chromosome 

  // now that we are done with all the processing, we can put the final
  // good sequences into the node's sequences (could not do earlier because
  // each call to Assign Sequences overwrote the node's sequences)
  seq_i = 0;
  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {
      strcpy(outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
	     final_sequences[seq_i]);

      // DEBUGGING
      //printf("NETGEN DEBUG: end of outg work hchromo %d, strand %d ns:%s, fs:%s\n", hchromo_i, strand_i, outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence, final_sequences[seq_i]);
      //printf("NETGEN DEBUG: end of outg work hchromo %d, strand %d, length of ns:%d, length of fs:%d\n", hchromo_i, strand_i,strlen(outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence), strlen(final_sequences[seq_i]));
      // END DEBUGGING

      seq_i++;
    }
  }

  // DEBUGGING
  //printf("NETGEN DEBUG:  After outgroup processing our outgroup has:\n");
  //for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nsphc; strand_i++)
    //{
      //printf("NETGEN DEBUG:%s\n", outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
  //END DEBUGGING
  
  // be good and clean up
  free(avg_h_dist_value);
  free(meets_goal);
  for (try_i = 0; try_i < max_outg_tries; try_i++)
  {
    free(outg_try[try_i]);
  }
  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    free(final_sequences[seq_i]);
  }

}

/******************************************************************************/
/*                           MakeOutgroupTry                                  */
/******************************************************************************/
int MakeOutgroupTry(int try_i, int *best_try, int hchromo_i, int strand_i,
		    int num_nodes, int num_extant_taxa, int seq_i,
		    LINEAGE NodeArray[], int node_index, double current_time, 
		    char *outg_try[], int * meets_goal, int * avg_h_dist_value)
{
  int node_i, h_dist;
  LINEAGE outg = NodeArray[node_index];


  int outg_found_yn = 0;
  int all_h_dist_sum = 0;

  // first get a set of potential sequences
  AssignSequences(1, NodeArray, node_index, 0, current_time);

  // copy the specific sequence from node into outg_try (because on the 
  // next try round the node's seq will be rewritten in AssignSequences)
  // this is because AssignSequences does all sequences and we want to 
  // deal with sequences one at a time here
      
  strcpy(outg_try[try_i], 
         outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

  // DEBUGGING
  //printf("NETGEN DEBUG: just copied NodeArray sequence(length %d) into outg_try (length %d)\n", strlen(outg_try[try_i]), strlen(outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence));
  // END DEBUGGING

  // first initialize meet goal so we don't get screwed up in 
  // the next part of the code
  for (node_i = 0; node_i < (num_nodes-1); node_i++)
  {
    meets_goal[node_i] = 0;
  }

  // now go through all the extant taxa of the tree/network and do distance work
  // (but don't compare it to itself which is the last entry in NodeArray!, and
  //  only do comparisons to extant taxa, not all leaves which may include
  //  lineages that died out prior to the end of the simulation)
      
  for (node_i = 0; node_i < (num_nodes-1); node_i++)
  {
    if (NodeArray[node_i]->extant)
    {
      // calc hamming distance between the two sequences
      h_dist = HammingDistanceStrings(outg_try[try_i], 
		                      NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

      // track
      all_h_dist_sum = all_h_dist_sum + h_dist;

      // DEBUGGING
      //printf("NETGEN DEBUG: all_h_dist_sum is now %d \n", all_h_dist_sum);

      // compare to see if dist meets min/max restrictions and track
      if ( (h_dist <= max_outg_diff_value) && 
           (h_dist >= min_outg_diff_value)) 
      {
        meets_goal[node_i] = 1;
      }
      else
      {
        meets_goal[node_i] = -1;
      }
    } // end of checking this extant taxon
  } // keep going through and looking for extant taxa

  // store avg_dist and keep all as integers to keep things easy -- remember
  // that outgroup leaf is not counted as part of num_extant_taxa; realize
  // this truncates avg dist but this is a minor factor when doing large strings
  avg_h_dist_value[try_i] = (int) ((all_h_dist_sum)/(num_extant_taxa));

  // DEBUGGING
  //printf("NETGEN DEBUG: We just calc avg ham dist as %d \n", avg_h_dist_value[try_i]);
  // END DEBUGGING

  // check to see if all entries of meet_goal are met -- if yes,
  // we have our sequence_i and can stop -- so assume we have it and
  // see if something does not meet our goal (note that the array
  // was initialized to 0 so the nodes that are not leafs have a 0
  // so look for -1s to be the problem)
      
  outg_found_yn = 1;

  for (node_i = 0; node_i < (num_nodes-1); node_i++)
  {
    if (meets_goal[node_i] < 0)
        outg_found_yn = 0;
  }

  // if we have our sequence assiged for seq_i, set vars and continue on
  // with rest of processing (do not keep doing tries for this sequence_i)
  if (outg_found_yn)
    (*best_try) = try_i;

  return (outg_found_yn);
}
