/*****************************************************************************/
/*                                                                           */
/*  NetGen kappa version                                                     */
/*  Simulator for phylogenetic networks with sequences.                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "globals.h"
#include "twister.h"
#include "utilities.h"

/******************************************************************************/
/*                                 MinInt                                     */
/******************************************************************************/
int MinInt(int a, int b)
{
  if (a < b)
    return a;
  else
    return b;
}

/******************************************************************************/
/*                                 MaxInt                                     */
/******************************************************************************/
int MaxInt(int a, int b)
{
  if (a < b)
    return b;
  else
    return a;
}

/******************************************************************************/
/*                               MaxDouble                                    */
/******************************************************************************/
double MaxDouble(double a, double b)
{
  if (a < b)
    return b;
  else
    return a;
}

/******************************************************************************/
/*                                MaxLong                                     */
/******************************************************************************/
long MaxLong(long a, long b)
{
  if (a < b)
    return b;
  else
    return a;
}

/******************************************************************************/
/*                                MinLong                                     */
/******************************************************************************/
long MinLong(long a, long b)
{
  if (a < b)
    return a;
  else
    return b;
}


/******************************************************************************/
/*                         HammingDistanceStrings                             */
/******************************************************************************/
int HammingDistanceStrings(char * str_1, char * str_2)
{
  int char_i, length, count;

  // if the strings differ in length we automatically have a hamming
  // distance > 0; so add that to count right away and then we can 
  // stop comparing strings when the shortest string is exhausted 

  length = MinInt(strlen(str_1), strlen(str_2));
  count = 0;
  count = count + (abs(strlen(str_1) - strlen(str_2))); 

  // DEBUGGING
  //printf("NETGEN DEBUG:  Our length for comparison is %d (from min(%d,%d))\n", length, strlen(str_1), strlen(str_2));

  for (char_i = 0; char_i < length; char_i++)
  {
    if (str_1[char_i] != str_2[char_i])
      count++;
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: Our two strings below have a distance of %d\n", count);
  //printf("NETGEN DEBUG: str1:%s, str2:%s\n", str_1, str_2);
  // END DEBUGGING


  return count;
}

/******************************************************************************/
/*                                MaxTime                                     */
/******************************************************************************/
double MaxTime(double t1, double t2, double t3)
{
  if ( (t1 >= t2) && (t1 >= t3) ) // t1 is the max
    return t1;

  if ( (t2 >= t1) && (t2 >= t3) ) // t2 is the max
    return t2;

  if ( (t3 >= t1) && (t3 >= t2) ) // t3 is the max
    return t3;

  // Should not get here....
  printf("NETGEN ERROR: we've got a problem with MaxTime routine\n");
  exit(0);
}


/******************************************************************************/
/*                                MinTime                                     */
/******************************************************************************/
double MinTime(double t1, double t2, double t3)
{
  if ( (t1 <= t2) && (t1 <= t3) ) // t1 is the min
    return t1;

  if ( (t2 <= t1) && (t2 <= t3) ) // t2 is the min
    return t2;

  if ( (t3 <= t1) && (t3 <= t2) ) // t3 is the min
    return t3;

  // Should not get here....
  printf("NETGEN ERROR: we've got a problem with MinTime routine\n");
  exit(0);
}

/******************************************************************************/
/*                               Gamma Dist                                   */
/******************************************************************************/
double GammaDist(int shape, double scale)
{
  // this routine generates a gamma random variable according to the
  // shape and scale values;  this algorithm assumes/requires that the
  // shape value is an integer as doing the inverse of a gamma is messy
  // otherwise -- see Hahn and Shapiro "Statistical Models in Engineering"
  // page 242 for details (the idea is to do a sum of exponential random
  // variables which we can get from the uniform random number generator)

  double exp_value;
  double gamma_value;
  double unif_rand_num;
  int i;

  gamma_value = 0;

  for (i = 1; i <= shape; i++)
  {
    unif_rand_num = genrand_real2();
    exp_value =  log(1-unif_rand_num);
    gamma_value = gamma_value + exp_value;
  }

  gamma_value = (-1/scale) * gamma_value;

  // DEBUGGING
  //printf("NETGEN DEBUG:  just generated a gamma value of %lf\n", gamma_value);
  // END DEBUGGING

  return gamma_value;
}

/******************************************************************************/
/*                           NextExpArrInterval                               */
/******************************************************************************/
double NextExpArrInterval(double rate)
{
  // knowing that poisson processes have exponential inter-arrival times,
  // we want a randomly drawn inter-arrival time which will follow the
  // exponential distribution;  however, we only have available a uniform
  // random number generator, so we take the cdf of the exponential and 
  // then our uniform number is used to index the cdf of the exponential 
  // and give us our exponential random number -- hence doing a mapping
  //
  // the following equation does all of this and returns the exponential
  // interarrival time we want -- see Hahn&Shapiro "Statistical Modeling 
  // in Engineering" page 258

  double next_arrival;
  double unif_rand_num;

  // this gives us a uniform random value in the range of [0,1)
  // (we don't want 1 to be a possibility, because 1 in the below equation
  //  leads to the ln(1-1) and the ln(0) is undefined)
  unif_rand_num = genrand_real2();

  // DEBUGGING
  //printf("NETGEN DEBUG: our unif rand num is %f\n", unif_rand_num);

  // NOTE:  we really should check that the rate is not 0 before calling
  //        this routine as rate = 0 will make this bomb and a zero rate
  //        is really undefined, but as a safety net, we return DBL_MAX
  //        if the rate is 0 (but realize this is problematic because 
  //        if DBL_MAX is used in calculation with another value, results
  //        may not be as expected due to overflow issues!)
  
  if (rate > 0)
  {
    next_arrival = (-1 * log(1-unif_rand_num))/rate;
  }
  else
  {
    next_arrival = DBL_MAX;

    printf("NETGEN WARNING: Exponential random number generator was given a rate of zero, which is problematic\n");
  }

  // DEBUGGING
  //printf("NETGEN DEBUG: our exp rand num is %f\n", next_arrival);
  //printf("NETGEN DEBUG: our passed rate was %f\n", rate);
 
  return next_arrival;
}

/******************************************************************************/
/*             Customized Exponential Function for Hamming Distance           */
/******************************************************************************/
long CustomizedExpHD(long max_hd)
{
  // here is where we want to select a random value from the exponential
  // distribution which is defined by the lambda_hd that the user provided

  // we have calculated the cdf in this case to be:
  //
  //                     e^(-HD/lambda_hd) - 1
  //   cdf =  P(H) =  -------------------------
  //                   e^(-HD_max/lambda_hd) - 1
  //

  // when this is solved for HD we get the following equation:
  //
  //   HD = -lambda_hd * ln ((e^(-HD_max/lambd_hd)-1) * P(H) + 1)
  //
  //   and P(H) is a random value from our uniform random number generator
  //   in the range of 0 to max_hd

  // note:  lambda_hd was sanity checked upon input

  long hd_value;
  double hd_value_double;
  double unif_rand_num;
  long unif_rand_num_scaled;
  double e_term;

  unif_rand_num = genrand_real2();

  e_term = exp(((-1*max_hd)/lambda_hd));

  hd_value_double = (-1 * lambda_hd) * log(((e_term - 1) * unif_rand_num) + 1);

  hd_value = (long) trunc(hd_value_double);

  // FIX LATER
  // DEBUGGING
  // printf("NETGEN DEBUG: our hd_value to return is %ld\n", hd_value);
  // printf("%f\n", hd_value_double);

  return hd_value;
}

/******************************************************************************/
/*                             NormRandDist                                   */
/******************************************************************************/
double NormRandHalfDist(double mean, double var)
{
  // this routine gets us random values from a normal distribution using
  // the uniform random number generator -- but we only want positive
  // values, so it is a "half" distribution as times should only be positive

  // this code is based upon the polar algorithm presented in Ross 
  // "A First Course in Probability" pages 450-452

  double u1, u2, v1, v2, s, x;

  s = 1.0;

  while (s >= 1.0)
  {
    u1 = genrand_real1();
    u2 = genrand_real1();

    v1 = (2.0 * u1) - 1.0;
    v2 = (2.0 * u2) - 1.0;

    s = v1 * v1 + v2 * v2;
  }

  x = fabs(v1 * sqrt ( (-2.0 * log(s))/s ));
  x = mean + (sqrt(var) * x);

  return x;
}
