/*****************************************************************************/
/*                                                                           */
/*  NSGW kappa version                                                       */
/*  Wrapper code for Seq-Gen in order to request/process sequences between   */
/*  NetGen and Seq-gen.                                                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>

#include "globals.h"
#include "network.h"
#include "twister.h"
#include "input.h"

/*****************************************************************************/
/*                        ProcessCommandLine                                 */
/*****************************************************************************/

void ProcessCommandLine(int argc, char *argv[], char * seq_gen_options, 
		        int * seq_length, unsigned long * rand_seed)
{

  // NOTE:  the only seven possible command lines are:
  //
  //
  // ********************
  // IF NO SEED IS GIVEN:
  // ********************
  // (A) argc=6  nsg_wrap -d -l <int> <inputfile> "<seq-gen-options>"
  // (B) argc=4  nsg_wrap -a <inputfile> "<seq-gen-options>"
  //
  //
  // ********************
  // IF A SEED IS GIVEN:
  // ********************
  // (C) argc=8  nsg_wrap -d -l <int> -z <int> <inputfile> "<seq-gen-options>"
  // (D) argc=6  nsg_wrap -a -z <int> <inputfile> "<seq-gen-options>"
  //
  //
  //**********************************
  // NOTE:  these are for testing only
  //**********************************
  // (E) argc=8  nsg_wrap -E -d -l <int> <inputfile> <rnd_num_test_file> "<seq-gen-opts>"
  // (F) argc=6  nsg_wrap -F -a <inputfile> <rnd_num_test_file> "<seq-gen-opts>"
  // (G) argc=6  nsg_wrap -G -t <inputfile> <rnd_num_test_file> "<seq-gen-opts>"
  //
  //

  /*********************************************************************/
  /*                  sanity check number of arguments                 */ 
  /*********************************************************************/

  if ((argc != 4)&&(argc != 5)&&(argc != 6)&&(argc != 7)&&(argc != 8)&&
      (argc != 9)&&(argc != 10))
  {
    fprintf(stderr, "NSGW ERROR:  Too many or too few arguments\n");
    fprintf(stderr, "NSGW ERROR:  Your command line had %d arguments\n", argc);
    fprintf(stderr, "NSGW ERROR: Command line must be one of the following:\n");
    fprintf(stderr, "NSGW ERROR:  (A) 'nsg_wrap -d -l <int> <inputfile> \"seq_gen_options\"'\n");
    fprintf(stderr, "NSGW ERROR:  (B) 'nsg_wrap -a <inputfile> \"seq_gen_options\"'\n");
    fprintf(stderr, "NSGW ERROR:  (C) 'nsg_wrap -d -l <int> -z <int> <inputfile> \"seq_gen_options\"'\n");
    fprintf(stderr, "NSGW ERROR:  (D) 'nsg_wrap -a -z <int> <inputfile> \"seq_gen_options\"'\n");
    fprintf(stderr, "NSGW ERROR: The commands lines for verification testing are:\n");
    fprintf(stderr, "NSGW ERROR:  (E) 'nsg_wrap -E -d -l <int> <inputfile> <rand_num_test_file> \"seq_gen_options\" '\n");
    fprintf(stderr, "NSGW ERROR:  (F) 'nsg_wrap -F -a <inputfile> <rand_num_test_file> \"seq_gen_options\" '\n");
    fprintf(stderr, "NSGW ERROR:  (G) 'nsg_wrap -G -t <inputfile> <rand_num_test_file> \"seq_gen_options\" '\n");
    exit(0);
  }

  /*********************************************************************/
  /*         process command line according to which one it is         */ 
  /*********************************************************************/
  // this is certainly repetitive and inefficient, but worth keeping
  // the processing straight

  /******************************************************/
  // (A) argc=6  
  // nsg_wrap -d -l <int> <inputfile> "<seq-gen-options>"
  /******************************************************/
  if ( (strcmp(argv[1],"-d")==0) && (strcmp(argv[4],"-z")!=0) )
  {
    ProcessCommandLineA(argc, argv, seq_gen_options, seq_length, rand_seed);
  }

  /******************************************************/
  // (B) argc=4 
  // nsg_wrap -a <inputfile> "<seq-gen-options>"
  /******************************************************/
  if ( (strcmp(argv[1],"-a")==0) && (strcmp(argv[2],"-z")!=0) )
  {
    ProcessCommandLineB(argc, argv, seq_gen_options, seq_length, rand_seed);
  }


  /*************************************************************/
  // (C) argc=8  
  // nsg_wrap -d -l <int> -z <int> <inputfile> <seq-gen-options>
  /*************************************************************/
  if ( (strcmp(argv[1],"-d")==0) && (strcmp(argv[4],"-z")==0) )
  {
    ProcessCommandLineC(argc, argv, seq_gen_options, seq_length, rand_seed);
  }

  /*****************************************************/
  // (D) argc=6 
  // nsg_wrap -a -z <int> <inputfile> <seq-gen-options>
  /*****************************************************/
  if ( (strcmp(argv[1],"-a")==0) && (strcmp(argv[2],"-z")==0) )
  {
    ProcessCommandLineD(argc, argv, seq_gen_options, seq_length, rand_seed);
  }

  /***********************************************************************/
  // (E) argc=8  
  // nsg_wrap -E -d -l <int> <inputfile> <rnd_num_test_file> <seq-gen-opts>
  /***********************************************************************/
  if (strcmp(argv[1],"-E")==0)
  {
    ProcessCommandLineE(argc, argv, seq_gen_options, seq_length, rand_seed);
  }

  /***************************************************************/
  // (F) argc=6 
  // nsg_wrap -F -a <inputfile> <rnd_num_test_file> <seq-gen-opts> 
  /***************************************************************/
  if (strcmp(argv[1],"-F")==0)
  {
    ProcessCommandLineF(argc, argv, seq_gen_options, seq_length, rand_seed);
  }

  /***************************************************************/
  // (G) argc=6 
  // nsg_wrap -G -t <inputfile> <rnd_num_test_file> <seq-gen-opts> 
  /***************************************************************/
  if (strcmp(argv[1],"-G")==0)
  {
    ProcessCommandLineG(argc, argv, seq_gen_options, seq_length, rand_seed);
  }
}

/******************************************************************************/
/*                          RandomNumberSetUp                                 */
/******************************************************************************/
void RandomNumberSetUp()
{
  int num_rand_values;

  /**************************************************/
  /* if comparing topologies between NG and NSGW(a) */  
  /*             (case G command line)              */
  /**************************************************/

  if (rand_tree_test_yn)
  {
    // first read in how many rows this file has and sanity check
    fscanf(fp_in_rand,"%d\n", &num_rand_values);
    if (num_rand_values > (MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES))
    {
      fprintf(stderr, "NSGW ERROR:  too many random numbers (%d) provided for tree test run\n", num_rand_values);
      exit(0);
    }

    // now read in the integers and the corresponding node id's
    // for which they are to be used (as recorded by NG)
    for (rand_num_test_index = 0; 
	 rand_num_test_index < num_rand_values;
	 rand_num_test_index++)
    {
      fscanf(fp_in_rand,"%lu %d\n", &rand_num_tree_test[rand_num_test_index], &rand_num_node_tree_test[rand_num_test_index]);
    }

    // DEBUGGING
    //fprintf(stderr, "NSGW DEBUG: first value of rand_num_tree_test is: %lu\n", rand_num_tree_test[0]);
    //fprintf(stderr, "NSGW DEBUG:   for node: %d\n", rand_num_node_tree_test[0]);
    //fprintf(stderr, "NSGW DEBUG: last value of rand_num_tree_test is: %lu\n", rand_num_tree_test[num_rand_values-1]);
    //fprintf(stderr, "NSGW DEBUG:   for node: %d\n", rand_num_node_tree_test[num_rand_values-1]);
  }

  /************************************************/
  /* if comparing sequences between SG and NSGW   */
  /*      (case E and case F command lines)       */
  /************************************************/

  // This section is not real robust as we are assuming there are 10,000
  // random values to be read in -- but not worth the effort at this time
  // to make flexible because our validation tests will always use the
  // same 10,000 valued file for now.....
  //
  // read in the random numbers (which we know there is 10,000) if provided
  // and see if this will have a chance of meeting our needs since
  // we know we will need AT LEAST a random number for every base
  if (rand_test_yn == 1)
  {
    for (rand_num_test_index = 0; 
         rand_num_test_index < 10000; 
         rand_num_test_index++)
    {
      fscanf(fp_in_rand,"%10lf\n", &rand_num_test[rand_num_test_index]);
    }

    // sanity check how many rand numbers we have	 
    if (10000 < (MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES))
    {
      fprintf(stderr, "NSGW ERROR: not enough random numbers available\n");
      fprintf(stderr, "NSGW ERROR: we know we may need at least a rand num for each base\n");
      fprintf(stderr, "NSGW ERROR: which is MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES\n");
      exit(0);
    }

    // DEBUGGING
    //fprintf(stderr, "NSGW DEBUG: the first rng is: %10lf\n", rand_num_test[0]);
    //fprintf(stderr, "NSGW DEBUG: the last rng is: %10lf\n", rand_num_test[9999]);

  }
  
  /**********************************/
  // for either random number tests */
  /**********************************/
  if (rand_test_yn || rand_tree_test_yn)
  {
    fclose(fp_in_rand);
    // reset the index counter
    rand_num_test_index = 0;
  }

}

/*****************************************************************************/
/*                            ReadNetwork                                    */
/*****************************************************************************/

void ReadNetwork(NETWORK nt)
{
  double brlen, junk_brlen, junk_double;
  char desc;
  int hyb, nhc, nsphc, sl, line_count;
  long nd, src, dest;

  line_count = 1;

  while (!feof(fp_in)) 
  {
    fscanf(fp_in, "%c", &desc);
    switch(desc) 
    {
      case 'c' :              // comment line -- we ignore
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);

        line_count++;
        break;

      case 's' :              // summary line
        fscanf(fp_in, "%d %d %d %d %d %lf", &(nt->num_of_nodes), &(nt->num_of_arcs), &(nt->num_of_leaves), &(nt->num_of_hybrids), &(nt->outgroup_id), &(nt->height));

        // addition to problem line is that if an outgroup is present, we
        // need to read in the min_outg_diff_value, max_outg_diff_value, 
        // and max_outg_tries
        if (nt->outgroup_id > -1)
        {
          fscanf(fp_in, "%d %d %d", &(nt->min_outg_diff_value), &(nt->max_outg_diff_value), &(nt->max_outg_tries));
        }

        fscanf(fp_in, "%c", &desc);
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);

        line_count++;
        break;


      case 'r' :              // root node
      case 'n' :              // internal node descriptor
        fscanf(fp_in, "%ld %d %d %d %d %lf %lf %lf %lf", &nd, &hyb, &nhc, &nsphc, &sl, &junk_double, &junk_double, &junk_double, &junk_double);
        NewNode(nt, nd, nhc, nsphc, sl);
        nt->ndarray[nd]->hybrid = hyb;

	if (desc == 'r')
	{
          nt->root = nt->ndarray[nd];
	}

        fscanf(fp_in, "%c", &desc);
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);      

        line_count++;
        break;


      case 'e' :              // extant leaf node descriptor
        fscanf(fp_in, "%ld %d %d %d %d %lf", &nd, &hyb, &nhc, &nsphc, &sl, &junk_double);
        NewNode(nt, nd, nhc, nsphc, sl);
        nt->ndarray[nd]->leaf = 1;
        nt->ndarray[nd]->extant = 1;

        // up our count of extant taxa
        (nt->num_of_extant_taxa)++;

        fscanf(fp_in, "%c", &desc);
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);

        line_count++;
        break;


      case 'o' :              // outgroup node descriptor
        fscanf(fp_in, "%ld %d %d %d %d %lf", &nd, &hyb, &nhc, &nsphc, &sl, &junk_double);
        NewNode(nt, nd, nhc, nsphc, sl);
        nt->ndarray[nd]->leaf = 1;
        nt->ndarray[nd]->extant = 0;
        nt->ndarray[nd]->outgroup = 1;

        fscanf(fp_in, "%c", &desc);
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);

        line_count++;
        break;


      case 'l' :              // non-extant leaf node descriptor
        fscanf(fp_in, "%ld %d %d %d %d %lf", &nd, &hyb, &nhc, &nsphc, &sl, &junk_double);
        NewNode(nt, nd, nhc, nsphc, sl);
        nt->ndarray[nd]->leaf = 1;
        nt->ndarray[nd]->extant = 0;

        fscanf(fp_in, "%c", &desc);
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);

        line_count++;
        break;

      case 't' :             // "tip" edge descriptor (end node is extant taxon)
      case 'b' :             // branch descriptor (all other edges)
        // we read in both branch lengths (evolutionary and time) but
        // we only care about keeping the evolutionary branch length
        fscanf(fp_in, "%ld %ld %lf %lf", &src, &dest, &brlen, &junk_brlen);
        if (nt->ndarray[src]->left == NULL) 
	{
	  nt->ndarray[src]->left = nt->ndarray[dest];
        } 
	else 
	{
	  if (nt->ndarray[src]->right == NULL) 
	  {
	    nt->ndarray[src]->right = nt->ndarray[dest];
	  } 
	  else 
	  {
	  printf("Error: Node %ld has more than two children\n", src);
	  exit(0);
	  }
        }
        if (nt->ndarray[dest]->parent1 == NULL) 
	{
	  nt->ndarray[dest]->parent1 = nt->ndarray[src];
	  nt->ndarray[dest]->brlen1 = brlen;
        } 
	else 
	{
	  if (nt->ndarray[dest]->parent2 == NULL) 
	  {
	    nt->ndarray[dest]->parent2 = nt->ndarray[src];
	    nt->ndarray[dest]->brlen2 = brlen;	 
	  } 
	  else 
	  {
	    printf("Error: Node %ld has more than two parents\n", dest);
	    exit(0);
	  }
        }

        fscanf(fp_in, "%c", &desc);
        while (desc != '\n')
	  fscanf(fp_in, "%c", &desc);      

        line_count++;
        break;  

        //default  : 
	  //printf("Error on line %d in the input file\n", line_count);
          //exit(0);
          //line_count++;
          //break;
    }
  }

}

/**************************************************************************/
/*                           Command Line A                               */
/**************************************************************************/
void ProcessCommandLineA(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{
   /**********************************************************/
   /* (A) argc=6                                             */
   /* nsg_wrap -d -l <int> <inputfile> "<seq-gen-options>"   */
   /**********************************************************/
   if (strcmp(argv[2],"-l")!=0)
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case A) is incorrect\n");
     exit(0);
   }

   if (atoi(argv[3]) > MAX_SEQ_LENGTH)
   {
     fprintf(stderr, "NSGW ERROR: sequence length too large; limit is %d\n", MAX_SEQ_LENGTH);
     exit(0);
   }
   else
   {
    (*seq_length) = atoi(argv[3]);
   }

   fp_in = fopen(argv[4], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[4]);
     exit(0);
   }

   strcpy(seq_gen_options,argv[5]);

   // must do since not specified
   (*rand_seed) = (unsigned long) time(NULL);
   init_genrand((*rand_seed));
}

/**************************************************************************/
/*                           Command Line B                               */
/**************************************************************************/
void ProcessCommandLineB(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{

   /************************************************/
   /* (B) argc=4                                   */
   /* nsg_wrap -a <inputfile> "<seq-gen-options>"  */
   /************************************************/
   if (strcmp(argv[1],"-a")!=0)
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case B) is incorrect\n");
     exit(0);
   }

   fp_in = fopen(argv[2], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[2]);
     exit(0);
   }

   strcpy(seq_gen_options,argv[3]);

   // must do since not specified
   (*rand_seed) = (unsigned long) time(NULL);
   init_genrand((*rand_seed));
 
}

/**************************************************************************/
/*                           Command Line C                               */
/**************************************************************************/
void ProcessCommandLineC(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{
   /********************************************************************/
   /* (C) argc=8                                                       */
   /* nsg_wrap -d -l <int> -z <int> <inputfile> <seq-gen-options>      */
   /********************************************************************/
   if (strcmp(argv[2],"-l")!=0)
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case C) is incorrect\n");
     exit(0);
   }

   if (atoi(argv[3]) > MAX_SEQ_LENGTH)
   {
     fprintf(stderr, "NSGW ERROR: sequence length too large; limit is %d\n", MAX_SEQ_LENGTH);
     exit(0);
   }
   else
   {
    (*seq_length) = atoi(argv[3]);
   }

   // the largest value is 4294967295 which is 2^32 - 1
   // the command line arg, format as unsigned long, put into rand_seed's
   // address which we have in this routine as rand_seed because it came
   // in as an address
   sscanf(argv[5], "%lu", rand_seed);
   init_genrand((*rand_seed));

   fp_in = fopen(argv[6], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[6]);
     exit(0);
   }

   strcpy(seq_gen_options,argv[7]);

}

/**************************************************************************/
/*                           Command Line D                               */
/**************************************************************************/
void ProcessCommandLineD(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{
   /*******************************************************/
   /* (D) argc=6                                          */
   /* nsg_wrap -a -z <int> <inputfile> <seq-gen-options>  */
   /*******************************************************/
   if (strcmp(argv[2],"-z")!=0)
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case D) is incorrect\n");
     exit(0);
   }

   // the largest value is 4294967295 which is 2^32 - 1
   // the command line arg, format as unsigned long, put into rand_seed's
   // address which we have as rand_seed since it came into this routine
   // as an address
   sscanf(argv[3], "%lu", rand_seed);
   init_genrand((*rand_seed));

   fp_in = fopen(argv[4], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[4]);
     exit(0);
   }

   strcpy(seq_gen_options,argv[5]);
}

/**************************************************************************/
/*                           Command Line E                               */
/**************************************************************************/
void ProcessCommandLineE(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{
   /**************************************************************************/
   /* (E) argc=8                                                             */
   /* nsg_wrap -E -d -l <int> <inputfile> <rnd_num_test_file> <seq-gen-opts> */
   /**************************************************************************/
   if ( (strcmp(argv[2],"-d")!=0) &&
        (strcmp(argv[3],"-l")!=0) )
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case E) is incorrect\n");
     exit(0);
   }

   if (atoi(argv[4]) > MAX_SEQ_LENGTH)
   {
     fprintf(stderr, "NSGW ERROR: sequence length too large; limit is %d\n", MAX_SEQ_LENGTH);
     exit(0);
   }
   else
   {
    (*seq_length) = atoi(argv[3]);
   }

   fp_in = fopen(argv[5], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[5]);
     exit(0);
   }

   // now do the work for the random number input test file
   rand_test_yn = 1;
   fp_in_rand = fopen(argv[6], "r");

   if (fp_in_rand == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open the random number test file '%s'\n", argv[6]);
     exit(0);
   }

   // allocate (and check) the space for the needed array
   rand_num_test = (double *) calloc((MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES),sizeof(double));

   if (rand_num_test == NULL)
   {
     fprintf(stderr, "NSGW ERROR:  Could not obtain space for rand_num_test array\n");
     exit(0);
   }

   strcpy(seq_gen_options,argv[7]);
}

/**************************************************************************/
/*                           Command Line F                               */
/**************************************************************************/
void ProcessCommandLineF(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{
   /****************************************************************/
   /* (F) argc=6                                                   */
   /* nsg_wrap -F -a <inputfile> <rnd_num_test_file> <seq-gen-opts>*/
   /****************************************************************/
   if(strcmp(argv[2],"-a")!=0) 
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case F) is incorrect\n");
     exit(0);
   }

   fp_in = fopen(argv[3], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[3]);
     exit(0);
   }

   // now do the work for the random number input test file
   rand_test_yn = 1;
   fp_in_rand = fopen(argv[4], "r");

   if (fp_in_rand == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open the random number test file '%s'\n", argv[4]);
     exit(0);
   }

   // allocate (and check) the space for the needed array
   rand_num_test = (double *) calloc((MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES),sizeof(double));

   if (rand_num_test == NULL)
   {
     fprintf(stderr, "NSGW ERROR:  Could not obtain space for rand_num_test array\n");
     exit(0);
   }

   strcpy(seq_gen_options,argv[5]);

}
/**************************************************************************/
/*                           Command Line G                               */
/**************************************************************************/
void ProcessCommandLineG(int argc, char *argv[], char * seq_gen_options, 
		         int * seq_length, unsigned long * rand_seed)
{
   /*****************************************************************/
   /* (G) argc=6                                                    */
   /* nsg_wrap -G -t <inputfile> <rnd_num_test_file> <seq-gen-opts> */
   /*****************************************************************/
   if(strcmp(argv[2],"-t")!=0) 
   {
     fprintf(stderr, "NSGW ERROR:  Command line (case G) is incorrect\n");
     exit(0);
   }

   fp_in = fopen(argv[3], "r");
   if (fp_in == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open input file '%s'\n", argv[3]);
     exit(0);
   }

   // now do the work for the random number input test file (of integers!)
   rand_tree_test_yn = 1;

   fp_in_rand = fopen(argv[4], "r");
   if (fp_in_rand == NULL)
   {
     fprintf(stderr, "NSGW ERROR: Cannot open the TREE random number test file '%s'\n", argv[4]);
    exit(0);
   }

   // allocate (and check) the space for the needed arrays
   rand_num_tree_test = (long unsigned *) calloc((MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES),sizeof(long unsigned));
   rand_num_node_tree_test = (int *) calloc((MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_NODES),sizeof(int));

   if (rand_num_tree_test == NULL)
   {
     fprintf(stderr, "NSGW ERROR:  Could not obtain space for rand_num_tree_test and rand_num_node_tree_test arrays\n");
     exit(0);
   }

   strcpy(seq_gen_options,argv[5]);
}

