/*****************************************************************************/
/*                                                                           */
/*  NSGW kappa version                                                       */
/*  Wrapper code for Seq-Gen in order to request/process sequences between   */
/*  NetGen and Seq-gen.                                                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include "globals.h"
#include "network.h"
#include "twister.h"
#include "interface.h"


/******************************************************************************/
/*                          ReadRequestNode                                   */
/******************************************************************************/
void ReadRequestNode(FILE *fp_in, int max_num_seqs, int max_seq_length, 
		     int *num_of_parents, int *num_of_seqs, int seq_length, 
		     char *parent_sequences_ptr[], double *brlen)
{
  char in_line[255];
  char temp_sequence[MAX_SEQ_LENGTH+2]; //null and newline characters if present
  char junk_char;
  int  junk_int, seq_i;

  /***************************************************************/
  /*         make sure the file has the correct format           */
  /***************************************************************/
  // first line indicates process id from NG (needed for file_id), 
  // number of parents and number of sequences for the node
  
  fgets(in_line, 255, fp_in);

  if (sscanf(in_line, "%d %d %d" , &file_id, num_of_parents, num_of_seqs) != 3)
  {
    fprintf(stderr, "NSGW ERROR: File id OR number of parents OR number of sequences missing from first line of input file\n");
    exit(0);
  }

  // NOTE: in the case of hybrid nodes (ie parents = 2), all the logic
  // is handled in NETGEN or the GetNetwork Sequences routine and all
  // we ever go to SeqGen asking for is a single sequence
  if ( ((*num_of_parents)!=0) && ((*num_of_parents)!=1) )
  {
    fprintf(stderr, "NSGW ERROR: Invalid number of parents (must be 0 or 1): %d \n", *num_of_parents);
    exit(0);
  }

  if ((*num_of_seqs) > max_num_seqs)
  {
    fprintf(stderr, "NSGW ERROR: Too many sequences\n");
    exit(0);
  }

  /***************************************************************/
  /*         now read the file!                                  */
  /***************************************************************/

  /* case of num_of_parents == 0 */
  // if there is no parent we are done reading!

  /* case of num_of_parents == 1 */
  // read in the extra information contained in file when there is
  // one parent specified (we just need the sequences for the parent)
  // (first reserve the memory for the parent sequences though!)
  
  for (seq_i = 0; seq_i < (*num_of_seqs); seq_i++)
  {
    parent_sequences_ptr[seq_i] = (char *) calloc(seq_length+2, sizeof(char));

    if (parent_sequences_ptr[seq_i] == NULL)
    {
      fprintf(stderr, "NSGW ERROR: Could not allocate memory for incoming parent sequences\n");
      exit(0);
    }
  }
  
  //fprintf(stderr, "NSGW DEBUGGING: num of parents is %d\n", *num_of_parents);
  
  if ((*num_of_parents) == 1)
  {
    // the next lines contain the sequences for the parent
    for (seq_i = 0; seq_i < (*num_of_seqs); seq_i++)
    {
      fgets(temp_sequence, MAX_SEQ_LENGTH+2, fp_in);

      // DEBUGGING 
      //fprintf(stderr, "NSGW DEBUGGING: the seq read in is:%s.\n", temp_sequence);
      temp_sequence[seq_length] = '\0'; //put null at end of string
      temp_sequence[seq_length+1] = '\0'; //replace the newline with a null
      strcpy(parent_sequences_ptr[seq_i], temp_sequence);
      
    } // seqs

    // the next 3 lines have data we do not need
    fgets(in_line, 255, fp_in);
    fgets(in_line, 255, fp_in);
    fgets(in_line, 255, fp_in);

    // now we need the branch length -- it comes in as a pointer so
    // don't put the & sign in front here!
    fscanf(fp_in, "%c %d %d %lf" , &junk_char, &junk_int, &junk_int, brlen);

    //fprintf(stderr, "NSGW DEBUGGING: the branch length is %lf\n", (*brlen));
  }

  /***************************************************************/
  /*   be good and close the file since we are now done with it  */
  /***************************************************************/
  fclose(fp_in);
}


/******************************************************************************/
/*                     GenerateSeqGenInputFile                                */
/******************************************************************************/
void GenerateSeqGenInputFile(char *outfile, int num_of_parents, int num_of_seqs,
		             int seq_length, char *parent_sequences_ptr[],
			     char *parent_seq_ptr, double brlen, int seq_index)
{
  FILE *fp_out;
  int have_seq_ptr;
  char file_id_str[11] = "\0";

  if (parent_seq_ptr != NULL)
    have_seq_ptr = 1;
  else
    have_seq_ptr = 0;

  /***************************************************************/
  /*          before starting make the output file name          */
  /*            (but make sure we first clear it out             */
  /*             from any other calls to this routine!)          */
  /***************************************************************/
  strcpy(outfile, "\0");
  strcat(outfile, "sg_temp_in");
  strcat(outfile, "_");
  sprintf(file_id_str, "%d", file_id);
  strcat(outfile, file_id_str);

  /***************************************************************/
  /*              now open the file we will write                */
  /***************************************************************/
  fp_out = fopen(outfile, "w");
  if (fp_out == NULL)
  {
    fprintf(stderr, "NSGW ERROR: Cannot open output file '%s'\n", outfile);
    exit(0);
  }

  /***************************************************************/
  /*     now print the sequences                                 */
  /***************************************************************/
  // get sequence for root case
  if (num_of_parents == 0)
  {
    // we are going to ask SEQGEN to give us sequences for 
    // a single "node" which to it means a parent and 
    // two kids (the smallest SEQGEN does) but we do not care about
    // the kids at all -- we really just want the root 
    // to be randomly generated by SEQGEN in this case
    fprintf(fp_out, "(:0,:0)\n"); 
  }
  if (num_of_parents == 1)
  {
    // we are going to ask SEQGEN to give us sequences for
    // a single node where the parent's sequence is known
    
    // first give number of sequences which will always
    // be 1 since this gets called from a loop in main
    // and also give the sequence length
    fprintf(fp_out, "1 %d\n", seq_length);

    // now we need to tell it the parent sequence
    // default for Seq-Gen appears to be a name in 10 spaces
    // followed by a space and then the string
    //
    // if have_seq_ptr is 1 then we have the actual sequence,
    // if not, then we need to do indexing for the sequence
    // (it all depends if we are in 'a' mode or 'd' mode)
    if (have_seq_ptr == 1)
      fprintf(fp_out, "temp       %s\n", parent_seq_ptr);
    else
      fprintf(fp_out, "temp       %s\n", parent_sequences_ptr[seq_index]);

    
   
    //fprintf(stderr, "NSGW DEBUGGING: From making the SG input file parent sequence:%s, and with hchromo_index:%d and strand_index:%d\n", parent_ptr->hom_chromos[hchromo_index]->strands[strand_i]->sequence, hchromo_index, strand_index);

    // now print the number of trees which is always 1 for us
    fprintf(fp_out, "1\n");
    fprintf(fp_out, "(:%lf,:0)\n", brlen); 
  }

  /***************************************************************/
  /*     close the file so it will be readable by SEQGEN         */
  /***************************************************************/
  fclose(fp_out);

  // DEBUGGING
  // if (seq_index == -1)
  //   fprintf(stderr, "NSGW DEBUGGING:  At the end of making the input file seq index is%d and the parent sequence we used for making the file was %s\n", seq_index, parent_seq_ptr);
  // else
  //  fprintf(stderr, "NSGW DEBUGGING:  At the end of making the input file seq%d:%s\n", seq_index, parent_sequences_ptr[seq_index]);
  // END DEBUGGING

}


/******************************************************************************/
/*                             GetSeqGenSequence                              */
/******************************************************************************/
void GetSeqGenSequence(char *seq_gen_input_file_name, char seq_gen_options[],
		       int num_of_parents, int seq_length, 
		       char *node_sequences_ptr[], int seq_index, 
		       char *node_seq_ptr, int re_index_yn, 
		       int outgroup_node_yn)
{
  char SG_command[500];
  char exec_SG[500];
  FILE *pipe_ptr;
  int rm_error;


  /*********************************************************/
  /*  first set-up the command line we will need based on  */
  /*  number of parents and whether we are testing or not  */
  /*********************************************************/

  ChooseSGCommand(SG_command, num_of_parents);


  /*****************************************************/
  /*  now piece together the executable call we need   */
  /*****************************************************/

  MakeCall(exec_SG, SG_command, seq_gen_options, seq_gen_input_file_name,
	   num_of_parents, seq_length, seq_index, re_index_yn, 
	   outgroup_node_yn);



  /**********************************************************/
  /*     now do the call and process the response           */
  /**********************************************************/

  // open the pipe and do work
  if ((pipe_ptr = popen(exec_SG, "r")) == NULL)
  {
    fprintf(stderr, "NSGW ERROR: There was a problem with the pipe call to Seq-Gen\n");
    exit(0);
  }
  else
  {

    // read back the sequences and store them where they belong
    ProcessSGResponse(pipe_ptr, num_of_parents, seq_length, node_sequences_ptr, 
		      seq_index, node_seq_ptr);

    // close the execution pipe
    (void) pclose(pipe_ptr);

    // note:  there may be times when we may want to keep this file around,
    //         but normally we want to discard it
    //
    // remove the input file to SG since we are done with it
    //
    rm_error = remove(seq_gen_input_file_name);
    
    if (rm_error != 0)
    {
      fprintf(stderr, "NSGW ERROR: There was a problem removing the input file for Seq-Gen\n");
      fprintf(stderr, "NSGW ERROR: program has therefore terminated\n");
      exit(0);
    }
  }
}


/******************************************************************************/
/*                           ChooseSGCommand                                  */
/******************************************************************************/
void ChooseSGCommand(char *SG_command, int num_of_parents)
{

  // this routine sets the command line for Seq-Gen depending on which
  // version of it we need
  
  // remember this testing is for NSGW-A vs SG which requires a modified
  // version of SG
  if (rand_test_yn) 
  {
    if (num_of_parents == 0)
    {
      strcpy(SG_command, 
	     "./rn_seq-gen -wa -z");
    }
    else
    {
      strcpy(SG_command,
	     "./rn_seq-gen -wa -k1 -z"); 
    }
  }

  // regular versions
  if (!rand_test_yn) 
  {
    if (num_of_parents == 0)
    {
      strcpy(SG_command, 
             "./seq-gen -wa -z"); 

    }
    else
    {
      strcpy(SG_command,
             "./seq-gen -wa -k1 -z");
    }
  }
}


/******************************************************************************/
/*                             MakeCall                                       */
/******************************************************************************/
void MakeCall(char *exec_SG, char *SG_command, char seq_gen_options[], 
	      char *seq_gen_input_file_name, int num_of_parents, 
	      int seq_length, int seq_index, int re_index_yn, 
	      int outgroup_node_yn)
{
  int begin_part_size, option_part_size, end_part_size;
  int seq_i;
  char file_id_str[11] = "\0";
  char rand_num_str[11] = "\0";
  char seq_length_str[11] = "\0";
  long unsigned new_rand;

  // check size limits (strlen does NOT count the terminating null if present)
  
  begin_part_size = strlen(SG_command);

  option_part_size = strlen(seq_gen_options);

  //fprintf(stderr, "NSGW DEBUGGING:  our seq_gen_option string is:%s\n",seq_gen_options);

  // add in redirection stuff that we know we will have
  end_part_size = strlen(seq_gen_input_file_name) + 3 + 26;

  if ( (begin_part_size + option_part_size + end_part_size + 1) > 500 )
  {
    fprintf(stderr, "NSGW ERROR: Seq-Gen command line size exceeded\n");
    exit(0);
  }

  // make the execution line we need 
  strcpy(exec_SG, SG_command);

  // RNG work
  //
  // NOTE:  if testing with prescribed random numbers (namely using the modified
  //        version of SG which reads in a list of doubles), this number needs
  //        to be the rand_num_index and pass it as if it were the random number
  //        seed -- the rand test version of SEQGEN will then use it 
  //        appropriately (i.e. SEQGEN should not be restarting with
  //        rand_test_index = 0 each time it is called from here under this
  //        scenario); this scenario is indicated by rand_test_yn=1
  //        
  //        if we are running the tree test (rand_tree_test_yn=1) we want to use
  //        a value from the input file as a seed, and then use the subsequent
  //        value for the call (this works because the input file contains
  //        comparison seeds which were given to NSGW(d) when NG was run and 
  //        the input file of seeds made)
  //
  //        if we are not either of the above tests, we will create this integer
  //        and it will be repeatable, given we started the NSG_wrapper with 
  //        its same own seed, which will dictate what this value ends up being
  //        (NSGW seeds can be tracked in NG by including track_nsgw_seeds in 
  //        its input file)

  //fprintf(stderr, "NSGW DEBUGGING:  about to do RNG STUFF\n");

  // if no validation testing is going on -- this is what we want
  if ( (rand_test_yn == 0) && (rand_tree_test_yn == 0) )
  {
    new_rand = genrand_int32(); 

    // DEBUGGING
    //fprintf(stderr, "NSGW DEBUGGING:  just got a new number off the stream and the new number is %lu\n", new_rand);
    
    sprintf(rand_num_str, "%lu", new_rand);
    strcat(exec_SG, rand_num_str);
  }

  // if the test of NSGW vs SG is happening this is rand_test_yn==1
  // and we want the value on the command line here to be the index value
  // to tell MODIFIED SG what index into its rng array to use
  if (rand_test_yn == 1)
  {
    sprintf(rand_num_str, "%d", rand_num_test_index);
    strcat(exec_SG, rand_num_str);
  }

  // if rand_num_tree_test==1 is happening, we are comparing NG with NSGW(a)
  // and we want to be seeding with the integer from the input file
  // so that we end up giving the same rng seed to the regular SG as
  // it happened when NG was run (however, remember that NSGW(d) is called
  // once for each node, not each sequence so must be careful on how
  // we advance the index and rngs here -- use re_index_yn to know if we
  // should advance the index -- basically do not want to when starting off
  // with root, otherwise the FIRST sequence for each node we want to,
  // but for an outgroup we want to advance index for each try -- so before
  // coming into this routine, it is determined whether to re-index or not
  // and test this value here (also note that the root is a special case
  // and the seed needs to be initialized only, no pointer advancing; 
  // another special case for advancing is when we have the outgroup
  // node and we must advance the rngs like NG -- no re_index, but
  // an eaten rng for each sequence not looked at here)
  //
  // after note:  it looks like we do send the re_index flag for outgroup
  //              work when testing, so we do re_index, but the key thing
  //              is to "eat" the correct number of rngs (one for each
  //              sequence not looked at here)
  //
  if (rand_tree_test_yn == 1)
  {
    if (num_of_parents == 0)
    {
      init_genrand(rand_num_tree_test[rand_num_test_index]);

      // DEBUGGING
      //fprintf(stderr, "NSGW DEBUG:  just reinitialized random number stream with %lu\n", rand_num_tree_test[rand_num_test_index]);
      // END DEBUGGING

    }

    if (re_index_yn == 1)
    {
      rand_num_test_index = rand_num_test_index + 1;

      // DEBUGGING
      //fprintf(stderr, "NSGW DEBUG:  just advanced the index by 1, current index is now %d and rng seed is now %lu\n", rand_num_test_index, rand_num_tree_test[rand_num_test_index]);
      // END DEBUGGING

      init_genrand(rand_num_tree_test[rand_num_test_index]);

      // DEBUGGING
      //fprintf(stderr, "NSGW DEBUG:  just reinitialized random number stream with %lu\n", rand_num_tree_test[rand_num_test_index]);
      // END DEBUGGING
    }

    if (outgroup_node_yn == 1)
    {
      for (seq_i = 0; seq_i < seq_index; seq_i++)
      {
        genrand_int32();
      }
    }

    new_rand = genrand_int32();

    // DEBUGGING
    //fprintf(stderr, "NSGW DEBUG:  just got a new number off the stream and the new number is %lu\n", new_rand);
    // END DEBUGGING
    
    sprintf(rand_num_str, "%lu", new_rand);
    strcat(exec_SG, rand_num_str);
  }

  // now do the sequence length
  strcat(exec_SG, " -l");
  sprintf(seq_length_str, "%d ", seq_length);
  strcat(exec_SG, seq_length_str);

  // now do any other options the user requested and they better
  // be legit! (we do not sanity check here)
  strcat(exec_SG, seq_gen_options);

  // now do the input file
  strcat(exec_SG, " < ");
  strcat(exec_SG, seq_gen_input_file_name);

  // redirect the standerror output from SeqGen just in case we need it,
  // (but go ahead and hardcode the name so it is overwritten each time)
  // no longer doing as an append because it takes too much disk space
  // when doing so many runs
  strcat(exec_SG, " 2> se_out_seq-gen");
  strcat(exec_SG, "_");
  sprintf(file_id_str, "%d", file_id);
  strcat(exec_SG, file_id_str);

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUG: Our command line to Seq-Gen is:%s\n",exec_SG);
  // END DEBUGGING

}


/******************************************************************************/
/*                         ProcessSGResponse                                  */
/******************************************************************************/
void ProcessSGResponse(FILE *pipe_ptr, int num_of_parents, int seq_length, 
		       char *node_sequences_ptr[], int seq_index,
		       char *node_seq_ptr)
{
  char junk_string[MAX_SEQ_LENGTH+11];  // allow for node name and misc chars
  char temp_sequence[MAX_SEQ_LENGTH+2]; //allow for null and newline characters
  int junk_int;
  int have_seq_ptr;

  if (node_seq_ptr != NULL)
    have_seq_ptr = 1;
  else
    have_seq_ptr = 0;

  // we never need the first line of data from Seq-Gen
  fgets(junk_string, (MAX_SEQ_LENGTH+11), pipe_ptr);

  // DEBUGGING
  fprintf(stderr, "NSGW DEBUG: our first line back from SG is %s\n", junk_string);
  // END DEBUGGING

  // process no ancestor case
  if (num_of_parents == 0)
  {
    // we need the second line of data and nothing else for this case
    // the first thing on the line is the node number and then a tab
    fscanf(pipe_ptr, "%d\t", &junk_int);

    // DEBUGGING
    fprintf(stderr, "NSGW DEBUG: our temp seq before going to SG is %s\n", temp_sequence);
    // END DEBUGGING

    // now get the ancestor sequence and the null, and the newline if there
    fgets(temp_sequence, MAX_SEQ_LENGTH+2, pipe_ptr);

    // DEBUGGING
    fprintf(stderr, "NSGW DEBUG: our temp seq back from SG is %s\n", temp_sequence);
    // END DEBUGGING

    // DEBUGGING
    fprintf(stderr, "NSGW DEBUG: seq_length + 1 is %d\n", seq_length+1);
    fprintf(stderr, "NSGW DEBUG: sizeof(char) is %d\n", sizeof(char));
    fprintf(stderr, "NSGW DEBUG: last char of seq is %c\n", temp_sequence[seq_length-1]);
    if (temp_sequence[seq_length] == '\0')
      fprintf(stderr, "NSGW DEBUG: the first character after the string is a null\n");
    if (temp_sequence[seq_length] == '\n')
      fprintf(stderr, "NSGW DEBUG: the first character after the string is a nl\n");
    // END DEBUGGING

    // get the string into its home in the sequence array
    // NOTE: it appears from testing the output of SEQ-GEN is a bunch
    // of characters (for the bases) followed only by a newline
    // there is no null character which terminates the string
      
    temp_sequence[seq_length] = '\0';   //replace the newline with a null
    temp_sequence[seq_length+1] = '\0'; //just to be safe
    if (have_seq_ptr == 1)
      strcpy(node_seq_ptr, temp_sequence); 
    else
      strcpy(node_sequences_ptr[seq_index], temp_sequence); 
      

    // we do not care about the next two lines in this case
    fgets(junk_string, (MAX_SEQ_LENGTH+11), pipe_ptr);
    fgets(junk_string, (MAX_SEQ_LENGTH+11), pipe_ptr);


    // make sure we advance the rand_num_test_index appropriately

    // if we are testing with prescribed random numbers into the modified
    // version of Seq-Gen -- it should have eaten up exactly the number
    // of bases in random number usage
    if (rand_test_yn)
      rand_num_test_index = rand_num_test_index + seq_length;

    // if we are doing the tree test, the index is incremented as needed
    // (when seq_index == 0) above in the call generation part of this
    // routine

  } // end of num_of_parents == 0

  // now deal with the single parent case
  if (num_of_parents == 1)
  {
    // we do not need the second line in this case because it has
    // the ancestor/root sequence we gave it
    fgets(junk_string, (MAX_SEQ_LENGTH+11), pipe_ptr);

    // DEBUGGING
    fprintf(stderr, "NSGW DEBUG: our temp seq before going to SG is %s\n", temp_sequence);
    // END DEBUGGING

    // we do want the next line which is the sequence of the node
    // we requested and since we did not give it a name there should
    // simply be a single space before the string even though in
    // printing it appears like there is a tab, but there really is not
    // NOTE that this is different from the parents == 0 case where there
    // is no name/space
    fgets(temp_sequence, (MAX_SEQ_LENGTH+11), pipe_ptr);
      
    // DEBUGGING
    fprintf(stderr, "NSGW DEBUG: Our seq back from SG is:%s.\n", temp_sequence);
    // END DEBUGGING

    temp_sequence[seq_length+1] = '\0';   //replace the newline with a null
    temp_sequence[seq_length+2] = '\0';   //just to be safe
    if (have_seq_ptr == 1)
      strcpy(node_seq_ptr, &(temp_sequence[1]));
    else
      strcpy(node_sequences_ptr[seq_index], &(temp_sequence[1]));

    // DEBUGGING
    //fprintf(stderr, "NSGW DEBUG: our temp_sequence is:%s\n", temp_sequence);
    //fprintf(stderr, "NSGW DEBUG: our node's seq is:%s\n", node_seq_ptr);
    // END DEBUGGING
  
    // we do not care about the last line in this case
    fgets(junk_string, (MAX_SEQ_LENGTH+11), pipe_ptr);

    // make sure we advance the rand_num_test_index appropriately

    // if we are testing with prescribed random numbers into the modified
    // version of Seq-Gen -- it should have eaten up exactly the number
    // of bases in random number usage
    if (rand_test_yn)
      rand_num_test_index = rand_num_test_index + seq_length;

    // if we are doing the tree test, the index is incremented as needed
    // (when seq_index == 0) above in the call generation part of this
    // routine
    
  } // end of num_of_parents == 1

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUG: at end of process sg response\n"); 
  //fprintf(stderr, "NSGW DEBUG: our junk string is %s\n", junk_string);
  //fprintf(stderr, "NSGW DEBUG: our junk int is %d\n", junk_int);
  //fprintf(stderr, "NSGW DEBUG: our temp_sequence is:%s\n", temp_sequence);
  // END DEBUGGING
}


/******************************************************************************/
/*                           PrintRequestedSequences                          */
/******************************************************************************/
void PrintRequestedSequences(int num_of_seqs, char *node_sequences_ptr[])
{
  int seq_i;

  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    fprintf(stdout, "%s\n", node_sequences_ptr[seq_i]);
  }
}
