/*****************************************************************************/
/*                                                                           */
/*  NSGW kappa version                                                       */
/*  Wrapper code for Seq-Gen in order to request/process sequences between   */
/*  NetGen and Seq-gen.                                                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/


/******************************************************************************/
/*                                 NSG_wrapper                                */
/*                                                                            */
/*  This code is written to be a wrapper in order to facilitate using         */
/*  Seq-Gen code from Oxford to generate sequences for our networks           */
/*  This routine is called either:                                            */
/*                  DURING the generation of a network from NETGEN            */
/*                  AFTER the generation of a newtork by NETGEN               */
/*  as determined by the incoming flags -d or -a                              */
/*                                                                            */
/*  NOTE:                                                                     */
/*   the command line is very restrictive see the code for the only possible  */
/*       legitimate exec lines                                                */
/*                                                                            */
/*                                                                            */
/*  COMMAND LINE GUIDELINES:                                                  */
/*                                                                            */
/*   * the length of the sequence must always be specified for "during"       */
/*     cases (for "after" cases it is read in for each node)                  */
/*                                                                            */
/*   * an input file is always specified                                      */
/*                                                                            */
/*   * the output file is always STDOUT (to facilitate piping)                */
/*     (so if running independently redirect the output with "> resultfile")  */
/*                                                                            */
/*   * a random number seed is optional but must be the last arg before       */
/*     the input file if present                                              */
/*                                                                            */
/*   * must always give the Seq-Gen options as the last arg, even if nothing  */
/*     must specify, and must use " " or ' '                                  */
/*     (e.g. "-m <model> -a <alpha>"  or '-m <model> - a <alpha>'             */
/*      or ' ' or " ")                                                        */
/*                                                                            */
/*   * the args to Seq-Gen must be legitimate per their manual and NOT        */
/*     interfere with the expectations of the piping (e.g. can specify things */
/*     like models and parameters but not request multiple trees or certain   */
/*     output formats Seq-Gen can generate)                                   */

// University of New Mexico
// Monique M. Morin


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include "globals.h"
#include "network.h"
#include "interface.h"
#include "input.h"
#include "output.h"
#include "sequences.h"
#include "twister.h"


/*****************************************************************************/
/*                              main                                         */
/*****************************************************************************/

int main(int argc, char *argv[])
{
  char seq_gen_options[100];
  int  seq_length;
  unsigned long rand_seed = 0;

  NETWORK NT;
  char option;
  char outfile[300] = "\0";  //dealt with in GenerateSeqGenInputFile

  int num_of_parents, num_of_seqs, seq_i, node_i;
  int hchromo_i, strand_i, nhc, nsphc;
  double brlen;
  char *node_sequences_ptr[MAX_NUM_HOM_CHROMOS * MAX_NUM_STRANDS_PER_HC];
  char *parent_sequences_ptr[MAX_NUM_HOM_CHROMOS * MAX_NUM_STRANDS_PER_HC];

  char seq_i_str[10];


  /***********************************/
  /*      process the command line   */
  /***********************************/
  ProcessCommandLine(argc, argv, &seq_gen_options[0], &seq_length, &rand_seed);

  //fprintf(stderr, "NSGW DEBUG:  Our arg count is %d\n", argc);
  //fprintf(stderr, "NSGW DEBUG:  Our seq_gen_options string is %s\n", &seq_gen_options[0]);

  /**********************************/
  /*  set-up of mode/run type       */
  /**********************************/

  if (strcmp(argv[1],"-d")==0)
    option = 'd';
  else if (strcmp(argv[1],"-a")==0)
    option = 'a';
  else if (strcmp(argv[1],"-E")==0)
    option = 'd';
  else if (strcmp(argv[1],"-F")==0)
    option = 'a';
  else if (strcmp(argv[1],"-G")==0)
    option = 't';


  /*********************************/
  /*   misc debugging statements   */
  /*********************************/
  
  //fprintf(stderr, "NSGW DEBUG: seed is:%10lu\n", rand_seed);


  /*********************************/
  /* random number setup if needed */
  /*********************************/

  if (rand_tree_test_yn || rand_test_yn)
    RandomNumberSetUp();


  /********************************/
  /********************************/
  /*                              */
  /*  MAIN PROCESSING WORK!!!!!!! */
  /*                              */
  /********************************/
  /********************************/
  
  /*****************************************************************/
  /*     deal with the "during" option here -- where we            */
  /*     want sequences during the topology generation             */
  /*     as it is being done by NETGEN                             */
  /*****************************************************************/
  if ( option == 'd')
  {
    //fprintf(stderr, "NSGW DEBUG: JUST STARTED NSGW with 'd' option (seed is %lu)\n", rand_seed);

    // Read input file
    ReadRequestNode(fp_in, (MAX_NUM_HOM_CHROMOS * MAX_NUM_STRANDS_PER_HC), 
		    MAX_SEQ_LENGTH, &num_of_parents, 
		    &num_of_seqs, seq_length, parent_sequences_ptr, &brlen);

    //fprintf(stderr, "NSGW DEBUG: seq_length from main is: %d\n", seq_length);
    //fprintf(stderr, "NSGW DEBUG: num_of_parents from main is: %d\n", num_of_parents);

    // now reserve the space for the node_sequences_ptr which we need to pass
    // to GetSeqGenSequence since we do not have a real node (with
    // homologous chromosomes and strands) made under this interactive mode
    for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
    {
      node_sequences_ptr[seq_i] = (char *) calloc(seq_length+2, sizeof(char));

      if (node_sequences_ptr[seq_i] == NULL)
      {
        fprintf(stderr, "NSGW ERROR:  Could not allocate memory for node sequences we will get back from SG\n");
	exit(0);
      }

      strcpy(node_sequences_ptr[seq_i], "init seq ");
      sprintf(seq_i_str, "%d",seq_i);
      strcat(node_sequences_ptr[seq_i], seq_i_str);
    }

    // this loop must be executed for each sequence we want from SEQ-GEN
    for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
    {
      // Generate the file needed for SEQ-GEN (we have the array, not a
      // specific sequence so pass NULL for the 6th arg)
      GenerateSeqGenInputFile(outfile, num_of_parents, num_of_seqs, seq_length, 
	    	              parent_sequences_ptr, NULL, brlen, seq_i); 

      // Request and Process Reply of sequence from SEQ-GEN
      // (note: under this mode re-indexing for test "G" not possible
      //        and do not distinguish outgroups, so last arg two args are 0's)
      GetSeqGenSequence(outfile, seq_gen_options, num_of_parents, seq_length,
		        node_sequences_ptr, seq_i, NULL, 0, 0);
    }

    // Report to STDOUT the info needed by NETGEN
    PrintRequestedSequences(num_of_seqs, node_sequences_ptr);


    // Be good and clean up memory
    if (num_of_parents == 0)
    {
      for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
      {
        free (node_sequences_ptr[seq_i]);
      }
    }
    if (num_of_parents > 0)
    {
      for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
      {
        free (parent_sequences_ptr[seq_i]);
        free (node_sequences_ptr[seq_i]);
      }
    }
  }

  /*****************************************************************/
  /*     deal with the "after" option here -- where we             */
  /*     want sequences AFTER the topology generation              */
  /*                                                               */
  /*     also if doing testing for tree comparisons between        */
  /*     NG and NSGW(a) (rand_tree_test_yn) we follow this route   */
  /*****************************************************************/
  if ( (option == 'a') || (option == 't'))
  {
    // set the file_id once from the process id
    file_id = getpid();

    // make the network -- just get the basic space and set
    // default values; ReadNetwork will take care of actually 
    // setting the real values and making all the nodes
    NT = NewNetwork();

    // read in the network
    ReadNetwork(NT);

    // work through the network and get sequences for each node
    GetNetworkSequences(outfile, seq_gen_options, NT);
    
    // print the network
    PrintNetwork(stdout, NT, rand_seed);

    // free the nodes (but first their homologous chromosomes/strands) 
    // and finally the network itself
    for (node_i = 0; node_i < NT->num_of_nodes; node_i++)
    {
      nhc = NT->ndarray[node_i]->num_of_hom_chromos;
      nsphc = NT->ndarray[node_i]->num_of_strands_per_hc;

      for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
      {
        for (strand_i = 0; strand_i < nsphc ; strand_i++)
        {
          free(NT->ndarray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
	}
      }

      free(NT->ndarray[node_i]);

      //fprintf(stderr, "NSGW DEBUGGING: done releasing node %d\n", node_i);
    }

    // now we can free the network
    free(NT);
  }
  
  /**********************************************************/
  /* clean up dynamically allocated arrays if we used them! */
  /**********************************************************/
  if (rand_test_yn)
    free(rand_num_test);

  if (rand_tree_test_yn)
    free(rand_num_tree_test);


  /*****************************/
  /* be good and return nicely */
  /*****************************/
  return 0;

}

