/*****************************************************************************/
/*                                                                           */
/*  NSGW kappa version                                                       */
/*  Wrapper code for Seq-Gen in order to request/process sequences between   */
/*  NetGen and Seq-gen.                                                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>

#include "globals.h"
#include "network.h"
#include "utilities.h"
#include "output.h"


/*****************************************************************************/
/*                            PrintNetwork                                   */
/*****************************************************************************/
void PrintNetwork(FILE *fp, NETWORK nt, unsigned long rand_seed)
{
  int hchromo_i, strand_i;

  ResetProcessed(nt);

  fprintf(fp,"c Output Report for NSGW\n");
  fprintf(fp,"c \n");
  fprintf(fp,"c The rand_seed for this run was: %lu\n", rand_seed);

  if (rand_tree_test_yn)
    fprintf(fp, "c This was a rand_test_tree run\n");

  if (rand_test_yn)
    fprintf(fp, "c This was a rand_test run\n");

  fprintf(fp,"c \n");

  if (nt->outgroup_id > -1)
  {
    fprintf(fp,"c There was an outgroup present -- its leaf is node: %d\n", nt->outgroup_id);
    fprintf(fp,"c The outgroup params --\n");
    fprintf(fp,"c min_outg_diff_value, max_outg_diff_value, max_outg_tries -- were:\n");
    fprintf(fp,"c %d %d %d \n", nt->min_outg_diff_value, nt->max_outg_diff_value, nt->max_outg_tries);
  }

  fprintf(fp,"c \n");
  fprintf(fp,"c \n");
  fprintf(fp,"c Root and internal nodes are:\n");
  fprintf(fp,"c For each node we list the:\n");
  fprintf(fp,"c   type, id, hybrid status, #hchromosomes, #strands per homologous chromosome,\n"); 
  fprintf(fp,"c   seq length, and sequences\n");
  fprintf(fp,"c    \n");
  fprintf(fp,"c       r = root node\n");
  fprintf(fp,"c       n = internal node\n");
  fprintf(fp,"c \n");

  fprintf(fp,"r %ld %d %d %d %d\n", nt->root->id, nt->root->hybrid, nt->root->num_of_hom_chromos, nt->root->num_of_strands_per_hc, nt->root->seq_length);

  for(hchromo_i = 0; hchromo_i < nt->root->num_of_hom_chromos; hchromo_i++)
  {
    for(strand_i = 0; strand_i < nt->root->num_of_strands_per_hc; strand_i++)
    {
      fprintf(fp,"%s\n",nt->root->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    }
  }

  nt->root->processed_yn = 1; // to avoid reprinting

  PrintInternalNodes(fp, nt);
  fprintf(fp,"c \n");
  fprintf(fp,"c \n");

  ResetProcessed(nt);
  fprintf(fp,"c The leaves are: \n");
  fprintf(fp,"c       e = extant taxa\n");
  fprintf(fp,"c       l = non-extant taxa\n");
  fprintf(fp,"c \n");
  PrintLeaves(fp, nt);
  fprintf(fp,"c \n");

  ResetProcessed(nt);
  fprintf(fp,"c For reference the edges are: \n");
  PrintEdges(fp, nt);

  ResetProcessed(nt);


  // this extra printing is done for making comparisons to other
  // output easier -- we print all the nodes (internal and leaves)
  // in order with their sequences (the lines start with a "c" to
  // indicate these are comment lines, in case someone wants to
  // read this file)

  fprintf(fp,"c \n");
  fprintf(fp,"c The node listing in sequential order is: \n");
  fprintf(fp,"c \n");
  PrintAllNodes(fp, nt);


  // print the modified newick version
  ResetProcessed(nt);

  fprintf(fp,"c \n");
  fprintf(fp,"c Modified Newick Report -- evolutionary branch lengths \n");
  fprintf(fp,"c");
  PrintNewickNetwork(fp, nt->root);
  fprintf(fp,");\n");

  ResetProcessed(nt);


}

/*****************************************************************************/
/*                            PrintInternalNodes                             */
/*****************************************************************************/
void PrintInternalNodes(FILE *fp, NETWORK nt)
{
  NODE nd;

  int node_i;
  int hchromo_i;
  int strand_i;

  for (node_i = 0; node_i < nt->num_of_nodes; node_i++)
  {
    if (!(nt->ndarray[node_i]->leaf))
    {
      nd = nt->ndarray[node_i];

      if (nd != nt->root)
      {

        fprintf(fp,"n %ld %d %d %d %d\n", nd->id, nd->hybrid, nd->num_of_hom_chromos, nd->num_of_strands_per_hc, nd->seq_length);

        for(hchromo_i = 0; hchromo_i < nd->num_of_hom_chromos; hchromo_i++)
        {
          for(strand_i = 0; 
	      strand_i < nd->num_of_strands_per_hc; 
	      strand_i++)
          {
            fprintf(fp,"%s\n",nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
	  }
        }
      }
    }
  }
}

/*****************************************************************************/
/*                            PrintLeaves                                    */
/*****************************************************************************/
void PrintLeaves(FILE *fp, NETWORK nt)
{
  int node_i, hchromo_i, strand_i;

  for (node_i = 0; node_i < nt->num_of_nodes; node_i++)
  {
    if (nt->ndarray[node_i]->leaf == 1)
    {
      if (nt->ndarray[node_i]->extant == 1)
      {
        fprintf(fp,"e %ld %d %d %d %d\n",nt->ndarray[node_i]->id,nt->ndarray[node_i]->extant, nt->ndarray[node_i]->num_of_hom_chromos, nt->ndarray[node_i]->num_of_strands_per_hc, nt->ndarray[node_i]->seq_length);
      }
      else
      {
        fprintf(fp,"l %ld %d %d %d %d\n",nt->ndarray[node_i]->id,nt->ndarray[node_i]->extant, nt->ndarray[node_i]->num_of_hom_chromos, nt->ndarray[node_i]->num_of_strands_per_hc, nt->ndarray[node_i]->seq_length);
      }

      for(hchromo_i = 0; 
          hchromo_i < nt->ndarray[node_i]->num_of_hom_chromos; 
	  hchromo_i++)
      {
        for(strand_i = 0; 
	    strand_i < nt->ndarray[node_i]->num_of_strands_per_hc; 
	    strand_i++)
        {
          fprintf(fp,"%s\n", nt->ndarray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
	}
      }

      // this really does not matter as leaves are designated leaves
      // but for consistency mark it!
      nt->ndarray[node_i]->processed_yn = 1; 
    }

  }
}

/*****************************************************************************/
/*                            PrintEdges                                     */
/*****************************************************************************/
void PrintEdges(FILE *fp, NETWORK nt)
{
  NODE p_node;
  NODE c_node;
  int node_i;

  fprintf(fp, "c \n");
  fprintf(fp, "c \n");
  fprintf(fp, "c The following are all the edges of the network\n");
  fprintf(fp, "c \n");

  fprintf(fp, "c For each edge we list the:\n");
  fprintf(fp, "c      type, parent, child, evolutionary branch length\n");

  fprintf(fp, "c \n");
  fprintf(fp, "c Types are:\n");
  fprintf(fp, "c     t = \"tip\" (edge from internal node to extant taxon),\n");
  fprintf(fp, "c     b = all other edges (between two internal nodes, or internal node and\n");
  fprintf(fp, "c         non-extant leaf)\n");
  fprintf(fp, "c \n");

  /***************************/
  /*       print edges       */
  /***************************/
  for (node_i = 0; node_i < nt->num_of_nodes; node_i++)
  {
    // print the edges associated with this node
    p_node = nt->ndarray[node_i];

    // first print the edge from it to the left child
    c_node = nt->ndarray[node_i]->left;

    if (c_node != NULL)
    {
      if (c_node->extant)
        fprintf(fp, "t ");
      else
        fprintf(fp, "b ");

      if (c_node->parent1 == p_node)
      {
        fprintf(fp, "%ld %ld %lf\n", p_node->id, c_node->id, c_node->brlen1);
      }
      else
      {
        fprintf(fp, "%ld %ld %lf\n", p_node->id, c_node->id, c_node->brlen2);
      }
    }
    
    // now print the edge from it to the right child
    c_node = nt->ndarray[node_i]->right;

    if (c_node != NULL)
    {
      if (c_node->extant)
        fprintf(fp, "t ");
      else
        fprintf(fp, "b ");

      if (c_node->parent1 == p_node)
      {
        fprintf(fp, "%ld %ld %lf\n", p_node->id, c_node->id, c_node->brlen1);
      }
      else
      {
        fprintf(fp, "%ld %ld %lf\n", p_node->id, c_node->id, c_node->brlen2);
      }

    }
  }
}


/*****************************************************************************/
/*                            PrintAllNodes                                  */
/*****************************************************************************/
void PrintAllNodes(FILE *fp, NETWORK nt)
{
  int node_i, hchromo_i, strand_i;
  int nhc, nsphc;

  for (node_i = 0; node_i < nt->num_of_nodes; node_i++)
  {

    fprintf(fp, "c l %ld %d %d %d\n", 
            nt->ndarray[node_i]->id, nt->ndarray[node_i]->hybrid, nt->ndarray[node_i]->num_of_hom_chromos, nt->ndarray[node_i]->num_of_strands_per_hc);

    nhc = nt->ndarray[node_i]->num_of_hom_chromos;
    nsphc = nt->ndarray[node_i]->num_of_strands_per_hc;

    for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
    {
      for (strand_i = 0; strand_i < nsphc; strand_i++)
      {
        fprintf(fp, "c %s\n",
		nt->ndarray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      }
    }

  }
}

/******************************************************************************/
/*                        PrintNewickNetwork                                  */
/******************************************************************************/
void PrintNewickNetwork(FILE *fp, NODE nd)
{
  /* this routine follows a postfix approach */
  /* the last ); must be put in by the calling routine */

  if (nd != NULL)
  {
    /**********************************/
    /*         left subtree           */
    /**********************************/
    if ( (nd->left != NULL) && (nd->processed_yn == 0) )
    {
      fprintf(fp, "(");
      PrintNewickNetwork(fp, nd->left);
    }
    // if we have a true left leaf node, or a hybrid node acting 
    // as a left leaf because it has been printed once already, 
    // we want to print out the "leaf"
    else if ( ((nd->leaf == 1) && (nd->parent1->left == nd))||
	      ((nd->hybrid != 0) && (nd->processed_yn == 1) &&
	         ((nd->parent1->left == nd) ||
		  (nd->parent2->left == nd))) )
    {
      if (nd->hybrid != 0)
      {
        fprintf(fp, "%ld#H:%f", nd->id, nd->brlen1);
      }
      else
      {
	// for a true leaf, label it extant (#E) if it is or if it is
	// the outgroup leaf, along with the correct type of branch length
	
        if (nd->extant == 1 || nd->outgroup == 1)
          fprintf(fp, "%ld#E:%f", nd->id, nd->brlen1);
        else
          fprintf(fp, "%ld:%f", nd->id, nd->brlen1);
      }

      // mark this leaf/node as printed
      // (this marker goes up to 2 hybrid nodes get printed twice)
      nd->processed_yn = nd->processed_yn + 1;
    }


    /**********************************/
    /*         right subtree          */
    /**********************************/
    if ( (nd->right != NULL) && (nd->processed_yn == 0) )
    {
      fprintf(fp, ",");
      PrintNewickNetwork(fp, nd->right);
    }
    // if we have a true right leaf node, or a hybrid node acting 
    // as a right leaf because it has been printed once already, 
    // we want to print out the "leaf"
    else if ( ((nd->leaf == 1) &&(nd->parent1->right == nd))||
	      ((nd->hybrid != 0) && (nd->processed_yn == 1) &&
	         ((nd->parent1->right == nd) ||
		  (nd->parent2->right == nd))) )
    {
      if (nd->hybrid != 0)
      {
        fprintf(fp, "%ld#H:%f", nd->id, nd->brlen1);
      }
      else
      {
	// for a true leaf, label it extant (#E) if it is or if it is
	// the outgroup leaf, along with the correct type of branch length
	
        if (nd->extant == 1 || nd->outgroup == 1)
          fprintf(fp, "%ld#E:%f", nd->id, nd->brlen1);
        else
          fprintf(fp, "%ld:%f", nd->id, nd->brlen1);
      }

      // mark this leaf/node as printed
      // (this marker goes up to 2 hybrid nodes get printed twice)
      nd->processed_yn = nd->processed_yn + 1;
    }


    /**********************************/
    /*             root               */
    /**********************************/
    /* root of (sub)tree */
    if ( (nd->leaf != 1) && (nd->parent1 != NULL) &&
	 (nd->processed_yn == 0) ) 
    {
      if (nd->hybrid != 0)
      {
        fprintf(fp,")%ld#H:%f",nd->id, nd->brlen1);
      }
      else
      {
        fprintf(fp,")%ld:%f",nd->id, nd->brlen1);
      }

      // mark this leaf/node as printed
      // (this marker goes up to 2 as hybrid nodes get printed twice)
      nd->processed_yn = nd->processed_yn + 1;
    }
  }
}

/*****************************************************************************/
/*                          OldPrintInternalNodes                            */
/*****************************************************************************/
void OldPrintInternalNodes(FILE *fp, NODE nd)
{
  int hchromo_i, strand_i;

  if (nd->processed_yn == 0)
  {
    if ((nd->left != NULL)  || (nd->right != NULL))
    {   
      nd->processed_yn = 1;

      fprintf(fp,"n %ld %d %d %d %d\n", nd->id, nd->hybrid, nd->num_of_hom_chromos, nd->num_of_strands_per_hc, nd->seq_length);

      for(hchromo_i = 0; hchromo_i < nd->num_of_hom_chromos; hchromo_i++)
      {
        for(strand_i = 0; strand_i < nd->num_of_strands_per_hc; strand_i++)
        {
          fprintf(fp,"%s\n",nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
	}
      }
    }
  }

  if (nd->left != NULL)
    OldPrintInternalNodes(fp, nd->left);
  if (nd->right != NULL)
    OldPrintInternalNodes(fp, nd->right);
}

/*****************************************************************************/
/*                           OldPrintEdges                                   */
/*****************************************************************************/
void OldPrintEdges(FILE *fp, NODE nd)
{
  if ( (nd->processed_yn == 0) && nd->left != NULL)
  {
    if (nd->left->parent1 == nd)
    {
      fprintf(fp,"a %ld %ld %lf\n", nd->id, nd->left->id, nd->left->brlen1);
    }
    else
    {
      fprintf(fp,"a %ld %ld %lf\n", nd->id, nd->left->id, nd->left->brlen2);
    }
    OldPrintEdges(fp, nd->left);
  }
  if ( (nd->processed_yn == 0) && nd->right != NULL)
  {
    if (nd->right->parent1 == nd)
    {
      fprintf(fp,"a %ld %ld %lf\n", nd->id, nd->right->id, nd->right->brlen1);
    }
    else
    {
      fprintf(fp,"a %ld %ld %lf\n", nd->id, nd->right->id, nd->right->brlen2);
    }
    OldPrintEdges(fp, nd->right);
  }
  if ((nd->processed_yn == 1) && (nd->left!=NULL))
    OldPrintEdges(fp, nd->left);
  if ((nd->processed_yn == 1) && (nd->right!=NULL))
    OldPrintEdges(fp, nd->right);

  nd->processed_yn = 1;
}
