/*****************************************************************************/
/*                                                                           */
/*  NSGW kappa version                                                       */
/*  Wrapper code for Seq-Gen in order to request/process sequences between   */
/*  NetGen and Seq-gen.                                                      */
/*  Copyright (C) 2007  Monique M. Morin                                     */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>

#include "globals.h"
#include "network.h"
#include "interface.h"
#include "utilities.h"
#include "twister.h"
#include "sequences.h"


/*****************************************************************************/
/*                         GetNetworkSequences                               */
/*****************************************************************************/
void GetNetworkSequences(char *outfile, char seq_gen_options[], NETWORK nt) 
{
   int node_i, nos, sl, nhc, nsphc, hchromo_i, strand_i;


  /*****************************************/
  /*           do the root                 */
  /*****************************************/
  // note: if doing "G" test, do not want to re_index here and this is not
  // the outgroup so the last two args to GetSeqGenSequence are both 0's
  //
  // note: we get a sequence for each homologous chromosome, and all strands in
  // the same homologous chromosome match strand 0

  nhc = nt->root->num_of_hom_chromos;
  nsphc = nt->root->num_of_strands_per_hc;
  nos = nhc * nsphc;
  sl = nt->root->seq_length;

  // for the root, we really only need the sequence for the first strand
  // for every homologous chromosome (since we want all root strands to 
  // be identical within each homologous chromosome), but we need to go 
  // through the process of calling
  // Seq-Gen for each sequence in order to keep validation testing working
  // correctly and the couple of extra calls here for the root are not a big 
  // performance hit -- afterwards we can do the correct copying
  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {
      GenerateSeqGenInputFile(outfile, 0, nos, sl, NULL, 
	  	              nt->root->hom_chromos[0]->strands[0]->sequence,
			      0, -1);

      GetSeqGenSequence(outfile, seq_gen_options, 0, sl, NULL, -1, 
		        nt->root->hom_chromos[0]->strands[0]->sequence, 0, 0);
    }

    // now that we have done all the calls, use the seq for strand 0, for
    // all the strands in this homologous chromosome
    for (strand_i = 1; strand_i < nsphc; strand_i++)
    {
     strcpy(nt->root->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
	    nt->root->hom_chromos[hchromo_i]->strands[0]->sequence);
    }

  }

  // mark root as processed
  nt->root->processed_yn = 1;

  /*****************************************/
  /*    sub-root if we have an outgroup    */
  /*****************************************/
  //
  // if we have an outgroup, the topology will already reflect
  // the 'super root' which has two kids -- the 'sub root' and the
  // outgroup leaf -- by merely having all the nodes in the input
  // file
  //
  // so we do not need to do anything special for the sub root
  // because it is already present with a branch length of 0
  // between it and the super root
  //
  // we do need to care about how the outgroup leaf is handled
  // though because we want it to meet the parameters of similarity
  // and/difference which were read in as part of the input
  // this is dealt with as part of the following loop


  /**************************************************************/
  /*   now that we have root seqs do the rest of the topology   */
  /**************************************************************/

  ResetProcessed(nt);
  nt->root->processed_yn = 1;

  // REGULAR PROCESSING 
  if (rand_tree_test_yn == 0)
  {
    for (node_i = 0; node_i < nt->num_of_nodes; node_i++)
    {
      if (nt->ndarray[node_i]->processed_yn == 0)
      {
	sl = nt->ndarray[node_i]->seq_length;
	nos = nt->ndarray[node_i]->num_of_hom_chromos * 
	      nt->ndarray[node_i]->num_of_strands_per_hc;

        if (nt->outgroup_id == nt->ndarray[node_i]->id)
        {
          ProcessOutgroupSequence(nt, nt->ndarray[node_i], outfile, 
	        		  seq_gen_options, sl, nos); 
        }
        else
        {
          ProcessNodeSequence(nt->ndarray[node_i], outfile, seq_gen_options, 
	                      sl, nos);
        }
      }
    }
  }

  // IRREGULAR PROCESSING 
  // if doing rand_num_tree_test, must process the 
  // nodes in the same order as NG
  if (rand_tree_test_yn == 1)
  {
    ProcessInNGOrder(outfile, seq_gen_options, nt);
  }

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUG:  we are done with GetNetworkSequences\n");

}


/******************************************************************************/
/*                           ProcessInNGOrder                                 */
/******************************************************************************/
void ProcessInNGOrder(char *outfile, char seq_gen_options[], NETWORK nt)
{
  int node_i, ng_node_i, sl, nos;

  // if we are in this routine, means we must process the nodes in the 
  // same order as NG did so that the random numbers will match up and 
  // SG will give the same sequences;  rand_num_node_tree_test has the 
  // order of nodes to do

  // start this indexing at 1 as root is already processed and in this testing
  // scenario, we are guaranteed the root was 0
  for (node_i = 1; node_i < nt->num_of_nodes; node_i++)
  {
    // find the node id to be processed 
    ng_node_i = rand_num_node_tree_test[node_i];

    // DEBUGGING
    //fprintf(stderr, "NSGW DEBUGGING: just pulled which node to process it as %d\n", ng_node_i);
    
    sl = nt->ndarray[ng_node_i]->seq_length;

    nos = nt->ndarray[ng_node_i]->num_of_hom_chromos * 
	  nt->ndarray[ng_node_i]->num_of_strands_per_hc;

    if (nt->ndarray[ng_node_i]->processed_yn == 0)
    {
      if (nt->outgroup_id == nt->ndarray[ng_node_i]->id)
      {
        ProcessOutgroupSequence(nt, nt->ndarray[ng_node_i], outfile, 
	        		seq_gen_options, sl, nos); 

      }
      else
      {
        ProcessNodeSequence(nt->ndarray[ng_node_i], outfile, seq_gen_options, 
	                    sl, nos);
      }
    }

    //fprintf(stderr, "NSGW DEBUG: just finished node_i %d in NG processing order %d\n", ng_node_i, node_i);

  } // next node
}

/*****************************************************************************/
/*                         ProcessNodeSequence                               */
/*****************************************************************************/
void ProcessNodeSequence(NODE nd, char *outfile, char seq_gen_options[], 
		         int seq_length, int num_of_seqs)
{
  // this should always be the case, but just being careful
  if (nd->processed_yn == 0)
  {
    if (nd->hybrid == 0) // tree node -- 1 parent
    {
      // see if the parent has its sequences yet, if not get before
      // proceeding in this routine, otherwise we can proceed to next if
      if (nd->parent1->processed_yn == 0)
      {
        ProcessNodeSequence(nd->parent1, outfile, seq_gen_options, seq_length,
			    num_of_seqs);
      }

      // now we are ready to process the tree node
      ProcessTreeNodeSequence(nd, outfile, seq_gen_options, seq_length);

    } // done processing tree node cases

    else if (nd->hybrid == 1) // diploid hybridization -- 2 parents

    {
      // check to see if BOTH parents have their seqs and if not, get them Both!
      if (nd->parent1->processed_yn == 0)
      {
        ProcessNodeSequence(nd->parent1, outfile, seq_gen_options, seq_length,
			    num_of_seqs);
      }
      if (nd->parent2->processed_yn == 0)
      {
        ProcessNodeSequence(nd->parent2, outfile, seq_gen_options, seq_length,
			    num_of_seqs);
      }

      // now we are ready to process the diploid node
      ProcessDipNodeSequence(nd, outfile, seq_gen_options, seq_length, num_of_seqs);
    }

    else if (nd->hybrid == 2) // polyploid hybridization -- 2 parents

    {
      // check to see if BOTH parents have their seqs and if not, get them Both!
      if (nd->parent1->processed_yn == 0)
      {
        ProcessNodeSequence(nd->parent1, outfile, seq_gen_options, seq_length,
			    num_of_seqs);
      }
      if (nd->parent2->processed_yn == 0)
      {
        ProcessNodeSequence(nd->parent2, outfile, seq_gen_options, seq_length,
			    num_of_seqs);
      }

      // now we are ready to process the polyploid node
      ProcessPolyNodeSequence(nd, outfile, seq_gen_options, seq_length, num_of_seqs);
    }

    else // other types of hybridization

    {
      fprintf(stderr, "NSGW ERROR: node is something other than a tree, diploid hybrid, or polyploid hybrid node\n");
      fprintf(stderr, "NSGW ERROR: other hybridizations are not yet implemented\n");
      exit(0);
    }
  } 
  else
  {
    fprintf(stderr, "NSGW ERROR: logical error processing the topology\n");
    exit(0);
  }
}

/*****************************************************************************/
/*                       ProcessTreeNodeSequence                             */
/*****************************************************************************/
void ProcessTreeNodeSequence(NODE nd, char *outfile, char seq_gen_options[], 
		             int seq_length)
{
   int hchromo_i, strand_i, re_index_yn;
   int nos;

  // we can go ahead and process this node because its parent 
  // knows its sequence which we can mutate with SeqGen
  if (nd->parent1->processed_yn == 1)
  {
    for (hchromo_i = 0; hchromo_i < nd->num_of_hom_chromos; hchromo_i++)
    {
      for (strand_i = 0; strand_i < nd->num_of_strands_per_hc; strand_i++)
      {
        // note: if doing "G" test here, want to re_index when seq_i==0
        //       but there is no outgroup so the last arg to GetSeqGenSequence
        //       here is always 0
        if ( (hchromo_i == 0) && (strand_i == 0) )
          re_index_yn = 1;
        else
          re_index_yn = 0;

        // note: NULL and -1 are given since we have the specific sequence
        nos = nd->num_of_hom_chromos * nd->num_of_strands_per_hc;
        GenerateSeqGenInputFile(outfile, 1, nos, nd->seq_length, NULL,
	           	        nd->parent1->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
			        nd->brlen1, -1);
        GetSeqGenSequence(outfile, seq_gen_options, 1, nd->seq_length, NULL, -1,
			  nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
			  re_index_yn, 0);

        //fprintf(stderr, "NSGW DEBUGGING: just set %s for node id %ld\n", nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, nd->id);
	
      } // next strand
    } // next homologous chromosome 

    // mark this node as processed so we don't redo it!
    nd->processed_yn = 1;
  }
}

/*****************************************************************************/
/*                       ProcessDipNodeSequence                              */
/*****************************************************************************/
void ProcessDipNodeSequence(NODE nd, char *outfile, char seq_gen_options[], 
	                    int seq_length, int num_of_seqs)
{
  int p1_num_strands_gives, p2_num_strands_gives;
  int nhc, nsphc, hchromo_i; 

  // DEBUGGING
  //int strand_i;

  nhc = nd->num_of_hom_chromos;
  nsphc = nd->num_of_strands_per_hc;

  // first see if our strand number is even -- otherwise, diploid
  // hybridization does not make sense to us
  if ( (nsphc%2) != 0 )
  {
    fprintf(stderr, "NSGW ERROR:  Number of strands (%d) for a hybrid is not an even number -- this is an undefined situation\n", nsphc);
    fprintf(stderr, "NSGW ERROR:  the two parents involved have ids of %ld and %ld\n", nd->parent1->id, nd->parent2->id);
    exit(0);
  }

  // now that we know we have an even strand number, we can establish
  // how many strands for each homologous chromosome will come from each parent;
  // note that for diploid hybridizations this is always a 50/50 split
  // per homologous chromosome contents

  p1_num_strands_gives = (int) (nsphc * 0.5);
  p2_num_strands_gives = p1_num_strands_gives;

  // make sure these add up right!
  if ( (p1_num_strands_gives + p2_num_strands_gives) != nsphc)
  {
    fprintf(stderr, "NSGW ERROR: We have a problem with doing a diploid hybridization\n");
    fprintf(stderr, "NSGW ERROR: number of strands do not add up for the hybrid id %ld\n", nd->id);
    exit(0);
  }
 
  // make sure the homologous chromosomes and strands between the 
  // two parents match
  if ( (nd->parent1->num_of_hom_chromos != nd->parent2->num_of_hom_chromos) ||
       (nd->parent1->num_of_strands_per_hc != nd->parent2->num_of_strands_per_hc) )
  {
    fprintf(stderr, "NSGW ERROR: There is a problem with this diploid hybrid node\n");
    fprintf(stderr, "NSGW ERROR: its parents do not agree in their number of seqs(either number of homologous chromosomes or strands is off)\n");
    exit(0);
  }

  /***************************************************************************/
  /*                                                                         */
  /*  For EACH homologous chromosome, half of the sequences come from each   */
  /*  of the two parents -- resulting in the same total                      */
  /*  number of strand sequences for the hybrid in the end.                  */
  /*  However, the selection of strand sequences for each                    */
  /*  homologous chromosome composition is random.                           */
  /*                                                                         */
  /***************************************************************************/

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    // do the first half of sequences which will come from p1
    RandomFillStrands(nd, nd->parent1, hchromo_i, 0, p1_num_strands_gives);
    
    // do the second half of sequences which will come from p2
    RandomFillStrands(nd, nd->parent1, hchromo_i, p1_num_strands_gives, nsphc);
  }

  // we are now done with this node
  nd->processed_yn = 1;

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUG: done assigning seqs for hybrid %ld\n", nd->id);
  //
  //for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nsphc; strand_i++)
    //{
      //fprintf(stderr, "NSGW DEBUG: hchromo_i %d, strand_i %d is %s \n", hchromo_i, strand_i, nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
 // END DEBUGGING
 
}

/*****************************************************************************/
/*                       RandomFillStrands                                   */
/*****************************************************************************/
void RandomFillStrands(NODE nd, NODE parent_ptr, int hchromo_i,
		          int start_strand, int end_strand)
{
  int strand_i;
  int rand_index, used_index_yn;
  int used_indices[MAX_NUM_STRANDS_PER_HC];

  /***************************************************************/
  /*  Here we select the correct number of sequences from the    */
  /*  specified parent.  We do this as random selection          */
  /*  without replacement and track which sequences are chosen,  */
  /*  as a sequence can be chosen only once.  This keeps the     */
  /*  random selection fair as we do not adjust the size of the  */
  /*  pool, simply rechoose if a duplicate is chosen.            */
  /***************************************************************/

  // initialize the tracking array
  for (strand_i = 0; strand_i < nd->num_of_strands_per_hc; strand_i++)
  {
    used_indices[strand_i] = -1;
  }


  // do the contribution
  for (strand_i = 0; strand_i < end_strand; strand_i++)
  {
    // set default so we enter the loop the first time
    used_index_yn = 1;

    while (used_index_yn)
    {
      // the following gives a rand in the [0,1) range multiplied by strand
      // number and finally cast to an int -- this in effect gives us a rand
      // int in the range of [0,num_of_strands_per_hc-1] which is how we want 
      // to select
      rand_index = (int) (genrand_real2()*(parent_ptr->num_of_strands_per_hc));

      if (used_indices[rand_index] == 1)
        used_index_yn = 1;
      else
	used_index_yn = 0;
    }

    // record the rand_index chosen this round
    used_indices[rand_index] = 1;

    // copy the chosen sequence into the node
    strcpy(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
	   parent_ptr->hom_chromos[hchromo_i]->strands[rand_index]->sequence);

    // record which parent gave the sequence
    nd->hom_chromos[hchromo_i]->strands[strand_i]->ancestor_id = parent_ptr->id;
  } // next strand
}

/*****************************************************************************/
/*                      ProcessPolyNodeSequence                              */
/*****************************************************************************/
void ProcessPolyNodeSequence(NODE nd, char *outfile, char seq_gen_options[], 
	                     int seq_length, int num_of_seqs)
{
  int nhc;
  int p1_nhc;
  int p2_nhc;
  int nsphc;
  int p1_nsphc;
  int p2_nsphc;
  int hchromo_i; 
  int strand_i;
  int p2_hchromo;

  nhc = nd->num_of_hom_chromos;
  nsphc = nd->num_of_strands_per_hc;

  p1_nhc = nd->parent1->num_of_hom_chromos;
  p2_nhc = nd->parent2->num_of_hom_chromos;

  p1_nsphc = nd->parent1->num_of_strands_per_hc;
  p2_nsphc = nd->parent2->num_of_strands_per_hc;

  // for polyploid hybridization the hybrid node gets all
  // homologous chromosomes/strands for both parents so make
  // sure the values add up/match correctly

  if (nhc != (p1_nhc + p2_nhc) )
  {
    fprintf(stderr, "NSGW ERROR:  Polyploid hybrid number of homologous chromosomes is not the sum of the parents' homologous chromosomes (%d vs %d)\n", nhc, (p1_nhc+p2_nhc));
  }

  if ( (nsphc != p1_nsphc) || (nsphc != p2_nsphc) )
  {
    fprintf(stderr, "NSGW ERROR:  Polyploid hybrid number of strands do not match the parental values for number of strands (%d vs %d and %d)\n", nsphc, p1_nsphc, p2_nsphc);
  }

  // now for polyploid hybridization we know that the
  // hybrid node gets all from both parents so go
  // ahead and do the copying

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    // deal with parent 1 sequences
    if (hchromo_i < nd->parent1->num_of_hom_chromos)
    {
      for (strand_i = 0; strand_i < nsphc; strand_i++)
      {
        strcpy(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
	       nd->parent1->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      }
    }
    else // deal with parent 2 sequences
    {
      p2_hchromo = hchromo_i - nd->parent1->num_of_hom_chromos;     

      for (strand_i = 0; strand_i < nsphc; strand_i++)
      {
        strcpy(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
	       nd->parent2->hom_chromos[p2_hchromo]->strands[strand_i]->sequence);
      }
    }
  }

  // we are now done with this node
  nd->processed_yn = 1;

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUG: done assigning seqs for polyploid hybrid %ld\n", nd->id);
  //
  //for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nsphc; strand_i++)
    //{
      //fprintf(stderr, "NSGW DEBUG: hchromo_i %d, strand_i %d is %s \n", hchromo_i, strand_i, nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
 // END DEBUGGING
 
}


/*****************************************************************************/
/*                       ProcessOutgroupSequence                             */
/*****************************************************************************/

void ProcessOutgroupSequence(NETWORK nt, NODE nd, char *outfile, 
		             char seq_gen_options[], int seq_length, 
			     int num_of_seqs)
{
  // this should always be the case, but just being careful
  if (nd->processed_yn == 0)
  {
    // see if the parent (the root in this case)  has its sequences yet, 
    // if not this is a processing error so stop
    if (nd->parent1->processed_yn == 0)
    {
      fprintf(stderr, "NSGW ERROR:  Root was not processed for sequences -- cannot proceed\n");
    }

    // now we are ready to process the outgroup node
    AssignOutgroupSequences(nt, nd, outfile, seq_gen_options, seq_length);

  }
  //fprintf(stderr, "NSGW DEBUGGING: done with ProcessOutgroupSequence\n");
}


/*****************************************************************************/
/*                       AssignOutgroupSequences                             */
/*****************************************************************************/
void AssignOutgroupSequences(NETWORK nt, NODE nd, char * outfile, 
		             char seq_gen_options[], int seq_length)
{
  // here we do the work of assigning sequences to the outgroup
  // the big issue is that we want a leaf that is somewhat similar to the
  // leaves of the phylogeny, but not too similar
  //
  // max_tries tells us the maximum number of sequences to try in assigning
  // each sequence; while max_outg_diff_value and min_outg_diff_value give
  // us the range for how dissimilar/similar the new sequence should be

  int seq_i, try_i;                 // loop variables
  int hchromo_i, strand_i;          // more loop variables
  int outg_found_yn;                // found flag
  int best_avg_h_dist, best_try;    // tracking variables

  // to store the tried sequences
  char * outg_try[MAX_OUTGROUP_TRIES]; 

  // to store the final sequences decided upon
  char * final_sequences[MAX_NUM_HOM_CHROMOS * MAX_NUM_STRANDS_PER_HC];

  // to store avg dist of tried seqs to all other extant taxa
  int * avg_h_dist_value; 

  // to track whether the tried sequence under study meets the range
  // for each extant taxon (-1/1 entry)
  int * meets_goal;       

  // to make things shorter and clearer
  NODE outg = nd;
  int nhc = outg->num_of_hom_chromos;
  int nsphc = outg->num_of_strands_per_hc;
  int num_of_seqs = nhc * nsphc; 

  // setup dynamically allocated stuff
  avg_h_dist_value = (int *) calloc(MAX_OUTGROUP_TRIES, sizeof(int));
  meets_goal = (int *) calloc(MAX_NODES, sizeof(int));

  if ( (avg_h_dist_value == NULL) || (meets_goal == NULL))
  {
    fprintf(stderr, "NSGW ERROR: Could not allocate space for outgroup processing\n");
    exit(0);
  }

  // set up memory for storing the tries
  for (try_i = 0; try_i < nt->max_outg_tries; try_i++)
  {
    outg_try[try_i] = (char *) calloc((MAX_SEQ_LENGTH+2),sizeof(char));
    if (outg_try[try_i] == NULL)
    {
      fprintf(stderr, "NSGW ERROR: Could not allocate space for tried sequences as part of the outgroup work\n");
      exit(0);
    }
  }

  // set up the memory for the final sequences
  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    final_sequences[seq_i] = (char *) calloc((MAX_SEQ_LENGTH+2),sizeof(char));
    
    if (final_sequences[seq_i] == NULL)
    {
      fprintf(stderr, "NSGW ERROR: Could not allocate space for final sequences as part of the outgroup work\n");
      exit(0);
    }
  }

  // DEBUGGING
  //fprintf (stderr, "NSGW DEBUGGING: Starting Outgroup Sequence Work\n");
  // END DEBUGGING

  seq_i = 0;

  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {

      // DEBUGGING
      //fprintf(stderr, "NSGW DEBUGGING: Working out outgroup seq %d\n", seq_i);
      // END DEBUGGING

      outg_found_yn = 0;

      for (try_i = 0; try_i < nt->max_outg_tries; try_i++)
      {
	outg_found_yn = MakeOutgroupTry(nt, outg, outfile, seq_gen_options,
			                try_i, &best_try, hchromo_i, strand_i, 
					seq_i, outg_try, meets_goal, 
					avg_h_dist_value);
        if (outg_found_yn)
          break;

        // DEBUGGING
        //fprintf(stderr, "NSGW DEBUGGING: Doing another try for seq_i %d \n", seq_i);
        // END DEBUGGING

      } // do next try


      // if outgroup not found, choose the 'best' one of the tries
      if (outg_found_yn == 0)
      {
        for (try_i = 0; try_i < nt->max_outg_tries; try_i++)
        {
          // cast avg_h_dist_value as an int so we can compare properly
          if ( (((int) avg_h_dist_value[try_i]) >= (nt->min_outg_diff_value)) &&
               (((int) avg_h_dist_value[try_i]) <= (nt->max_outg_diff_value)))
          {
            outg_found_yn = 1;
            best_try = try_i;
	    break;
	  }
        }
      }

      // if still have not found outgroup, pick the one with the highest
      // average hamming distance
      if (outg_found_yn == 0)
      {
        best_avg_h_dist = 0;
        best_try = 0;
        for (try_i = 0; try_i < nt->max_outg_tries; try_i++)
        {
          if (avg_h_dist_value[try_i] > best_avg_h_dist)
	  {
            best_try = try_i;
	    best_avg_h_dist = avg_h_dist_value[try_i];
	  }
        }
	// we are guaranteed to have found something!
	outg_found_yn = 1;
      }

      // now take our best found (from any of the above methods) and
      // put it into the final sequences array

      strcpy(final_sequences[seq_i], outg_try[best_try]);

      // DEBUGGING
      //fprintf(stderr, "NSGW DEBUGGING:  We have our final %d seq as:%s\n", seq_i, final_sequences[seq_i]);
      //fprintf(stderr, "NSGW DEBUGGING:  Our best try was %d and its avg hamming distance was %d \n", best_try, avg_h_dist_value[best_try]);
      //fprintf(stderr, "NSGW DEBUG:  The length of outg_try[best_try] was %d \n", strlen(outg_try[best_try]));
      //
      //fprintf(stderr, "NSGW DEBUGGING:  max_outg_tries is %d \n", nt->max_outg_tries);
      //for (try_i = 0; try_i < nt->max_outg_tries; try_i++)
      //{
        //fprintf(stderr, "NSGW DEBUGGING:  The length of outg_try[%d] was %d \n", try_i, strlen(outg_try[try_i]));
      //}
      // END DEBUGGING
 
      seq_i++;

    } // next strand
  } // next homologous chromosome


  // now that we are done with all the processing, we can put the final
  // good sequences into the node's sequences (could not do earlier because
  // each call to Assign Sequences overwrote the node's sequences)

  seq_i = 0;
  for (hchromo_i = 0; hchromo_i < nhc; hchromo_i++)
  {
    for (strand_i = 0; strand_i < nsphc; strand_i++)
    {
      strcpy(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, 
	     final_sequences[seq_i]);

      // DEBUGGING
      //fprintf(stderr, "NSGW DEBUGGING: end of outg work hchromo %d, strand %d ns:%s, fs:%s\n", hchromo_i, strand_i,  nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence, final_sequences[seq_i]);
      //fprintf(stderr, "NSGW DEBUGGING: end of outg work hchromo %d, strand %d, length of ns:%d, length of fs:%d\n", hchromo_i, strand_i, strlen(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence), strlen(final_sequences[seq_i]));
      // END DEBUGGING

      seq_i++;
    }
  }

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUGGING:  After outgroup processing our outgroup has:\n");
  //for (hchromo_i = 0; hchromo_i < nd->num_of_hom_chromos; hchromo_i++)
  //{
    //for (strand_i = 0; strand_i < nd->num_of_strands_per_hc; strand_i++)
    //{
      //fprintf(stderr, "NSGW DEBUGGING:%s\n",nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    //}
  //}
  //fprintf(stderr, "NSGW DEBUGGING: done with outgroup processing really\n");
  // END DEBUGGING

  // be good and clean up
  free(avg_h_dist_value);
  free(meets_goal);
  for (seq_i = 0; seq_i < num_of_seqs; seq_i++)
  {
    free(final_sequences[seq_i]);
  }
  for (try_i = 0; try_i < nt->max_outg_tries; try_i++)
  {
    free(outg_try[try_i]);
  }

}

/******************************************************************************/
/*                           MakeOutgroupTry                                  */
/******************************************************************************/
int MakeOutgroupTry(NETWORK nt, NODE outg, char * outfile, 
		    char seq_gen_options[], int try_i, int *best_try, 
		    int hchromo_i, int strand_i, int seq_i, char *outg_try[], 
		    int *meets_goal, int *avg_h_dist_value)
{
  int node_i, h_dist;

  int outg_found_yn = 0;
  int all_h_dist_sum = 0;
  int seq_index = hchromo_i * outg->num_of_strands_per_hc + strand_i;
  int num_of_seqs = outg->num_of_hom_chromos * outg->num_of_strands_per_hc;
  

  // first get a set of potential sequences
  //
  // note: if doing test "G" we always want to re_index here for
  //       the outgroup and we need to tell it we have the outgroup
  //       so the last two args to GetSeqGenSequence are both 1's
  //
  // note: we need to tell GetSeqGenSequence which SEQUENCE we are
  //       on so that we index the random numbers correctly if
  //       doing a comparison test, so calculate which one we
  //       are on by hchromo_i * num_of_strands_per_hc + strand_i and 
  //       give this value
	
	
  GenerateSeqGenInputFile(outfile, 1, num_of_seqs, outg->seq_length, NULL, outg->parent1->hom_chromos[hchromo_i]->strands[strand_i]->sequence, outg->brlen1, -1);

  GetSeqGenSequence(outfile, seq_gen_options, 1, outg->seq_length, NULL, 
		    seq_index, 
		    outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
		    1, 1);

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUGGING:  The parent sequence given to GenerateSeqGenInputFile was %s\n", outg->parent1->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
  // END DEBUGGING

  // copy the specific sequence from node into outg_try (because on the 
  // next try round the node's seq will be rewritten in AssignSequences)
  // this is because AssignSequences does all sequences and we want 
  // to deal with sequences one at a time here
      
  strcpy(outg_try[try_i], 
	 outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence);


  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUGGING: just copied ndarray sequence(length %d) into outg_try (length %d)\n", strlen(outg_try[try_i]), strlen(outg->hom_chromos[hchromo_i]->strands[strand_i]->sequence));
  // END DEBUGGING
  
  // first initialize meet goal so we don't get screwed up in 
  // the next part of the code
  for (node_i = 0; node_i < (nt->num_of_nodes+1); node_i++)
  {
    meets_goal[node_i] = 0;
  }

  // now go through all the leafs of the tree/network and do distance work
  // (but don't compare it to itself which is the last entry in ndarray!)
  // only do comparisons to extant taxa, not 'all leaves' as these may include
  // nodes that died out prior to the end of the simulation
      
  for (node_i = 0; node_i < (nt->num_of_nodes-1); node_i++)
  {
    if (nt->ndarray[node_i]->extant)
    {
      // calc hamming distance between the two sequences
      h_dist = HammingDistance(outg_try[try_i], 
	     	               nt->ndarray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->sequence);


      // track
      all_h_dist_sum = all_h_dist_sum + h_dist;

      //fprintf(stderr, "NSGW DEBUGGING: all_h_dist_sum is now %d \n", all_h_dist_sum);

      // compare to see if dist meets min/max restrictions and track
      if ( (h_dist <= nt->max_outg_diff_value) && 
           (h_dist >= nt->min_outg_diff_value)) 
      {
        meets_goal[node_i] = 1;
      }
      else
      {
        meets_goal[node_i] = -1;
      }
    } // end of checking this extant taxon
  } // keep going through and looking for extant taxa

  // store avg_dist and keep all as integers to keep things easy -- remember
  // that outgroup leaf is not counted as part of num_of_extant_taxa; realize
  // this truncates avg dist but this is a minor factor when doing large strings
  avg_h_dist_value[try_i] = (int) ((all_h_dist_sum)/(nt->num_of_extant_taxa));

  // DEBUGGING
  //fprintf(stderr, "NSGW DEBUGGING: We just calc avg ham dist as %d \n", avg_h_dist_value[try_i]);
  // END DEBUGGING

  // check to see if all entries of meet_goal are met -- if yes,
  // we have our sequence_i and can stop -- so assume we have it and
  // see if something does not meet our goal (note that the array
  // was initialized to 0 so the nodes that are not leafs have a 0
  // so look for -1s to be the problem)
      
  outg_found_yn = 1;

  for (node_i = 0; node_i < (nt->num_of_nodes-1); node_i++)
  {
    if (meets_goal[node_i] < 0)
      outg_found_yn = 0;
  } 

  // if we have our sequence assiged for seq_i, set vars and continue on
  // with rest of processing (do not keep doing tries for this sequence_i)
  if (outg_found_yn)
    (*best_try) = try_i;

  return (outg_found_yn);
}
